/*******************************************************************************

    File:     action.c
    Author:   Rick Cooper
    Created:  03/04/07
    Edited:   03/04/07
    Contents: Generic action handling functions

    Public procedures:

*******************************************************************************/

#include "ss.h"
#include "wcst.h"
#include "tol.h"
#include "rng.h"

/******************************************************************************/
/* Section 1: Utility functions ----------------------------------------------*/

static double holdsNothing(ContentionScheduler *cs, Effector *effector)
{
    return( effector->itemsHeld == NULL ? 1.0 : 0.0 );
}

/*----------------------------------------------------------------------------*/

static double holdsAnything(ContentionScheduler *cs, Effector *effector)
{
    return( effector->itemsHeld == NULL ? 0.0 : 1.0 );
}

/*----------------------------------------------------------------------------*/

static Object *mostActiveObjectOnNthCycle(int cycle, Object *o1, Object *o2, ObjectRole r)
{
    if (r == EMPTY) {
        fprintf(stderr, "WARNING: mostActiveObject() called with EMPTY ObjectRole: NULL returned\n");
        return(NULL);
    }
    else if (o1 == NULL) {
        return(o2);
    }
    else {
        return((o1->activation[r][cycle] > o2->activation[r][cycle])?o1:o2);
    }
}

/*----------------------------------------------------------------------------*/

static Object *bestObjectForRole(ContentionScheduler *cs, ObjectRole role)
{
    Object *best = NULL;
    ObjectList *os;

    for (os = cs->objects; os; os = os->tail) {
        best = mostActiveObjectOnNthCycle(cs->cycle, best, os->head, role);
    }

    return(best);
}

/*----------------------------------------------------------------------------*/

static Effector *bestEffector(ContentionScheduler *cs, double effectorType(ContentionScheduler *, Effector *))
{
    Effector *best = NULL;
    EffectorList *es;
    double max = 0.0, this;

    for (es = cs->effectors; es; es = es->tail) {
        if ((this = effectorType(cs, es->head) * es->head->activation[cs->cycle]) > max) {
            max = this;
            best = es->head;
        }
    }
    /* If nothing found, look again without restrictions */
    if (best == NULL) {
        for (es = cs->effectors; es; es = es->tail) {
            if (es->head->activation[cs->cycle] > max) {
                best = es->head;
                max = best->activation[cs->cycle];
            }
        }
    }
    return(best);
}

/*----------------------------------------------------------------------------*/

/* This function attempts to find a full hand holding something not useful    */
/* to the current schema's selected parent.                                   */

/* If the current schema has no selected parent, or both hands are empty,     */
/* or both hands are full and useful, this function returns the less active   */
/* hand.                                                                      */

/* If both hands are full and not useful to the selected parent, we           */
/* recursively search for usefulness to the selected grandparent etc.         */

static Effector *leastActiveEffector(ContentionScheduler *cs)
{
    EffectorList *el;
    Effector *worst = NULL;

    for (el = cs->effectors; el != NULL; el = el->tail) {
        if ((worst == NULL) || (worst->activation[cs->cycle] > el->head->activation[cs->cycle])) {
            worst = el->head;
        }
    }
    return(worst);
}

static Boolean usesObjectsHeldBy(ContentionScheduler *cs, Schema *this, Effector *effector)
{
    ObjectTriggerFunctionList *tf;
    ObjectList *os;

    for (os = effector->itemsHeld; os != NULL; os = os->tail) {
        for (tf = this->objectTriggers; tf != NULL; tf = tf->next) {
            if (tf->trigger(cs, os->head) > 0) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

static Effector *worstEffectorForSchema(ContentionScheduler *cs, Schema *schema)
{
    Effector *leftHand = cs->effectors->head;
    Boolean leftHandUseful = FALSE;
    Boolean leftHandFull = FALSE;
    Effector *rightHand = cs->effectors->tail->head;
    Boolean rightHandUseful = FALSE;
    Boolean rightHandFull = FALSE;
    Schema *parent;

    if ((schema != NULL) && ((parent = selected_parent(cs, schema)) != NULL)) {
        if (isFullEffectorBoolean(leftHand)) {
            leftHandFull = TRUE;

            if (usesObjectsHeldBy(cs, parent, leftHand)) {
                leftHandUseful = TRUE;
            }
        }

        if (isFullEffectorBoolean(rightHand)) {
            rightHandFull = TRUE;

            if (usesObjectsHeldBy(cs, parent, rightHand)) {
                rightHandUseful = TRUE;
            }
        }

        if (leftHandFull && rightHandFull && !leftHandUseful && !rightHandUseful) {
            return(worstEffectorForSchema(cs, parent));
        }
        else if (leftHandFull && !leftHandUseful) {
            return(leftHand);
        }
        else if (rightHandFull && !rightHandUseful) {
            return(rightHand);
        }
    }
    return(leastActiveEffector(cs));
}

static Schema *selected_basic_schema(ContentionScheduler *cs)
{
    SchemaList *sl;

    for (sl = cs->schemas; sl != NULL; sl = sl->tail) {
        if (sl->head->selected && (sl->head->corresponding_action != ACTION_NONE)) {
            return(sl->head);
        }
    }
    return(NULL);
}

Effector *effectorGetWorst(ContentionScheduler *cs)
{
    return(worstEffectorForSchema(cs, selected_basic_schema(cs)));
}

/*----------------------------------------------------------------------------*/

/* This function is called during action execution to bind an object to a     */
/* functional role.                                                           */

void objectBindToRole(ContentionScheduler *cs, Object *object, ObjectRole role)
{
    char buffer[128];

    if (role == EMPTY) {
        g_snprintf(buffer, 128, "BIND ERROR: Attempt to bind %s to invalid role (%s)", object->name, roleName[role]);
        message(BIND, cs, buffer);
    }
    else if (cs->current_object[role] == NULL) {
        object->inhibit[role] = FALSE; /* Just in case ... */
        cs->current_object[role] = object;
        g_snprintf(buffer, 128, "Binding %s as %s", object->name, roleName[role]);
        message(BIND, cs, buffer);
    }
    else if (cs->current_object[role] != object) {
        g_snprintf(buffer, 128, "BIND ERROR: Cannot bind %s as %s because %s is already bound to that role", object->name, roleName[role], cs->current_object[role]->name);
        message(BIND, cs, buffer);
    }
}

/******************************************************************************/
/* Section 2: Error reporting ------------------------------------------------*/

void errorObjectIncorrect(ContentionScheduler *cs, char *template, char *arg1, char *arg2, char *arg3)
{
    char buffer[256];

    g_snprintf(buffer, 256, template, arg1, arg2, arg3);
    action_log_add_warning(cs, "OBJECT ERROR", buffer);
    fprintf(stderr, "%5d\tOBJECT ERROR: %s\n", cs->cycle, buffer);
}

void errorEffectorIncorrect(ContentionScheduler *cs, char *template, char *e1n, char *e2n)
{
    char buffer[256];

    g_snprintf(buffer, 256, template, e1n, e2n);
    action_log_add_warning(cs, "EFFECTOR ERROR", buffer);
    fprintf(stderr, "%5d\tEFFECTOR ERROR: %s\n", cs->cycle, buffer);
}

void errorObjectState(ContentionScheduler *cs, char *template, char *arg1, char *arg2, char *arg3)
{
    char buffer[256];

    g_snprintf(buffer, 256, template, arg1, arg2, arg3);
    action_log_add_warning(cs, "OBJECT STATE ERROR", buffer);
    fprintf(stderr, "%5d\tSTATE ERROR: %s\n", cs->cycle, buffer);
}

void errorEffectorState(ContentionScheduler *cs, char *template, char *effector)
{
    char buffer[256];

    g_snprintf(buffer, 256, template, effector);
    action_log_add_warning(cs, "EFFECTOR STATE ERROR", buffer);
    fprintf(stderr, "%5d\tEFFECTOR STATE ERROR: %s\n", cs->cycle, buffer);
}

void errorObjectNull(ContentionScheduler *cs, char *role, char *action)
{
    char buffer[256];

    g_snprintf(buffer, 256, "%s of %s", role, action);
    action_log_add_warning(cs, "NULL OBJECT ERROR", buffer);
    fprintf(stderr, "%5d\tNULL OBJECT ERROR: %s\n", cs->cycle, buffer);
}

void errorEffectorNull(ContentionScheduler *cs, char *action)
{
    action_log_add_warning(cs, "NULL EFFECTOR ERROR", action);
    fprintf(stderr, "%5d\tNULL EFFECTOR ERROR: %s\n", cs->cycle, action);
}

void actionMessage(ContentionScheduler *cs, char *action, char *message)
{
    if (messages[RUN] != NULL) {
        fprintf(messages[RUN], "%5d\t%s: %s\n", cs->cycle, action, message);
    }
    action_log_add_action(cs, action, message);
}

/******************************************************************************/
/* Section 4: Recording events -----------------------------------------------*/

void record_intervention(ContentionScheduler *cs, char *action)
{
    FILE *fp;

    if (cs->logfile != NULL) {
        if ((fp = fopen(cs->logfile, "a")) == NULL) {
            fprintf(stderr, "LOG WARNING: Cannot write to log file\n");
        }
        else {
            fprintf(fp, "%4d: intervention %s\n", cs->cycle, action);
            fclose(fp);
        }
    }
}

void record_action(ContentionScheduler *cs, char *action, char *arg1, char *arg2, char *arg3)
{
    FILE *fp;

    if (cs->logfile != NULL) {
        if ((fp = fopen(cs->logfile, "a")) == NULL) {
            fprintf(stderr, "LOG WARNING: Cannot write to log file\n");
        }
        else {
            fprintf(fp, "%4d: %s", cs->cycle, action);
            if (arg1 != NULL) { fprintf(fp, " %s", arg1); }
            if (arg2 != NULL) { fprintf(fp, " %s", arg2); }
            if (arg3 != NULL) { fprintf(fp, " %s", arg3); }
            fprintf(fp, "\n");
            fclose(fp);
        }
    }
}

/******************************************************************************/
/* Section 3: Generic actions: -----------------------------------------------*/

static char buffer[256];

/* effectorDiscardContents(HandContainingObject) -----------------------------*/

/* When an object is discarded the hand no longer holds the object and the    */
/* object's position changes to the position where the object was placed.     */
/* effectorDiscardContents only takes place if the selected hand is non-NULL  */
/* and full, ie no error is registered otherwise.                             */

void effectorDiscardContents(ContentionScheduler *cs, Effector *effector)
{
    Object *theme = NULL;

    if ((effector != NULL) && (effectorGetContents(effector) != NULL)) {
        theme = effectorGetContents(effector);
        g_snprintf(buffer, 256, "Discard %s with %s", theme->name, effector->name);
        actionMessage(cs, "Discard", buffer);
        record_action(cs, "Discard", theme->name, effector->name, NULL);

        traceEffectorPreAction(cs, effector);
        effector->itemsHeld = objectListDeleteElement(effector->itemsHeld, theme);
        traceEffectorPostAction(cs, effector);

        traceObjectPreAction(cs, theme);
        objectSetPosition(theme, effectorDiscardCoordinate(effector), 0.0);
        traceObjectPostAction(cs, theme);
    }
}

/* action_pick_up_theme() ----------------------------------------------------*/

static Boolean action_pick_up_theme(ContentionScheduler *cs, void *task, Object *theme, Effector *effector)
{
    if ((effector == NULL) || isFullEffectorBoolean(effector)) {
        effector = effectorGetWorst(cs);
        effectorDiscardContents(cs, effector);
    }

    if (theme == NULL) {
        errorObjectNull(cs, "Theme", "PickUp");
    }
    else if (isTable(theme) > 0.9) {
        errorObjectState(cs, "%s: desired object (%s) cannot be picked up", "PickUp", theme->name, NULL);
    }
    else if (isHeldBoolean(theme)) {
        errorObjectState(cs, "%s: desired object (%s) is already held by %s", "PickUp", theme->name, theme->holder->name);
    }
    else {
        g_snprintf(buffer, 256, "Picking up %s with %s", theme->name, effector->name);
        actionMessage(cs, "PickUp", buffer);
        record_action(cs, "PickUp", theme->name, effector->name, NULL);
        
        traceEffectorPreAction(cs, effector);
        effectorAddItemToContents(effector, theme);
        traceEffectorPostAction(cs, effector);

        traceObjectPreAction(cs, theme);
        if (cs->task.id == TASK_WCST) {
            wcst_action_pick_up((WcstTaskData *)task, theme, effector);
        }
        else if (cs->task.id == TASK_TOL) {
            tol_action_pick_up((TolTaskData *)task, theme, effector);
        }
        theme->holder = effector;
        traceObjectPostAction(cs, theme);
        objectBindToRole(cs, theme, THEME);
    }
    return(TRUE);
}

/*----------------------------------------------------------------------------*/

static Boolean action_put_down_target(ContentionScheduler *cs, void *task, Effector *effector, Object *target)
{
    Object *theme = NULL;

    if (effector == NULL) {
        errorEffectorNull(cs, "PutDown");
    }
    else if (!isFullEffectorBoolean(effector)) {
        errorEffectorState(cs, "PutDown: effector (%s) is empty", effector->name);
    }
    else {
        while ((theme = effectorGetContents(effector)) != NULL) {
            g_snprintf(buffer, 256, "Putting down %s with %s on %s", theme->name, effector->name, target->name);
            actionMessage(cs, "PutDown", buffer);
            record_action(cs, "PutDown", theme->name, effector->name, NULL);

            traceEffectorPreAction(cs, effector);
            effector->itemsHeld = objectListDeleteElement(effector->itemsHeld, theme);
            traceEffectorPostAction(cs, effector);

            traceObjectPreAction(cs, theme);
            if (cs->task.id == TASK_WCST) {
                wcst_action_put_down((WcstTaskData *)task, theme, target);
            }
            else if (cs->task.id == TASK_TOL) {
                tol_action_put_down((TolTaskData *)task, theme, target);
            }
            traceObjectPostAction(cs, theme);
        }
        objectBindToRole(cs, target, TARGET);
    }
    return(TRUE);
}

static Boolean action_attend_target(ContentionScheduler *cs, void *task)
{
// For the RNG task: get the label of o and stick it in WM, in the form:
// << <status contemplating> <response l> >>

    Boolean r = FALSE;

    if (cs->task.id == TASK_RNG) {
        Object *disk = rng_select_best_target(cs);

        if ((disk != NULL) && (disk->activation[TARGET][cs->cycle] > 0.6)) {
            char *letter = rng_get_disk_label(disk);

            if (letter != NULL) {
                Chunk *cnk = chunk_build("status", "contemplating");
                wm_delete_all(cs, cnk);
                chunk_extend(cnk, "response", letter);
                wm_add(cs, cnk);
                r = TRUE;
            }
        }
        else {
            // Do nothing - basically we're waiting for a target to be activated
        }
    }
    return(r);
}

static Boolean action_disengage_target(ContentionScheduler *cs, void *task)
{
// For the RNG task: get the label of the most active disk. If it is being
// contemplated, then fail. If it isn't, remove anything that is being
// contemplated and succeed. This should cause CS to run until the currently
// contemplated disk is not the most active.

    Boolean r = FALSE;

    if (cs->task.id == TASK_RNG) {
        Object *disk = rng_select_best_target(cs);
        char *letter = rng_get_disk_label(disk);

        if (letter != NULL) {
            Chunk *cnk1, *cnk2, *cnk3;

            cnk1 = chunk_build("status", "contemplating");
            chunk_extend(cnk1, "response", letter);

            cnk2 = chunk_build("status", "generated");
            chunk_extend(cnk2, "response", letter);

            if (wm_match(cs, cnk1) || wm_match(cs, cnk2)) {
                disk->inhibit[TARGET] = TRUE;
                r = FALSE;
            }
            else {
                cnk3 = chunk_build("status", "contemplating");
                wm_delete_all(cs, cnk3);
                r = TRUE;
                chunk_dispose(cnk3);
	    }
            chunk_dispose(cnk1);
            chunk_dispose(cnk2);
        }
    }
    return(r);
}

static Boolean action_select_target(ContentionScheduler *cs, void *task)
{
// For the RNG task: get the label of o and stick it in WM, in the form:
// << <status contemplating> <response l> >>

    Chunk *template, *old_chunk, *new_chunk;
    char *l;

    template = chunk_build("status", "contemplating");
    chunk_extend(template, "response", NULL);

    old_chunk = wm_query(cs, template);
    chunk_dispose(template);

    if ((old_chunk != NULL) && ((l = chunk_get_value(old_chunk, "response")) != NULL)) {
        rng_select_disk(cs, task, l);

        /* Remove any old "generated" records: */
        template = chunk_build("status", "generated");
        wm_delete_all(cs, template);
        chunk_dispose(template);

        /*  Update the "contemplated" to "generated": */
        new_chunk = chunk_build("status", "generated");
        chunk_extend(new_chunk, "response", l);
        wm_delete(cs, old_chunk);
        wm_add(cs, new_chunk);

        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/******************************************************************************/
/* Section 4: Top-level interface --------------------------------------------*/

Boolean execute_action(ContentionScheduler *cs, Schema *schema, void *task)
{
    Boolean r = FALSE;

    switch (schema->corresponding_action) {
        case ACTION_PICK_UP_THEME: {
            r = action_pick_up_theme(cs, task, bestObjectForRole(cs, THEME), bestEffector(cs, holdsNothing));
            break;
        }
        case ACTION_PUT_DOWN_TARGET: {
            r = action_put_down_target(cs, task, bestEffector(cs, holdsAnything), bestObjectForRole(cs, TARGET));
            break;
        }
        case ACTION_ATTEND_TARGET: {
            r = action_attend_target(cs, task);
            break;
	}
        case ACTION_DISENGAGE_TARGET: {
            r = action_disengage_target(cs, task);
            break;
        }
        case ACTION_SELECT_TARGET: {
            r = action_select_target(cs, task);
            break;
        }
        default: {
            fprintf(stderr, "Invalid action: %d\n", schema->corresponding_action);
            break;
        }
    }
    return(r);
}

/******************************************************************************/
