/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper (translated from C++ written by SVW)
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "ss.h"

CsParameters parameter;
FILE *messages[MAX_MESSAGE_TYPES];
char *roleName[MAX_ROLES];

/******************************************************************************/

static Actions *action_list_free(Actions *actions)
{
    while (actions != NULL) {
        Actions *tmp = actions;
        actions = tmp->next;
        free(tmp);
    }
    return(actions);
}

/*----------------------------------------------------------------------------*/

ContentionScheduler *cs_create()
{
    ContentionScheduler *cs;
    ObjectRole role;

    roleName[SOURCE] =            "Source";
    roleName[TARGET] =            "Target";
    roleName[THEME] =             "Theme";
    roleName[IMPLEMENT] =         "Implement";

    parameter.restAct =            0.10;
    parameter.selectionThreshold = 0.60;
    parameter.persistence =        0.87;
    parameter.goalCutoff =         0.50;
    parameter.goalPrecision =    100.00;
    parameter.maths =              TANH;
    parameter.interventions =      TRUE;
    parameter.parental[0] =        0.50;
    parameter.parental[1] =        0.40;
    parameter.parental[2] =        0.40;
    parameter.environmental[0] =   0.10;
    parameter.environmental[1] =   0.10;
    parameter.environmental[2] =   0.00;
    parameter.self[0] =            0.23;
    parameter.self[1] =            0.23;
    parameter.self[2] =            0.23;
    parameter.lateral[0] =         0.46;
    parameter.lateral[1] =         0.46;
    parameter.lateral[2] =         0.46;
    parameter.noiseSD[0] =        0.005;
    parameter.noiseSD[1] =        0.005;
    parameter.noiseSD[2] =        0.005;

    parameter.wm_excite =          0.25;
    parameter.wm_inhibit =         0.75;
    parameter.wm_noise =           0.05;
    parameter.wm_persistence =     0.95;
    parameter.wm_inhibition =      0.00;

    // Supress messages ... could set to stdout;
    messages[RUN] =          NULL;
    messages[SELECTION] =    NULL;
    messages[BIND] =         NULL;
    messages[WARNING] =      NULL;
    messages[WORLD] =        NULL;
    messages[MONITOR] =      NULL;

    cs = (ContentionScheduler *)malloc(sizeof(ContentionScheduler));

    cs->cycle = 0;
    cs->logfile = NULL;
    cs->actions = NULL;
    cs->cl_actions = NULL;
    cs->cl_analysis = NULL;
    cs->cl_schemas = NULL;
    for (role = 0; role < MAX_ROLES; role++) {
        cs->cl_objects[role] = NULL;
    }
    cs->cl_effectors = NULL;

    cs->schemas = NULL;
    cs->objects = NULL;
    cs->effectors = NULL;

    cs->update_goal_queue = NULL;
    cs->selection_queue = NULL;
    cs->deselection_queue = NULL;
    cs->execution_queue = NULL;

    cs->wmes = NULL;
    cs->temporary_schema = NULL;
    cs->strategy = NULL;


    return(cs);
}

/*----------------------------------------------------------------------------*/

Boolean cs_initialise(ContentionScheduler *cs)
{
    ObjectRole role;

    cs->cycle = 0;
    cs->logfile = NULL;
    cs->actions = NULL;
    cs->cl_actions = NULL;
    cs->cl_analysis = NULL;
    cs->cl_schemas = NULL;
    for (role = 0; role < MAX_ROLES; role++) {
        cs->cl_objects[role] = NULL;
    }
    cs->cl_effectors = NULL;

    cs->schemas = NULL;
    cs->objects = NULL;
    cs->effectors = NULL;

    cs->update_goal_queue = NULL;
    cs->selection_queue = NULL;
    cs->deselection_queue = NULL;
    cs->execution_queue = NULL;

    cs->wmes = NULL;
    cs->temporary_schema = NULL;
    cs->strategy = NULL;

    random_seed_load("random_seed", cs);

    fprintf(stderr, "%5d\tInitialising... [Task: %s] ", cs->cycle, task_names[cs->task.id]); fflush(stderr);

    initialise_schemas(cs);
    initialise_objects(cs);
    initialise_effectors(cs);

    cs->fixated_object = NULL;
    for (role = 0; role < MAX_ROLES; role++) {
        cs->current_object[role] = NULL;
    }

    fprintf(stderr, "Done\n"); fflush(stderr);
    return(TRUE);
}

/*----------------------------------------------------------------------------*/

void cs_reinitialise(ContentionScheduler *cs, char *logfile)
{
    ObjectRole role;

    cs->actions = action_list_free(cs->actions);
    cs->cycle = 0;

    if (cs->logfile != NULL) {
        free(cs->logfile);
    }

    cs->logfile = (logfile == NULL ? NULL : string_copy(logfile));

    if (cs->logfile != NULL) {
        fprintf(stderr, "%5d\tReinitialising... [Task: %s; Log: %s] ", cs->cycle, task_names[cs->task.id], cs->logfile); fflush(stderr);
        fclose(fopen(cs->logfile, "w"));
    }
    else {
        fprintf(stderr, "%5d\tReinitialising... [Task: %s] ", cs->cycle, task_names[cs->task.id]); fflush(stderr);
    }

    random_seed_load("random_seed", cs);

    reinitialise_schemas(cs);
    reinitialise_objects(cs);
    reinitialise_effectors(cs);

    cs->fixated_object = NULL;
    for (role = 0; role < MAX_ROLES; role++) {
        cs->current_object[role] = NULL;
    }

    fprintf(stderr, "Done\n"); fflush(stderr);
}

/*----------------------------------------------------------------------------*/

void cs_destroy(ContentionScheduler *cs)
{
    extern void destroy_schemas(ContentionScheduler *cs);
    extern void destroy_objects(ContentionScheduler *cs);
    extern void destroy_effectors(ContentionScheduler *cs);

    destroy_schemas(cs);
    destroy_objects(cs);
    destroy_effectors(cs);

    if (cs->logfile != NULL) {
        free(cs->logfile);
    }

    free(cs);
}

/******************************************************************************/

static void step_activations(ContentionScheduler *cs)
{
    SchemaList *sl;
    ObjectList *ol;
    EffectorList *el;

    for (sl = cs->schemas; sl != NULL; sl = sl->tail) {
        schema_calculate_activation(cs, sl->head);
    }
    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        object_calculate_activation(cs, ol->head);
    }
    for (el = cs->effectors; el != NULL; el = el->tail) {
        effector_calculate_activation(cs, el->head);
    }

    if (parameter.maths == NORMALISED) {
        normalise_schema_activations(cs);
        normalise_object_activations(cs);
        normalise_effector_activations(cs);
    }
}

/*----------------------------------------------------------------------------*/
/* Run one cycle, logging progress if necessary:                              */

static void run_one_cycle(ContentionScheduler *cs, void *task)
{
    cs->cycle++;

    wm_step_activations(cs);
    step_activations(cs);
    step_selections(cs, task);
    dump_activations(cs);

//{
//Object *o = rng_select_best_target(cs);
//fprintf(stdout, "%4d: Best target is %s\n", cs->cycle, o->name);
//list_object_activation(cs, o, TARGET);
//print_excitation_of_object_by_schemas(stdout, cs, o);
//}

}

/*----------------------------------------------------------------------------*/

void cs_run_task(ContentionScheduler *cs, void *task, int time_out)
{
    /* Run until any intended schemata are complete:        */
    while (!all_intended_schemata_complete(cs) && (time_out-- > 0)) {
        run_one_cycle(cs, task);
    }
    /* Then set any intended schemata to "unintended":      */
    cs_schema_set_intended_all(cs, FALSE);
    /* And run until all schemata are deselected:           */
    while (cs_schema_is_selected(cs) && (time_out-- > 0)) {
        run_one_cycle(cs, task);
    }
}

/******************************************************************************/
