/*******************************************************************************

    File:       lib_gtkx.c
    Contents:   Miscellaneous GTK routines
    Author:     Rick Cooper
    Copyright (c) 1999 Richard P. Cooper

    This file is part of COGENT.

    COGENT is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    COGENT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with COGENT; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include "lib_gtkx.h"

#ifdef MALLOC_CHECK
#include "zmalloc.h"
#endif

/* Useful externally defined system functions: -------------------------------*/

extern int read(int fd, char *buffer, int length);
extern int write(int fd, const char *buffer, int length);
extern int lseek(int fd, int pos, int ref);
extern int close(int fd);

/* Global GDK variables: -----------------------------------------------------*/

GdkColor colour[29];

/******************************************************************************/
/* SECTION 1: Initialise gloabl GDK variables: ********************************/

/* colour[29] is an array of 29 different colours for inserting in graphics   */
/* contexts. colour[0] is black. colour[26] is white. The other colour        */
/* elements are currently used for colouring the different elements of        */
/* analogue buffers, but could be used to index box classes (by adding        */
/* another field to ClassStructure which indexes the class' colour).          */

/* The colours are (approx.):
    0 = black
    1 = navy blue
    2 = royal blue
    3 = dark kahki green
    4 = mud green
    5 = sky blue
    6 = bright green
    7 = pale green
    8 = pale blue
    9 = brown
   10 = mauve
   11 = vibrant lilac blue
   12 = light kahki green
   13 = grey
   14 = blue tending toward lilac
   15 = vibrant light green
   16 = limish green
   17 = sky blue
   18 = red
   19 = pink
   20 = light purple
   21 = orange
   22 = apricot
   23 = very light purple
   24 = vibrant yellow
   25 = pale yellow
   26 = white
*/

void initialise_colours(GtkWidget *window)
{
    GdkColormap *colour_map = gdk_window_get_colormap(window->window);
    int i;

    /*** First, allocate 27 colours: */

    for (i = 0; i < 27; i++) {
	short r = i / 9;
	short g = (i - (9 * r)) / 3;
	short b = i - (9 * r) - (3 * g);

	colour[i].pixel = 0;
	colour[i].red   = (unsigned short) (65535.0 * (r / 2.0));
	colour[i].green = (unsigned short) (65535.0 * (g / 2.0));
	colour[i].blue  = (unsigned short) (65535.0 * (b / 2.0));

	gdk_color_alloc(colour_map, &colour[i]);
    }

    /* Special colours (beyond the standard 27): */

    /* Very light grey, for the grid */

    colour[27].pixel = 0;
    colour[27].red = (unsigned short) (65535.0 * 0.875);
    colour[27].green = (unsigned short) (65535.0 * 0.875);
    colour[27].blue = (unsigned short) (65535.0 * 0.875);
    gdk_color_alloc(colour_map, &colour[27]);

    /* Blue, for selections */

    colour[28].pixel = 0;
    colour[28].red = (unsigned short) 16448;
    colour[28].green = (unsigned short) 26985;
    colour[28].blue = (unsigned short) 51400;
    gdk_color_alloc(colour_map, &colour[28]);
}

/*----------------------------------------------------------------------------*/

GdkGC *create_gcs(GtkWidget *window)
{
    GdkGC *theGC = gdk_gc_new(window->window);
    gdk_gc_set_foreground(theGC, &colour[0]);
    gdk_gc_set_background(theGC, &colour[26]);
    gdk_gc_set_function(theGC, GDK_COPY);
    return(theGC);
}

/******************************************************************************/
/* SECTION 3: Simple gtkx functions that extend standard capabilities: ********/

GtkWidget *gtkx_button_create(GtkWidget *parent, const char *label, void callback(), void *arg)
{
    GtkWidget *button = gtk_button_new_with_label(label);

    if (callback != NULL) {
        g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(callback), arg);
    }
    gtk_container_add(GTK_CONTAINER(parent), button);
    gtk_widget_show(button);
    return(button);
}

GtkWidget *gtkx_stock_button_create(GtkWidget *parent, const char *label, int left, void callback(), void *arg)
{
    GtkWidget *button = gtk_button_new_with_label(label);

    if (callback != NULL) {
        g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(callback), arg);
    }
    gtk_button_set_use_stock(GTK_BUTTON(button), TRUE);
    if (left) {
        gtk_box_pack_start(GTK_BOX(parent), button, FALSE, FALSE, 0);
    }
    else {
        gtk_box_pack_end(GTK_BOX(parent), button, FALSE, FALSE, 0);
    }
    gtk_widget_set_size_request(button, 85, -1);
    gtk_widget_show(button);
    return(button);
}

/*----------------------------------------------------------------------------*/

void gtkx_widget_set_font_bold(GtkWidget *widget)
{
    GtkStyle *style = gtk_style_copy(widget->style);

    pango_font_description_set_weight(style->font_desc, PANGO_WEIGHT_BOLD);
    /* Unreference the original style since we don't want it anymore: */
    g_object_unref(widget->style);
    gtk_widget_set_style(widget, style);
    /* Unreference the new style so it dies with the widget: */
    g_object_unref(widget->style);
}

/*----------------------------------------------------------------------------*/

void gtkx_widget_set_foreground_colour(GtkWidget *widget, int index)
{
    gtk_widget_modify_fg(widget, GTK_STATE_NORMAL, &colour[index]);
    gtk_widget_modify_fg(widget, GTK_STATE_INSENSITIVE, &colour[index]);
    gtk_widget_modify_fg(widget, GTK_STATE_PRELIGHT, &colour[index]);

    if (GTK_IS_CONTAINER(widget)) {
        GList *children = gtk_container_get_children(GTK_CONTAINER(widget));
        while (children != NULL) {
            GList *tmp = g_list_next(children);
            gtkx_widget_set_foreground_colour((GtkWidget *)children->data, index);
            children = tmp;
        }
    }
}

/*----------------------------------------------------------------------------*/

void gtkx_widget_unset_foreground(GtkWidget *widget)
{
    gtk_widget_modify_fg(widget, GTK_STATE_NORMAL, NULL);
    gtk_widget_modify_fg(widget, GTK_STATE_INSENSITIVE, NULL);
    gtk_widget_modify_fg(widget, GTK_STATE_PRELIGHT, NULL);

    if (GTK_IS_CONTAINER(widget)) {
        GList *children = gtk_container_get_children(GTK_CONTAINER(widget));
        while (children != NULL) {
            GList *tmp = g_list_next(children);
            gtkx_widget_unset_foreground((GtkWidget *)children->data);
            children = tmp;
        }
    }
}

/*----------------------------------------------------------------------------*/

int gtkx_widget_is_mapped(GtkWidget *widget)
{
    return(GTK_WIDGET_MAPPED(widget));
}

/******************************************************************************/

void gtkx_container_destroy_children(GtkContainer *panel)
{
    GList *children = gtk_container_get_children(panel);

    while (children != NULL) {
        GList *tmp = g_list_next(children);
        gtk_container_remove(panel, (GtkWidget *)children->data);
        children = tmp;
    }
    gtk_widget_queue_resize(GTK_WIDGET(panel));
}

/*----------------------------------------------------------------------------*/

GtkWidget *gtkx_menu_separator_create(GtkWidget *menu)
{
    GtkWidget *tmp = gtk_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), tmp);
    gtk_widget_show(tmp);
    return(tmp);
}

/*----------------------------------------------------------------------------*/

GtkWidget *gtkx_menu_item_create(GtkWidget *menu, char *label, void callback(), void *arg)
{
    GtkWidget *tmp = gtk_menu_item_new_with_label(label);
    if (callback != NULL) {
        g_signal_connect(G_OBJECT(tmp), "activate", G_CALLBACK(callback), arg);
    }
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), tmp);
    gtk_widget_show(tmp);
    return(tmp);
}


/*----------------------------------------------------------------------------*/

GtkWidget *gtkx_check_menu_item_create(GtkWidget *menu, char *label, int active, void callback(), void *arg)
{
    GtkWidget *tmp = gtk_check_menu_item_new_with_label(label);

    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tmp), active);
    if (callback != NULL) {
        g_signal_connect(G_OBJECT(tmp), "activate", G_CALLBACK(callback), arg);
    }
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), tmp);
    gtk_widget_show(tmp);
    return(tmp);
}

/*----------------------------------------------------------------------------*/

GtkWidget *gtkx_radio_menu_item_create(GtkWidget *menu, char *label, GSList *group, void callback(), void *arg)
{
    GtkWidget *tmp = gtk_radio_menu_item_new_with_label(group, label);

    if (callback != NULL) {
        g_signal_connect(G_OBJECT(tmp), "activate", G_CALLBACK(callback), arg);
    }
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), tmp);
    gtk_widget_show(tmp);
    return(tmp);
}

/*----------------------------------------------------------------------------*/

void gtkx_notebook_append_page(GtkWidget *notebook, GtkWidget *page, char *text)
{
    GtkWidget *label;

    label = gtk_label_new(text);
    gtk_widget_show(label);
    gtk_widget_show(page);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page, label);
}

/******************************************************************************/

void io_message(GtkWidget *window, const char *message)
{
    GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                GTK_MESSAGE_WARNING,
                                GTK_BUTTONS_OK,
                                message);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

int io_query(GtkWidget *window, const char *question)
{
    int answer;
    GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                                GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                GTK_MESSAGE_QUESTION,
                                GTK_BUTTONS_YES_NO,
                                question);

    answer = (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_YES);
    gtk_widget_destroy(dialog);
    return(answer);
}



void gtkx_flush_events()
{
    while (gtk_events_pending()) {
        gtk_main_iteration();
    }
}


/******************************************************************************/

void gtkx_canvas_reallocate(GtkWidget *da)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(da), "pixmap");

    if (canvas != NULL) {
        g_object_unref(canvas);
    }
    canvas = gdk_pixmap_new(da->window, da->allocation.width, da->allocation.height, -1);
    g_object_set_data(G_OBJECT(da), "pixmap", canvas);
}

/*----------------------------------------------------------------------------*/

int gtkx_canvas_event_expose(GtkWidget *da, GdkEventExpose *event, void *dummy)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(da), "pixmap");

    if (canvas != NULL) {
        gdk_draw_drawable(da->window,
                da->style->fg_gc[GTK_WIDGET_STATE(da)], canvas,
                event->area.x, event->area.y,
                event->area.x, event->area.y,
                event->area.width, event->area.height);
    }
    return(0);
}

/*----------------------------------------------------------------------------*/

void gtkx_canvas_callback_destroy(GtkWidget *da, void *dummy)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(da), "pixmap");

    if (canvas != NULL) {
        g_object_unref(canvas);
    }
    g_object_set_data(G_OBJECT(da), "pixmap", NULL);
}

/******************************************************************************/
