/*******************************************************************************

    File:     rng.c
    Author:   Rick Cooper
    Created:  24/06/08
    Edited:   24/06/08
    Contents: Routines for administering and scoring RNG

    Public procedures:
        RngTaskData *rng_create_task()
        void rng_destroy_task(RngTaskData *task)
        void rng_initialise_task(RngTaskData *task)
        void rng_initialise_next_trial(ContentionScheduler *cs, RngTaskData *task)
        Boolean rng_trial_complete(RngTaskData *task)
        Boolean rng_task_complete(RngTaskData *task)
        void rng_step_event(ContentionScheduler *cs, RngTaskData *task)
        void rng_synchronise_world(ContentionScheduler *cs, RngTaskData *task)

*******************************************************************************/

#include "ss.h"
#include "rng.h"
#include <string.h>
#include <math.h>

/******************************************************************************/
/* Section 1: Data structure creation/disposal -------------------------------*/

RngTaskData *rng_create_task()
{
    return((RngTaskData *)malloc(sizeof(RngTaskData)));
}

void rng_destroy_task(RngTaskData *task)
{
    if (task != NULL) {
        free(task);
    }
}

/******************************************************************************/
/* Section 2: Task initialisation --------------------------------------------*/

static void initialise_trial(RngTaskData *task)
{
    if (task->target_selected != NULL) {
        task->target_selected->colour = 6;
        task->target_selected = NULL;
    }
    task->event_count = 0;
    task->trial_complete = FALSE;
}

static void rng_initialise_block(RngTaskData *task)
{
    int i;

    for (i = 0; i < RNG_TRIALS_MAX; i++) {
        task->sequence[task->block_count][i] = -1;
    }
    task->trial_count = 0;
    initialise_trial(task);
}

/******************************************************************************/

static void rng_add_disk(RngTaskData *task, double x, double y, int colour, int id)
{
    RngDisks *new = (RngDisks *)malloc(sizeof(RngDisks));

    if (new != NULL) {
        new->x = x;
        new->y = y;
        new->colour = colour;
        new->id = id;
        new->next = task->disks;
        task->disks = new;
    }
}

void rng_initialise_task(RngTaskData *task)
{
    int i;

    task->block_count = 0;
    task->target_selected = NULL;
    rng_initialise_block(task);

    /* Initialise world: */
    task->disks = NULL;
    rng_add_disk(task, 0.0, 0.0, 26, -1);
    for (i = 0; i < 10; i++) {
        rng_add_disk(task, -150.0 * cos(M_PI * i/5.0), -150.0 * sin(M_PI * i/5.0), 6, i);
    }

    /* Initialise scoring: */
    task->score_r[task->block_count] = -1.0;
    task->score_rng[task->block_count] = -1.0;
    task->score_rr[task->block_count] = -1.0;
    task->score_oa[task->block_count] = -1.0;
    task->score_aa[task->block_count] = -1.0;
    task->score_tpi[task->block_count] = -1.0;
    task->score_rg[task->block_count] = -1.0;
}

/******************************************************************************/

static void rng_score_sequence(RngTaskData *task);

void rng_initialise_next_trial(ContentionScheduler *cs, RngTaskData *task)
{
    task->trial_count++;
    if (task->trial_count >= RNG_TRIALS_MAX) {
        rng_score_sequence(task);
        task->block_count++;
        if (task->block_count < RNG_BLOCKS) {
            cs_initialise_task(cs, TASK_RNG);
            wm_destroy(cs);
            rng_initialise_block(task);
            rng_initialise_wm(cs);
        }
    }
    else { 
        initialise_trial(task);
    }
}

/******************************************************************************/
/* Section 3: Score a block of trials ----------------------------------------*/

static double calculate_mean(int list[], int i)
{
    int sum = 0, n;

    for (n = 0; n < i; n++) {
        sum += list[n];
    }
    return(sum / (double) i);
}

static void rng_score_data(int sequence[RNG_TRIALS_MAX], double *rt_mean, int *misses, int *hits, double *r, double *rng, double *rr, double *aa, double *oa,  double *tpi, double *rg)
{
    /* Read a data file and return the various RNG dvs. */
    /* Also return misses (times participant failed to generate a response) */
    /* and hits (responses generated, excluding misses). */

    int f[10], ff[10][10], last[10];
    int rg_list[100];
    double rt, rt_sum, sum, h_s, h_m;
    double num, denom;
    long v1, v2, fr;
    int tp_direction, tp_count, n, d, i, j;

    *misses = 0;
    *hits = 0;

    for (v1 = 0; v1 < 10; v1++) {
        f[v1] = 0;
        last[v1] = -1;
        for (v2 = 0; v2 < 10; v2++) {
            ff[v1][v2] = 0;
        }
    }

    /* Score the data: */

    v1 = -1; fr = -1; tp_direction = 0; n = 0; i = 0; tp_count = 0; rt_sum = 0.0;

    for (j = 0; j < RNG_TRIALS_MAX; j++) {
        v2 = sequence[j];
        rt = 0.0; // TO DO: Handle RT?

        rt_sum += rt;
        /* Record the first valid response, for wrap-around: */
        if ((v2 >= 0) && (fr == -1)) {
            fr = v2;
        }

        /* Score hits and misses, and count frequencies for hits: */
        if (v2 < 0) {
            *misses = *misses + 1;
        }
        else {
            *hits = *hits + 1;
            f[v2]++;
        }

        /* Count frequencies for pairs: */
        if ((v1 >= 0) && (v2 >= 0)) {
            ff[v1][v2]++;
        }

        /* Score turning points: */
        d = (v2 < v1) ? (v2 + 10 - v1) : (v2 - v1);
        if ((v1 >= 0) && (d != 0)) {
            if ((tp_direction == 0) && (d != 5)) {
                /* First response ... just set the initial direction: */
                tp_direction = (d > 5) ? -1 : 1;
            }
            else if ((tp_direction > 0) && (d > 5)) {
                /* Sequence was going clockwise, but is no longer: */
                tp_direction = -1;
                tp_count++;
            }
            else if ((tp_direction < 0) && (d < 5)) {
                /* Sequence was going counter-clockwise, but is no longer: */
                tp_direction = +1;
                tp_count++;
            }
        }

        if (v2 >= 0) {
            /* If v2 is -1 we ignore this in ff; */
            v1 = v2;
        }

        if ((v2 >= 0) && (i < 100)) {
            if (last[v2] == -1) {
                last[v2] = n;
            }
            else {
                rg_list[i++] = n - last[v2];
                last[v2] = n;
            }
        }

        n++;
    }
    *rt_mean = rt_sum / (double) n;

    /* Add in the wrap-around: */
    if ((fr >= 0) && (v2 >= 0)) {
        ff[v2][fr]++;
    }

    /* Calculation of R: */
    sum = 0.0;
    for (v1 = 0; v1 < 10; v1++) {
        if (f[v1] > 0) {
            sum = sum + f[v1] * log(f[v1]) / log(2);
        }
    }
    h_s = log(*hits) / log(2.0) - (1.0 / (double) *hits) * sum;
    h_m = log(10.0) / log(2.0);

    *r = 100.0 * (1 - h_s / h_m);

    /* Calculation of RNG: */
    num = 0.0;
    denom = 0.0;

    for (v1 = 0; v1 < 10; v1++) {
	for (v2 = 0; v2 < 10; v2++) {
	    if (ff[v1][v2] > 1) {
		num = num + ff[v1][v2] * log(ff[v1][v2]) / log(2.0);
	    }
	}
	if (f[v1] > 1) {
	    denom = denom + f[v1] * log(f[v1]) / log(2.0);
	}
    }
    *rng = num / denom;

    /* Calculation of RR (repeat responses): */

    num = 0.0;
    denom = 0.0;
    for (v1 = 0; v1 < 10; v1++) {
	num = num + ff[v1][v1];
    }
    for (v1 = 0; v1 < 10; v1++) {
	denom = denom + f[v1];
    }
    *rr = num / denom;

    /* Calculation of AA (adjacent associates): */

    num = 0.0;
    denom = 0.0;
    for (v1 = 0; v1 < 10; v1++) {
	num = num + ff[v1][(v1+1) % 10] + ff[v1][(v1+9) % 10];
    }
    for (v1 = 0; v1 < 10; v1++) {
	denom = denom + f[v1];
    }
    *aa = num / denom;

    /* Calculation of OA (opposite associates): */

    num = 0.0;
    denom = 0.0;
    for (v1 = 0; v1 < 10; v1++) {
	num = num + ff[v1][(v1+5) % 10];
    }
    for (v1 = 0; v1 < 10; v1++) {
	denom = denom + f[v1];
    }
    *oa = num / denom;

    /* Calculation of tpi: */

    *tpi = 100.0 * (tp_count / (2.0 * (n - 2.0) / (3.0)));

    /* Calculate RG (repetition gap): */

    *rg = calculate_mean(rg_list, i);
}


static void rng_score_sequence(RngTaskData *task)
{
    double rt_mean, r, rng, rr, aa, oa, tpi, rg;
    int misses, hits;

    rng_score_data(task->sequence[task->block_count], &rt_mean, &misses, &hits, &r, &rng, &rr, &aa, &oa, &tpi, &rg);

    task->score_r[task->block_count] = r;
    task->score_rng[task->block_count] = rng;
    task->score_rr[task->block_count] = rr;
    task->score_aa[task->block_count] = aa;
    task->score_oa[task->block_count] = oa;
    task->score_tpi[task->block_count] = tpi;
    task->score_rg[task->block_count] =rg;
}

/******************************************************************************/
/* Section 4: RNG events -----------------------------------------------------*/

RngDisks *rng_get_disk(RngTaskData *task, char *l)
{
    RngDisks *d = task->disks;
    int id = l[0] - 'A';

    while ((d != NULL) && (d->id != id)) {
        d = d->next;
    }
    return(d);
}

extern double cs_random();

Boolean rng_response_fits_pattern(ContentionScheduler* cs, RngTaskData *task)
{
//    return(TRUE); // Reject the response ... should generate a different response
// Randomly reject the first response...
//    return(cs_random() > 0.1);
    return(FALSE);
}

char *rng_get_disk_label(Object *disk)
{
    if (disk == NULL) {
        return(NULL);
    }
    else if (object_has_feature(disk, IS_LETTER_A)) {
        return("A");
    }
    else if (object_has_feature(disk, IS_LETTER_B)) {
        return("B");
    }
    else if (object_has_feature(disk, IS_LETTER_C)) {
        return("C");
    }
    else if (object_has_feature(disk, IS_LETTER_D)) {
        return("D");
    }
    else if (object_has_feature(disk, IS_LETTER_E)) {
        return("E");
    }
    else if (object_has_feature(disk, IS_LETTER_F)) {
        return("F");
    }
    else if (object_has_feature(disk, IS_LETTER_G)) {
        return("G");
    }
    else if (object_has_feature(disk, IS_LETTER_H)) {
        return("H");
    }
    else if (object_has_feature(disk, IS_LETTER_I)) {
        return("I");
    }
    else if (object_has_feature(disk, IS_LETTER_J)) {
        return("J");
    }
    else {
        return(NULL);
    }
}

Object *rng_get_matching_disk(ContentionScheduler *cs, char *letter)
{
    ObjectList *ol;
    Object *this;
    char *this_letter;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        this = ol->head;
        this_letter = rng_get_disk_label(this);

        if ((this_letter != NULL) && (strcmp(letter, this_letter) == 0)) {
            return(this);
        }
    }
    return(NULL);
}

Boolean rng_disk_attended(ContentionScheduler *cs)
{
    Chunk *template;
    Boolean result;

    template = chunk_build("status", "contemplating");
    chunk_extend(template, "response", NULL);

    result = wm_match(cs, template);
    chunk_dispose(template);
    return(result);
}

Boolean rng_disk_selected(ContentionScheduler *cs)
{
    ObjectList *ol;
    Object *this;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        this = ol->head;
        if (object_has_feature(this, IS_DISK) && object_has_feature(this, IS_COLOUR_RED)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

Object *rng_select_best_target(ContentionScheduler *cs)
{
// Run through the list of (target) objects looking for a disk that is
// a) unselected, and b) above some threshold

    ObjectList *ol;
    Object *this, *disk;
    double disk_max = 0.0;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        this = ol->head;
        if (object_has_feature(this, IS_DISK) && !object_has_feature(this, IS_COLOUR_WHITE)) {
            if (this->activation[TARGET][cs->cycle-1] > disk_max) {
                disk = this;
                disk_max = disk->activation[TARGET][cs->cycle-1];
            }
        }
    }
    if (disk_max > 0.0) {
        return(disk);
    }
    else {
        return(NULL);
    }
}

static Boolean rng_unsuitable_response(ContentionScheduler *cs, RngTaskData *task)
{
    return(rng_response_fits_pattern(cs, task));
}

void rng_step_event(ContentionScheduler *cs, RngTaskData *task)
{
    task->event_count++;

    if (!rng_disk_attended(cs)) {
        cs_schema_set_intended(cs, "schemaThinkResponse", TRUE);
        cs_run_task(cs, task, 80);
    }
    else if (rng_unsuitable_response(cs, task)) {
        cs_schema_set_intended(cs, "schemaInhibitResponse", TRUE);
        cs_run_task(cs, task, 80);
    }
    else {
        cs_schema_set_intended(cs, "schemaProduceResponse", TRUE);
        cs_run_task(cs, task, 120);
    }
}

Boolean rng_trial_complete(RngTaskData *task)
{
    return(task->target_selected != NULL);
}

Boolean rng_task_complete(RngTaskData *task)
{
    return(task->block_count >= RNG_BLOCKS);
}

/******************************************************************************/
/* Section 5: Synchronise CS and external world representations --------------*/

void rng_select_disk(ContentionScheduler *cs, RngTaskData *task, char *l)
{
    Object *o = rng_get_matching_disk(cs, l);
    RngDisks *d = rng_get_disk(task, l);

    if (o != NULL) {
        o->features = (o->features & ~IS_COLOUR_GREEN) | IS_COLOUR_RED;
    }

    if (d != NULL) {
        d->colour = 18;
    }
    task->target_selected = d;
    task->sequence[task->block_count][task->trial_count] = d->id;

//    fprintf(stdout, "task->sequence[%d][%d] = %d\n", task->block_count, task->trial_count, d->id);
}


static Object *set_disk_features(ContentionScheduler *cs, char *name, Boolean selected)
{
    Object *this = objectGetByName(cs, name);

    if (this != NULL) {
        if (selected) {
            this->features = (this->features | IS_COLOUR_RED) & ~IS_COLOUR_GREEN;
        }
        else {
            this->features = (this->features | IS_COLOUR_GREEN) & ~IS_COLOUR_RED;
        }
    }
    return(this);
}

void rng_synchronise_world(ContentionScheduler *cs, RngTaskData *task)
{
    char buffer[16], label[2];
    int i;

    for (i = 0; i < 10; i++) {
        g_snprintf(label, 2, "%c", 'A' + i);
        g_snprintf(buffer, 16, "target: %s", label);

        set_disk_features(cs, buffer, (task->target_selected == rng_get_disk(task, label)));
    }
}

/******************************************************************************/
/* Section 6: Dump the analysis results: -------------------------------------*/

void rng_dump_analyses(FILE *fp, RngTaskData *task)
{
    rng_score_sequence(task);

    fprintf(fp, "\n");
    fprintf(fp, "Items generated:      %4d\n", task->trial_count);
    fprintf(fp, "Redundancy score:     %5.3f\n", task->score_r[task->block_count]);
    fprintf(fp, "RNG score:            %5.3f\n", task->score_rng[task->block_count]);
    fprintf(fp, "Repeat responses:     %5.3f\n", task->score_rr[task->block_count]);
    fprintf(fp, "Opposite associates:  %5.3f\n", task->score_oa[task->block_count]);
    fprintf(fp, "Adjacent associates:  %5.3f\n", task->score_aa[task->block_count]);
    fprintf(fp, "Turning point index:  %5.3f\n", task->score_tpi[task->block_count]);
    fprintf(fp, "Repetition gap:       %5.3f\n", task->score_rg[task->block_count]);
}

/******************************************************************************/
