/*******************************************************************************

    File:     rng_object.c
    Author:   Rick Cooper
    Created:  27/06/08
    Edited:   27/06/08
    Contents: Create/maintain RNG object representations

    Public procedures:
        void install_rng_basic_objects(ContentionScheduler *cs)
        void write_rng_object_attributes(FILE *fp, Object *this)

*******************************************************************************/

#include "ss.h"
#include "rng.h"
#include "math.h"

/******************************************************************************/
/* Section 1: Install RNG object representations -----------------------------*/

static int get_label(int i)
{
    switch (i) {
        case 0: { return(IS_LETTER_A); break; }
        case 1: { return(IS_LETTER_B); break; }
        case 2: { return(IS_LETTER_C); break; }
        case 3: { return(IS_LETTER_D); break; }
        case 4: { return(IS_LETTER_E); break; }
        case 5: { return(IS_LETTER_F); break; }
        case 6: { return(IS_LETTER_G); break; }
        case 7: { return(IS_LETTER_H); break; }
        case 8: { return(IS_LETTER_I); break; }
        case 9: { return(IS_LETTER_J); break; }
        default: { return(0); break; }
    }
}

static Boolean rng_is_opposite_disk(Object *d1, Object *d2)
{
    char *l1 = rng_get_disk_label(d1);
    char *l2 = rng_get_disk_label(d2);

    if ((l1 != NULL) && (l2 != NULL)) {
        return(labs(l1[0]-l2[0]) == 5);
    }
    else {
        return(FALSE);
    }
}

static Boolean rng_is_adjacent_disk(Object *d1, Object *d2)
{
    char *l1 = rng_get_disk_label(d1);
    char *l2 = rng_get_disk_label(d2);

    if ((l1 != NULL) && (l2 != NULL)) {
        return(labs(l1[0]-l2[0]) == 1);
    }
    else {
        return(FALSE);
    }
}

double rng_object_gradient(ContentionScheduler *cs, Object *this, ObjectRole r)
{
    ObjectList *ol;
    double g = 0.0;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        Object *tmp  = ol->head;

        if (rng_is_opposite_disk(this, tmp)) {
//            g += tmp->activation[r][cs->cycle-1];
        }
        else if (rng_is_adjacent_disk(this, tmp)) {
            g += tmp->activation[r][cs->cycle-1];
        }
    }
    return(g);
}


void install_rng_basic_objects(ContentionScheduler *cs)
{
    Object *this;
    char buffer[16];
    int i;

    /* Central fixation: */
    this = objectInstall(cs, "centre_disk", IS_DISK | IS_COLOUR_WHITE, 0);
    objectSetPosition(this, 0.0, 0.0);

    /* Ten labelled locations: */
    for (i = 0; i < 10; i++) {
        g_snprintf(buffer, 16, "target: %c", 'A' + i);
        this = objectInstall(cs, buffer, IS_DISK | IS_COLOUR_GREEN | get_label(i), 0);
        objectSetPosition(this, 100.0 * cos(M_PI * i/5.0), 50.0 * sin(M_PI * i/5.0));
        objectSetObjectTrigger(this, rng_object_gradient);
    }
}

/******************************************************************************/
/* Section 2: Write the object attributes (for debugging) --------------------*/

void write_rng_object_attributes(FILE *fp, Object *this)
{
    if (object_has_feature(this, IS_DISK)) {
        fprintf(fp, "disk: ");
    }

    if (object_has_feature(this, IS_COLOUR_WHITE)) {
        fprintf(fp, "white ");
    }
    else if (object_has_feature(this, IS_COLOUR_GREEN)) {
        fprintf(fp, "green ");
    }
    else if (object_has_feature(this, IS_COLOUR_RED)) {
        fprintf(fp, "red   ");
    }

    if (object_has_feature(this, IS_LETTER_A)) {
        fprintf(fp, "A ");
    }
    else if (object_has_feature(this, IS_LETTER_B)) {
        fprintf(fp, "B ");
    }
    else if (object_has_feature(this, IS_LETTER_C)) {
        fprintf(fp, "C ");
    }
    else if (object_has_feature(this, IS_LETTER_D)) {
        fprintf(fp, "D ");
    }
    else if (object_has_feature(this, IS_LETTER_E)) {
        fprintf(fp, "E ");
    }
    else if (object_has_feature(this, IS_LETTER_F)) {
        fprintf(fp, "F ");
    }
    else if (object_has_feature(this, IS_LETTER_G)) {
        fprintf(fp, "G ");
    }
    else if (object_has_feature(this, IS_LETTER_H)) {
        fprintf(fp, "H ");
    }
    else if (object_has_feature(this, IS_LETTER_I)) {
        fprintf(fp, "I ");
    }
    else if (object_has_feature(this, IS_LETTER_J)) {
        fprintf(fp, "J ");
    }
}

/******************************************************************************/

Boolean rng_disk_is_selected(ContentionScheduler *cs)
{
    ObjectList *ol;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        Object *this  = ol->head;

        if (object_has_feature(this, IS_DISK) && object_has_feature(this, IS_COLOUR_RED)) {
fprintf(stdout, "Found red disk\n");
            return(TRUE);
        }
    }
    return(FALSE);
}


