/*******************************************************************************

    File:     rng_schema.c
    Author:   Rick Cooper
    Created:  27/06/08
    Edited:   27/06/08
    Contents: Create/maintain schemas for the RNG task

    Public procedures:
        void install_rng_schemas(ContentionScheduler *cs)
        void rng_initialise_wm(ContentionScheduler *cs)

*******************************************************************************/

#include "ss.h"
#include "rng.h"

Object *rng_object_by_feature(ContentionScheduler *cs, int f)
{
    ObjectList *ol = cs->objects;

    while (ol != NULL) {
        if (object_has_feature(ol->head, f)) {
            return(ol->head);
        }
        ol = ol->tail;
    }
    return(NULL);
}

Object *rng_opposite_disk(ContentionScheduler *cs, Object *this)
{
    if (object_has_feature(this, IS_LETTER_A)) {
        return(rng_object_by_feature(cs, IS_LETTER_F));
    }
    else if (object_has_feature(this, IS_LETTER_B)) {
        return(rng_object_by_feature(cs, IS_LETTER_G));
    }
    else if (object_has_feature(this, IS_LETTER_C)) {
        return(rng_object_by_feature(cs, IS_LETTER_H));
    }
    else if (object_has_feature(this, IS_LETTER_D)) {
        return(rng_object_by_feature(cs, IS_LETTER_I));
    }
    else if (object_has_feature(this, IS_LETTER_E)) {
        return(rng_object_by_feature(cs, IS_LETTER_J));
    }
    else if (object_has_feature(this, IS_LETTER_F)) {
        return(rng_object_by_feature(cs, IS_LETTER_A));
    }
    else if (object_has_feature(this, IS_LETTER_G)) {
        return(rng_object_by_feature(cs, IS_LETTER_B));
    }
    else if (object_has_feature(this, IS_LETTER_H)) {
        return(rng_object_by_feature(cs, IS_LETTER_C));
    }
    else if (object_has_feature(this, IS_LETTER_I)) {
        return(rng_object_by_feature(cs, IS_LETTER_D));
    }
    else if (object_has_feature(this, IS_LETTER_J)) {
        return(rng_object_by_feature(cs, IS_LETTER_E));
    }
    else {
        return(NULL);
    }
}

/******************************************************************************/
/* Section 1: RNG-specific object test/manipulation functions ----------------*/

static GoalType goalDiskAttended(ContentionScheduler *cs)
{
    return(rng_disk_attended(cs) ? GOAL_TRUE : GOAL_FALSE);
}

static GoalType goalDiskSelected(ContentionScheduler *cs)
{
    return(rng_disk_selected(cs) ? GOAL_TRUE : GOAL_FALSE);
}

static GoalType goalAttentionFree(ContentionScheduler *cs)
{
    return(rng_disk_attended(cs) ? GOAL_FALSE : GOAL_TRUE);
}

static GoalType goalSelectionInhibited(ContentionScheduler *cs)
{
    Chunk *template, *match;

    template = chunk_build("response", NULL);
    match = wm_query(cs, template);
    chunk_dispose(template);

    if (match == NULL) {
        return(GOAL_TRUE);
    }
    else {
        char *l1 = chunk_get_value(match, "response");
        char *l2 = rng_get_disk_label(rng_select_best_target(cs));

        if (strcmp(l1, l2) == 0) {
            return(GOAL_FALSE);
        }
        else {
            return(GOAL_TRUE);
        }
    }
}

/******************************************************************************/
/* Section 2: Schema triggering functions ------------------------------------*/

static double schemaAttendOppositeTriggersTarget(ContentionScheduler *cs, Object *o)
{
    Object *d = rng_opposite_disk(cs, o);

//    if (d != NULL) {
//        return(5.0 * d->activation[TARGET][cs->cycle-1]);
//    }
//    else {
        return(0.0);
//    }
}

static double schemaThinkAnyResponseTriggersTarget(ContentionScheduler *cs, Object *o)
{
    if (object_has_feature(o, IS_DISK) && !object_has_feature(o, IS_COLOUR_WHITE)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

static double schemaThinkNotResponseTriggersTarget(ContentionScheduler *cs, Object *o)
{
//    if (o == rng_select_best_target(cs)) {
//        return(-5.0);
//    }
//    else {
        return(0.0);
//    }
}

static double schemaProduceResponseTriggersEffector(ContentionScheduler *cs, Effector *e)
{
    return(e == bestEmptyHand(cs) ? 1.0 : 0.0);
}

/******************************************************************************/
/* Section 3: Install RNG schemas -------------------------------------------*/

void install_rng_schemas(ContentionScheduler *cs)
{
    Schema *s;

    /* Part 1: Think of a response: */

    s = install_schema(cs, "schemaThinkResponse", "goalGenerate", ACTION_NONE, MEDIUM);
    install_schema_subgoal(s, "goalAttendDisk", goalTrue, goalDiskAttended);
    install_schema_object_trigger(s, TARGET, schemaThinkAnyResponseTriggersTarget);

    s = install_schema(cs, "schemaAttendAnyDisk",  "goalAttendDisk", ACTION_ATTEND_TARGET,  LOW);
    install_schema_object_trigger(s, TARGET, schemaThinkAnyResponseTriggersTarget);

//    s = install_schema(cs, "schemaAttendOppositeDisk",  "goalAttendDisk", ACTION_ATTEND_TARGET,  LOW);
//    install_schema_object_trigger(s, TARGET, schemaAttendOppositeTriggersTarget);

    /* Part 2: Inhibit an inappropriate response: */

    s = install_schema(cs, "schemaInhibitResponse", "goalInhibit", ACTION_NONE, MEDIUM);
    install_schema_subgoal(s, "goalDeattendDisk", goalDiskAttended, goalAttentionFree);
    install_schema_object_trigger(s, TARGET, schemaThinkNotResponseTriggersTarget);

    s = install_schema(cs, "schemaDeattendDisk",  "goalDeattendDisk", ACTION_DISENGAGE_TARGET,  LOW);
    install_schema_object_trigger(s, TARGET, schemaThinkNotResponseTriggersTarget);

    /* Part 3: Produce and then inhibit the response: */

    s = install_schema(cs, "schemaProduceResponse", "goalRespond", ACTION_NONE, MEDIUM);
    install_schema_effector_trigger(s, schemaProduceResponseTriggersEffector);
    install_schema_subgoal(s, "goalSelectDisk", goalDiskAttended, goalDiskSelected);
    install_schema_subgoal(s, "goalDeattendDisk", goalDiskSelected, goalSelectionInhibited);

    s = install_schema(cs, "schemaSelectDisk",  "goalSelectDisk", ACTION_SELECT_TARGET,  LOW);
}

/******************************************************************************/

void rng_initialise_wm(ContentionScheduler *cs)
{

}

/******************************************************************************/
