/*******************************************************************************

    File:     tol.c
    Author:   Rick Cooper
    Created:  29/03/07
    Edited:   29/03/07
    Contents: Routines administering and scoring the WCST 

    Public procedures:
        TolTaskData *tol_create_task()
        void tol_destroy_task(TolTaskData *task)
        void tol_initialise_task(TolTaskData *task)
        void tol_initialise_next_trial(ContentionScheduler *cs, TolTaskData *task)
        Boolean tol_trial_complete(TolTaskData *task)
        Boolean tol_task_complete(TolTaskData *task)
        void tol_action_pick_up(TolTaskData *task, Object *theme, Effector *effector)
        void tol_action_put_down(TolTaskData *task, Object *theme, Object *target)
        void tol_step_event(ContentionScheduler *cs, TolTaskData *task)

*******************************************************************************/

#include "ss.h"
#include "tol.h"
#include <math.h>

/******************************************************************************/
/* Section 1: Data structure creation/disposal -------------------------------*/

TolTaskData *tol_create_task()
{
    return((TolTaskData *)malloc(sizeof(TolTaskData)));
}

void tol_destroy_task(TolTaskData *task)
{
    if (task != NULL) {
        free(task);
    }
}

/******************************************************************************/
/* Section 2: Task initialisation --------------------------------------------*/

static void tol_state_copy(TolConfiguration *target, TolConfiguration *source)
{
    target->red_ball.peg = source->red_ball.peg;
    target->red_ball.position = source->red_ball.position;
    target->green_ball.peg = source->green_ball.peg;
    target->green_ball.position = source->green_ball.position;
    target->blue_ball.peg = source->blue_ball.peg;
    target->blue_ball.position = source->blue_ball.position;
}

static void tol_initialise_trial(TolTaskData *task)
{
    TolProblem *p = &(task->problems[task->trial_count]);
 
fprintf(stdout, "TRIAL %d ****************************************************************\n", task->trial_count);
    task->event_count = 0;
    tol_state_copy(&(p->current_state), &(p->initial_state));
    p->moves_actual = 0;
    p->rule_breaks = 0;
    p->balls_correct = 0;
    p->trial_done = FALSE;
}

static void tol_set_problem(TolTaskData *task, int n, int m, int sr1, int sr2, int sg1, int sg2, int sb1, int sb2, int dr1, int dr2, int dg1, int dg2, int db1, int db2)
{
    task->problems[n].moves_minimum = m;

    task->problems[n].initial_state.red_ball.peg = sr1;
    task->problems[n].initial_state.red_ball.position = sr2;
    task->problems[n].initial_state.green_ball.peg = sg1;
    task->problems[n].initial_state.green_ball.position = sg2;
    task->problems[n].initial_state.blue_ball.peg = sb1;
    task->problems[n].initial_state.blue_ball.position = sb2;

    task->problems[n].desired_state.red_ball.peg = dr1;
    task->problems[n].desired_state.red_ball.position = dr2;
    task->problems[n].desired_state.green_ball.peg = dg1;
    task->problems[n].desired_state.green_ball.position = dg2;
    task->problems[n].desired_state.blue_ball.peg = db1;
    task->problems[n].desired_state.blue_ball.position = db2;
}


void tol_initialise_task(TolTaskData *task)
{
    int n = 0;

    /* Karolina's practice tasks: */
    tol_set_problem(task, n++, 3, 3, 2, 3, 3, 3, 1, 2, 2, 2, 1, 1, 1);
    tol_set_problem(task, n++, 4, 2, 1, 1, 1, 2, 2, 3, 2, 3, 1, 3, 3);
    /* Karolina's "direct" tasks: */
    tol_set_problem(task, n++, 5, 3, 3, 3, 1, 3, 2, 3, 2, 1, 1, 3, 1);
    tol_set_problem(task, n++, 5, 3, 2, 1, 1, 3, 1, 3, 1, 2, 1, 3, 2);
    tol_set_problem(task, n++, 5, 1, 1, 2, 2, 2, 1, 1, 1, 2, 1, 3, 1);
    /* Karolina's "indirect" tasks: */
    tol_set_problem(task, n++, 5, 3, 1, 2, 1, 2, 2, 1, 1, 2, 2, 2, 1);
    tol_set_problem(task, n++, 6, 3, 1, 2, 1, 3, 2, 3, 2, 3, 1, 1, 1);
    tol_set_problem(task, n++, 6, 2, 2, 2, 1, 1, 1, 2, 1, 2, 2, 1, 1);

    task->trial_count = 0;
    tol_initialise_trial(task);
}

void tol_initialise_next_trial(ContentionScheduler *cs, TolTaskData *task)
{
    task->trial_count++;
    tol_initialise_trial(task);
}

/******************************************************************************/
/* Section 3: Detect trial/task completion -----------------------------------*/

Boolean tol_trial_complete(TolTaskData *task)
{
    return(task->problems[task->trial_count].trial_done);
}

Boolean tol_task_complete(TolTaskData *task)
{
    return(tol_trial_complete(task) && (task->trial_count >= (TOL_PROBLEM_MAX-1)));
}

/******************************************************************************/
/* Section 4: Score the current trail ----------------------------------------*/

static Boolean position_match(BallPosition *p1, BallPosition *p2)
{
    return((p1->peg == p2->peg) && (p1->position == p2->position));
}

static void tol_score_trial(TolTaskData *task)
{
    /* We just need to score how many balls are in place. Everything else   */
    /* should be scored along the way.                                      */

    TolProblem *p = &(task->problems[task->trial_count]);

    p->balls_correct = 0;
    if (position_match(&(p->current_state.red_ball), &(p->desired_state.red_ball))) {
        p->balls_correct++;
    }
    if (position_match(&(p->current_state.green_ball), &(p->desired_state.green_ball))) {
        p->balls_correct++;
    }
    if (position_match(&(p->current_state.blue_ball), &(p->desired_state.blue_ball))) {
        p->balls_correct++;
    }
}

/******************************************************************************/
/* Section 5: ToL events -----------------------------------------------------*/

typedef struct tol_goal_state {
    char                *colour;
    BallPosition         position;
} TolGoalState;

typedef struct subgoal_list {
    TolGoalState           this;
    struct subgoal_list   *tail;
} SubgoalList;

static void print_state(FILE *fp, char *label, TolConfiguration *current)
{
    fprintf(fp, "%s: \n", label);
    fprintf(fp, "    RED on %d.%d\n", current->red_ball.peg, current->red_ball.position);
    fprintf(fp, "    GREEN on %d.%d\n", current->green_ball.peg, current->green_ball.position);
    fprintf(fp, "    BLUE on %d.%d\n", current->blue_ball.peg, current->blue_ball.position);
}

static void schema_name_list_free(SchemaNameList *s)
{
    while (s != NULL) {
        SchemaNameList *t = s->next;
        free(s->name);
        free(s);
        s = t;
    }
}

static SchemaNameList *schema_name_list_append(SchemaNameList *s, char *name)
{
    SchemaNameList *new;

fprintf(stdout, "    Appending %s to strategy\n", name);

    if ((new = (SchemaNameList *)malloc(sizeof(SchemaNameList))) != NULL) {
        new->name = string_copy(name);
        new->next = NULL;
    }
    if (s == NULL) {
        return(new);
    }
    else {
        SchemaNameList *t = s;
        while (t->next != NULL) {
            t = t->next;
        }
        t->next = new;
        return(s);
    }
}

static char *tol_schema_get_name(char *colour, int peg)
{
fprintf(stdout, "    We want to move %s to peg %d\n", colour, peg);

    if (strcmp(colour, "red") == 0)  {
        if (peg == 1) {
            return("schemaMoveRedToOne");
        }
        else if (peg == 2) {
            return("schemaMoveRedToTwo");
        }
        else if (peg == 3) {
            return("schemaMoveRedToThree");
        }
    }
    if (strcmp(colour, "green") == 0)  {
        if (peg == 1) {
            return("schemaMoveGreenToOne");
        }
        else if (peg == 2) {
            return("schemaMoveGreenToTwo");
        }
        else if (peg == 3) {
            return("schemaMoveGreenToThree");
        }
    }
    if (strcmp(colour, "blue") == 0)  {
        if (peg == 1) {
            return("schemaMoveBlueToOne");
        }
        else if (peg == 2) {
            return("schemaMoveBlueToTwo");
        }
        else if (peg == 3) {
            return("schemaMoveBlueToThree");
        }
    }
    return(NULL);
}

static char *top_most_ball(TolConfiguration *state, int peg)
{
    char *colour = NULL;
    int position = 0;

    if ((state->red_ball.peg == peg) && (state->red_ball.position > position)) {
        position = state->red_ball.position;
        colour = "red";
    }
    if ((state->green_ball.peg == peg) && (state->green_ball.position > position)) {
        position = state->green_ball.position;
        colour = "green";
    }
    if ((state->blue_ball.peg == peg) && (state->blue_ball.position > position)) {
        position = state->blue_ball.position;
        colour = "blue";
    }

    return(colour);
}

BallPosition *locate_ball(TolConfiguration *state, char *colour)
{
    if (strcmp(colour, "red") == 0) {
        return(&state->red_ball);
    }
    if (strcmp(colour, "green") == 0) {
        return(&state->green_ball);
    }
    if (strcmp(colour, "blue") == 0) {
        return(&state->blue_ball);
    }
    return(NULL);
}

static int count_balls_on_peg(TolConfiguration *state, int peg)
{
    int count = 0;

    if (state->red_ball.peg == peg) {
        count++;
    }
    if (state->green_ball.peg == peg) {
        count++;
    }
    if (state->blue_ball.peg == peg) {
        count++;
    }
    return(count);
}

static void apply_move(TolConfiguration *state, char *ball, int peg)
{
    BallPosition *bp = NULL;

    if ((bp = locate_ball(state, ball)) != NULL) {
        bp->peg = peg;
        bp->position = count_balls_on_peg(state, bp->peg);
    }
}

static Boolean peg_is_full(TolConfiguration *state, int peg)
{
    return(count_balls_on_peg(state, peg) >= peg);
}

/*----------------------------------------------------------------------------*/
        
static int get_best_temporary_peg(TolConfiguration *state, int source_peg, char *colour)
{
    Boolean possible_pegs[4] = {TRUE, TRUE, TRUE, TRUE};
    BallPosition *bp;

    possible_pegs[source_peg] = FALSE;

    if ((source_peg != 1) && peg_is_full(state, 1)) {
        possible_pegs[1] = FALSE;
    }
    if ((source_peg != 2) && peg_is_full(state, 2)) {
        possible_pegs[2] = FALSE;
    }
    if ((source_peg != 3) && peg_is_full(state, 3)) {
        possible_pegs[3] = FALSE;
    }

    if ((bp = locate_ball(state, colour)) != NULL) {
        possible_pegs[bp->peg] = FALSE;
    }

    if (possible_pegs[1]) {
        return(1);
    }
    else if (possible_pegs[2]) {
        return(2);
    }
    else if (possible_pegs[3]) {
        return(3);
    }
    else {
        return(0);
    }
}

static SchemaNameList *tol_ps_generate_level1_strategy_for_subgoal(TolConfiguration *current, TolGoalState *subgoal)
{
    TolConfiguration hypothetical_state;
    BallPosition *source_position;
    int non_target_peg;
    SchemaNameList *s = NULL;
    char *ball;

    tol_state_copy(&hypothetical_state, current);
    source_position = locate_ball(&hypothetical_state, subgoal->colour);
    if (subgoal->position.peg == source_position->peg) {

        /* The ball is already on the target peg. Presumably we need to     */
        /* remove one or more from under it, and we don't want to put one   */
        /* that is currently under it on top of it while we're working.     */
        /* Now, the source peg can't be peg 1 since that can't have two     */
        /* balls on it. So the source peg is peg 2 or peg 3.                */

        /* If the source peg is 3, then we need to move any non-focus ball  */
        /* from peg 3 to peg 2, then move the focus-ball to peg 1, then     */
        /* move any non-focus ball from peg 3 to peg 2, then move the       */
        /* focus-ball back to peg 3.                                        */

        /* If the source peg is 2, then we need to move any non-focus ball  */
        /* to peg 3, the focus-ball to peg 1, any non-focus ball to peg 3,  */
        /* then the focus-ball back to peg 2.                               */

        /* Either way, if peg 1 is occupied then the subgoal can't be       */
        /* achieved so simply.                                              */

        /* Hence if peg 1 is occupied the first thing to do is to move the  */
        /* ball from peg 1 to a temporary peg (not 1 and not the source     */
        /* peg).                                                            */

        if ((ball = top_most_ball(&hypothetical_state, 1)) != NULL) {
            non_target_peg = 5 - source_position->peg;
            s = schema_name_list_append(s, tol_schema_get_name(ball, non_target_peg));
            apply_move(&hypothetical_state, ball, non_target_peg);
        }

        while (count_balls_on_peg(&hypothetical_state, subgoal->position.peg) >= subgoal->position.position) {
            ball = top_most_ball(&hypothetical_state, subgoal->position.peg);
            if ((non_target_peg = get_best_temporary_peg(&hypothetical_state, subgoal->position.peg, subgoal->colour)) == 0) {
                schema_name_list_free(s);
                return(NULL);
            }
            s = schema_name_list_append(s, tol_schema_get_name(ball, non_target_peg));
            apply_move(&hypothetical_state, ball, non_target_peg);
        }

    }
    else {
        non_target_peg = 6 - subgoal->position.peg - source_position->peg;

        /* First plan to move any balls above the focus ball off the focus peg: */
        ball = top_most_ball(&hypothetical_state, source_position->peg);
        while ((ball != NULL) && (strcmp(ball, subgoal->colour) != 0)) {
            if (peg_is_full(&hypothetical_state, non_target_peg)) {
                // non-target peg is full: abort strategy
                schema_name_list_free(s);
                return(NULL);
            }
            s = schema_name_list_append(s, tol_schema_get_name(ball, non_target_peg));
            apply_move(&hypothetical_state, ball, non_target_peg);
            ball = top_most_ball(&hypothetical_state, source_position->peg);
        }

        /* Now plan to move balls off the target peg if necessary:              */
        while (count_balls_on_peg(&hypothetical_state, subgoal->position.peg) >= subgoal->position.position) {
            ball = top_most_ball(&hypothetical_state, subgoal->position.peg);
            if (peg_is_full(&hypothetical_state, non_target_peg)) {
                // non-target peg is full: abort strategy
                schema_name_list_free(s);
                return(NULL);
            }
            s = schema_name_list_append(s, tol_schema_get_name(ball, non_target_peg));
            apply_move(&hypothetical_state, ball, non_target_peg);
        }
    }

    /* Finally, plan to move the focus ball to its target peg/position:     */

    source_position = locate_ball(&hypothetical_state, subgoal->colour);

    if (peg_is_full(&hypothetical_state, subgoal->position.peg)) {
        // non-target peg is full: abort strategy
        schema_name_list_free(s);
        return(NULL);
    }
    else if (strcmp(top_most_ball(&hypothetical_state, source_position->peg), subgoal->colour) != 0) {
        // focus ball is not on top: abort strategy
        schema_name_list_free(s);
        return(NULL);
    }
    else if (source_position->peg == subgoal->position.peg) {
        // focus ball is already on the target peg, but in the wrong position: abort strategy
        schema_name_list_free(s);
        return(NULL);
    }
    else {
        s = schema_name_list_append(s, tol_schema_get_name(subgoal->colour, subgoal->position.peg));
        return(s);
    }
}

static SubgoalList *subgoals_prepend(SubgoalList *subgoals, char *colour, BallPosition *p)
{
    SubgoalList *new;

    if ((new = (SubgoalList *)malloc(sizeof(SubgoalList))) != NULL) {
        new->this.colour = string_copy(colour);
        new->this.position.position = p->position;
        new->this.position.peg = p->peg;
        new->tail = subgoals;
        return(new);
    }
    else {
        return(subgoals);
    }
}

static Boolean subgoals_member(SubgoalList *subgoals, char *colour, BallPosition *pos)
{
    while (subgoals != NULL) {
        if ((strcmp(subgoals->this.colour, colour) == 0) && position_match(&(subgoals->this.position), pos)) {
            return(TRUE);
        }
        subgoals = subgoals->tail;
    }
    return(FALSE);
}

static TolGoalState *tol_goal_copy(TolGoalState *old)
{
    TolGoalState *new;

    if ((new = (TolGoalState *)malloc(sizeof(TolGoalState))) != NULL) {
        new->colour = string_copy(old->colour);
        new->position.position = old->position.position;
        new->position.peg = old->position.peg;
    }
    return(new);
}

static TolGoalState *subgoal_select_at_random(SubgoalList *subgoals)
{
    int l = 0, p;
    SubgoalList *sg;

    for (sg = subgoals; sg != NULL; sg = sg->tail) {
        l++;
    }
    p = (int) floor(l * (random() / pow(2, 31)));
    for (sg = subgoals; sg != NULL; sg = sg->tail) {
        if (--p < 0) {
            return(tol_goal_copy(&(sg->this)));
	}
    }
    return(NULL);
}

static void subgoal_dispose_list(SubgoalList *subgoals)
{
    while (subgoals != NULL) {
        SubgoalList *tmp = subgoals;
        subgoals = subgoals->tail;
        free(tmp->this.colour);
        free(tmp);
    }
}

static Boolean goal_blocked(TolConfiguration *current, TolConfiguration *desired, char *goal_colour, BallPosition *goal_position)
{
    /* Goal is blocked if there's a ball under the goal ball in the     */
    /* goal configuration (on the same peg) that is not there in the    */
    /* current configuration.                                           */

    if ((strcmp(goal_colour, "red") != 0) && (desired->red_ball.peg == goal_position->peg) && (desired->red_ball.position < goal_position->position) && !position_match(&(desired->red_ball), &(current->red_ball))) {
        return(TRUE);
    }
    if ((strcmp(goal_colour, "green") != 0) && (desired->green_ball.peg == goal_position->peg) && (desired->green_ball.position < goal_position->position) && !position_match(&(desired->green_ball), &(current->green_ball))) {
        return(TRUE);
    }
    if ((strcmp(goal_colour, "blue") != 0) && (desired->blue_ball.peg == goal_position->peg) && (desired->blue_ball.position < goal_position->position) && !position_match(&(desired->blue_ball), &(current->blue_ball))) {
        return(TRUE);
    }
    return(FALSE);
}

static TolGoalState *tol_ps_generate_subgoal(TolConfiguration *current, TolConfiguration *desired, SubgoalList *bad_subgoals)
{
    /* Locate a ball that is out of place and return a goal state       */
    /* structure representing it's correct/desired position. Favour     */
    /* balls that should be on lower peg positions over those that      */
    /* belong on higher positions. Return NULL if we can't find a ball  */
    /* that is out of place.                                            */

    SubgoalList *possible_subgoals = NULL;
    TolGoalState *result;

    if (!position_match(&(current->red_ball), &(desired->red_ball))) {
        if (subgoals_member(bad_subgoals, "red", &(desired->red_ball))) {
            /* We've already tried and failed to generate a strategy for this subgoal */
        }
        else if (goal_blocked(current, desired, "red", &(desired->red_ball))) {
            /* The subgoal can't be achieved until some other subgoal is achieved */
        }
        else {
            /* The subgoal looks plausible... */
            possible_subgoals = subgoals_prepend(possible_subgoals, "red", &(desired->red_ball));
        }
    }

    if (!position_match(&(current->green_ball), &(desired->green_ball))) {
        if (subgoals_member(bad_subgoals, "green", &(desired->green_ball))) {
            /* We've already tried and failed to generate a strategy for this subgoal */
        }
        else if (goal_blocked(current, desired, "green", &(desired->green_ball))) {
            /* The subgoal can't be achieved until some other subgoal is achieved */
        }
        else {
            /* The subgoal looks plausible... */
            possible_subgoals = subgoals_prepend(possible_subgoals, "green", &(desired->green_ball));
        }
    }

    if (!position_match(&(current->blue_ball), &(desired->blue_ball))) {
        if (subgoals_member(bad_subgoals, "blue", &(desired->blue_ball))) {
            /* We've already tried and failed to generate a strategy for this subgoal */
        }
        else if (goal_blocked(current, desired, "blue", &(desired->blue_ball))) {
            /* The subgoal can't be achieved until some other subgoal is achieved */
        }
        else {
            /* The subgoal looks plausible... */
            possible_subgoals = subgoals_prepend(possible_subgoals, "blue", &(desired->blue_ball));
        }
    }

    if (possible_subgoals != NULL) {
        result = subgoal_select_at_random(possible_subgoals);
        subgoal_dispose_list(possible_subgoals);
fprintf(stdout, "Selected subgoal: %s to %d.%d\n", result->colour, result->position.peg, result->position.position);
        return(result);
    }
    else {
        return(NULL);
    }
}


static SchemaNameList *tol_ps_generate_strategy(ContentionScheduler *cs, TolConfiguration *current, TolConfiguration *desired)
{
    /* Generate a strategy (schema sequence) to move from the current state */
    /* towards the desired state.                                           */

    SchemaNameList *strategy;
    SubgoalList *bad_subgoals;
    TolGoalState *subgoal;

// bad_subgoals is effectively a buffer used to store temporary
// results (subgoals that are unhelpful)
// strategy is an ordered list of schema names. This too requires
// a buffer - but now a buffer with order
// Is working memory used to store all of these things?

    bad_subgoals = NULL;
    while ((subgoal = tol_ps_generate_subgoal(current, desired, bad_subgoals)) != NULL) {
        if ((strategy = tol_ps_generate_level1_strategy_for_subgoal(current, subgoal)) != NULL) {
            subgoal_dispose_list(bad_subgoals);
            return(strategy);
        }
        else {
            bad_subgoals = subgoals_prepend(bad_subgoals, subgoal->colour, &(subgoal->position));
        }
    }
    subgoal_dispose_list(bad_subgoals);

// Try again, but consider more complex strategies...

    return(NULL);
}

char *strategy_pop(ContentionScheduler *cs)
{
    char *top = NULL;

    if (cs->strategy != NULL) {
        SchemaNameList *this = cs->strategy;
        cs->strategy = this->next;
        top = this->name;
        free(this);
    }
    return(top);
}

/*----------------------------------------------------------------------------*/

void tol_step_event(ContentionScheduler *cs, TolTaskData *task)
{
    /* We use a generic subgoaling approach. If we're in the process of */
    /* performing a series of planned schemas, attempt the next step.   */
    /* Otherwise attempt to generate a subgoal. This may fail if the    */
    /* task is complete, or if we're just stuck. If we succeed in       */
    /* generating a subgoal, attempt to generate a strategy to achieve  */
    /* that subgoal. If we can't generate a strategy, give up. This     */
    /* loop should probably be supplemented with a monitoring mechanism */
    /* to allow abandonment of a strategy that doesn't appear to be     */
    /* working. Perhaps a strategy should bring with it a list of       */
    /* anticipated states that can be used to check progress.           */

    TolConfiguration *current_state = &(task->problems[task->trial_count].current_state);
    TolConfiguration *desired_state = &(task->problems[task->trial_count].desired_state);
    SchemaNameList *strategy;
    char *schema_name;

    task->event_count++;

fprintf(stdout, "Stepping to event %d\n", task->event_count);
print_state(stdout, "CURRENT", current_state);
print_state(stdout, "DESIRED", desired_state);

    if ((schema_name = strategy_pop(cs)) != NULL) {
        cs_schema_set_intended(cs, schema_name, TRUE);
        cs_run_task(cs, task, 50);
        free(schema_name);
    }
    else if ((strategy = tol_ps_generate_strategy(cs, current_state, desired_state)) != NULL) {
        cs->strategy = strategy;
    }
    else { /* Giving up: */
        task->problems[task->trial_count].trial_done = TRUE;
        tol_score_trial(task);
    }
}

/******************************************************************************/
/* Section 6: Synchronise CS and external world representations --------------*/

static Object *set_ball_features(ContentionScheduler *cs, char *name, BallPosition *p)
{
    Object *this = objectGetByName(cs, name);

    if (this != NULL) {
        objectSetPosition(this, p->peg, p->position);
    }
    return(this);
}

void tol_synchronise_world(ContentionScheduler *cs, TolTaskData *task)
{
    set_ball_features(cs, "ball_current_red", &(task->problems[task->trial_count].current_state.red_ball));
    set_ball_features(cs, "ball_current_green", &(task->problems[task->trial_count].current_state.green_ball));
    set_ball_features(cs, "ball_current_blue", &(task->problems[task->trial_count].current_state.blue_ball));
    set_ball_features(cs, "ball_goal_red", &(task->problems[task->trial_count].desired_state.red_ball));
    set_ball_features(cs, "ball_goal_green", &(task->problems[task->trial_count].desired_state.green_ball));
    set_ball_features(cs, "ball_goal_blue", &(task->problems[task->trial_count].desired_state.blue_ball));
}

/******************************************************************************/
/* Section 7: Dump the analysis results --------------------------------------*/

void tol_dump_analyses(FILE *fp, TolTaskData *task)
{
    int mm = 0, ma = 0, rb = 0, bc = 0;
    int i;

    fprintf(fp, "\nProblem Id   Min. Moves    Actual Moves   Rule Breaks   Correct Balls\n");
    fprintf(fp, "------------+-------------+-------------+-------------+--------------\n");

    for (i = 0; i <= task->trial_count; i++) {
        TolProblem *problem = &(task->problems[i]);

        mm += problem->moves_minimum;
        ma += problem->moves_actual;
        rb += problem->rule_breaks;
        bc += problem->balls_correct;
        fprintf(fp, " Problem %d     %4d          %4d          %4d          %4d\n", i + 1, problem->moves_minimum, problem->moves_actual, problem->rule_breaks, problem->balls_correct);
    }

    fprintf(fp, "------------+-------------+-------------+-------------+--------------\n");
    fprintf(fp, "Mean           %6.3f        %6.3f        %6.3f        %6.3f\n", mm / (double) i, ma / (double) i, rb / (double) i, bc / (double) i);
}

/******************************************************************************/
/* Section 8: Actions that alter the world: ----------------------------------*/

static int get_peg_number(Object *target)
{
    if (strcmp(target->name, "peg_current_1") == 0) {
        return(1);
    }
    else  if (strcmp(target->name, "peg_current_2") == 0) {
        return(2);
    }
    else if (strcmp(target->name, "peg_current_3") == 0) {
        return(3);
    }
    else {
        return(0);
    }
}

/*----------------------------------------------------------------------------*/

void tol_action_pick_up(TolTaskData *task, Object *theme, Effector *effector)
{
    TolProblem *p = &(task->problems[task->trial_count]);
    BallPosition *bp = NULL;

    if (strcmp(theme->name, "ball_current_red") == 0) {
        bp = &(p->current_state.red_ball);
    }
    else if (strcmp(theme->name, "ball_current_green") == 0) {
        bp = &(p->current_state.green_ball);
    }
    else if (strcmp(theme->name, "ball_current_blue") == 0) {
        bp = &(p->current_state.blue_ball);
    }
    else if (strcmp(theme->name, "ball_goal_red") == 0) {
        bp = &(p->desired_state.red_ball);
    }
    else if (strcmp(theme->name, "ball_goal_green") == 0) {
        bp = &(p->desired_state.green_ball);
    }
    else if (strcmp(theme->name, "ball_goal_blue") == 0) {
        bp = &(p->desired_state.blue_ball);
    }

    if (bp != NULL) {
        bp->peg = 0;
        bp->position = 0;
        objectSetPosition(theme, bp->peg, bp->position);
    }
}

/*----------------------------------------------------------------------------*/

void tol_action_put_down(TolTaskData *task, Object *theme, Object *target)
{
    TolProblem *p = &(task->problems[task->trial_count]);
    BallPosition *bp = NULL;

    if (strcmp(theme->name, "ball_current_red") == 0) {
        bp = &(p->current_state.red_ball);
    }
    else if (strcmp(theme->name, "ball_current_green") == 0) {
        bp = &(p->current_state.green_ball);
    }
    else if (strcmp(theme->name, "ball_current_blue") == 0) {
        bp = &(p->current_state.blue_ball);
    }
    else if (strcmp(theme->name, "ball_goal_red") == 0) {
        bp = &(p->desired_state.red_ball);
    }
    else if (strcmp(theme->name, "ball_goal_green") == 0) {
        bp = &(p->desired_state.green_ball);
    }
    else if (strcmp(theme->name, "ball_goal_blue") == 0) {
        bp = &(p->desired_state.blue_ball);
    }

    if (bp != NULL) {
        bp->peg = get_peg_number(target);
        bp->position = count_balls_on_peg(&(p->current_state), bp->peg);
        objectSetPosition(theme, bp->peg, bp->position);
    }

    task->problems[task->trial_count].moves_actual++;
}

/******************************************************************************/
