/*******************************************************************************

    File:     tol_schema.c
    Author:   Rick Cooper
    Created:  30/03/07
    Edited:   30/03/07
    Contents: Create/maintain schemas for the ToL task

    Public procedures:
        void install_tol_schemas(ContentionScheduler *cs)

*******************************************************************************/

#include "ss.h"
#include "tol.h"

/******************************************************************************/
/* Section 1: ToL-specific object test/manipulation functions ----------------*/

static Boolean is_movable_ball(ContentionScheduler *cs, Object *o)
{
    ObjectList *ol;

    if ((o->features & IS_BALL) && (o->features & IS_APPARATUS)) { 
        for (ol = cs->objects; ol != NULL; ol = ol->tail) {
            Object *this = ol->head;
            if ((this->features & IS_BALL) && (this->features & IS_APPARATUS)) {
                if ((this->position[0] == o->position[0]) && (this->position[1] > o->position[1])) {
                    return(FALSE);
                }
            }
        }
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static int balls_on_peg(ContentionScheduler *cs, int peg)
{
    ObjectList *ol;
    int count = 0;

    for (ol = cs->objects; ol != NULL; ol = ol->tail) {
        Object *this = ol->head;
        if ((this->features & IS_BALL) && (this->features & IS_APPARATUS)) {
            if (this->position[0] == peg) {
                count++;
            }
        }
    }
    return(count);
}


static Boolean is_peg_with_space(ContentionScheduler *cs, Object *o)
{
    if (strcmp(o->name, "peg_current_1") == 0) {
        return(balls_on_peg(cs, 1) < 1);
    }
    else if (strcmp(o->name, "peg_current_2") == 0) {
        return(balls_on_peg(cs, 2) < 2);
    }
    else if (strcmp(o->name, "peg_current_3") == 0) {
        return(balls_on_peg(cs, 3) < 3);
    }
    else {
        return(FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static double ball_is_movable(ContentionScheduler *cs, Object *o, char *name)
{
    if ((strcmp(o->name, name) == 0) && is_movable_ball(cs, o)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double peg_is_potential_target(ContentionScheduler *cs, Object *o, char *name)
{
    if ((strcmp(o->name, name) == 0) && is_peg_with_space(cs, o)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static GoalType tol_goal_red_on_one(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_red");

    if ((this != NULL) && (this->position[0] == 1.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_red_on_two(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_red");

    if ((this != NULL) && (this->position[0] == 2.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_red_on_three(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_red");

    if ((this != NULL) && (this->position[0] == 3.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static GoalType tol_goal_green_on_one(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_green");

    if ((this != NULL) && (this->position[0] == 1.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_green_on_two(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_green");

    if ((this != NULL) && (this->position[0] == 2.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_green_on_three(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_green");

    if ((this != NULL) && (this->position[0] == 3.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static GoalType tol_goal_blue_on_one(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_blue");

    if ((this != NULL) && (this->position[0] == 1.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_blue_on_two(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_blue");

    if ((this != NULL) && (this->position[0] == 2.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

static GoalType tol_goal_blue_on_three(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "ball_current_blue");

    if ((this != NULL) && (this->position[0] == 3.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

/******************************************************************************/
/* Section 2: Schema triggering functions ------------------------------------*/

static double schemaMoveRedBallTriggersTheme(ContentionScheduler *cs, Object *o)
{
    return(ball_is_movable(cs, o, "ball_current_red"));
}

static double schemaMoveBlueBallTriggersTheme(ContentionScheduler *cs, Object *o)
{
    return(ball_is_movable(cs, o, "ball_current_blue"));
}

static double schemaMoveGreenBallTriggersTheme(ContentionScheduler *cs, Object *o)
{
    return(ball_is_movable(cs, o, "ball_current_green"));
}

/*----------------------------------------------------------------------------*/

static double schemaMoveBallTriggersPeg1(ContentionScheduler *cs, Object *o)
{
    return(peg_is_potential_target(cs, o, "peg_current_1"));
}

static double schemaMoveBallTriggersPeg2(ContentionScheduler *cs, Object *o)
{
    return(peg_is_potential_target(cs, o, "peg_current_2"));
}

static double schemaMoveBallTriggersPeg3(ContentionScheduler *cs, Object *o)
{
    return(peg_is_potential_target(cs, o, "peg_current_3"));
}

/*----------------------------------------------------------------------------*/

static double schemaPickUpBallTriggersTheme(ContentionScheduler *cs, Object *o)
{
/*  An ball is triggered as a theme for picking up if it is a ball,     */
/*  is not part of the goal apparatus, and it is not already held.      */

    if (isHeldBoolean(o)) {
        return(0.0);
    }
    else if (is_movable_ball(cs, o)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaPutDownBallTriggersTarget(ContentionScheduler *cs, Object *o)
{
/*  An ball is triggered as a theme for picking up if it is a ball,     */
/*  is not part of the goal apparatus, and it is not already held.      */

    if (is_peg_with_space(cs, o)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaPickUpBallTriggersEffector(ContentionScheduler *cs, Effector *e)
{
/*  Effectors are triggered for pick-up actions if they hold nothing. */

    return(isEmptyEffectorFuzzy(e));
}

/*----------------------------------------------------------------------------*/

static double schemaMoveBallTriggersEffector(ContentionScheduler *cs, Effector *e)
{
    return(0.0);
}

/*----------------------------------------------------------------------------*/

static double schemaPutDownBallTriggersEffector(ContentionScheduler *cs, Effector *e)
{
/*  Effectors are triggered for put-down actions if they hold something. */
    
    return(isFullEffectorFuzzy(e));
}

/******************************************************************************/
/* Section 3: Install ToL schemas --------------------------------------------*/

void install_tol_schemas(ContentionScheduler *cs)
{
    Schema *s;

    s = install_schema(cs, "schemaMoveRedToThree", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveRedBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg3);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_red_on_three);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveRedToTwo", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveRedBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg2);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_red_on_two);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveRedToOne", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveRedBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg1);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_red_on_one);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveGreenToThree", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveGreenBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg3);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_green_on_three);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveGreenToTwo", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveGreenBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg2);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_green_on_two);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveGreenToOne", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveGreenBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg1);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_green_on_one);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveBlueToThree", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveBlueBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg3);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_blue_on_three);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveBlueToTwo", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveBlueBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg2);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_blue_on_two);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaMoveBlueToOne", "goalMoveBall", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMoveBlueBallTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMoveBallTriggersPeg1);
    install_schema_effector_trigger(s, schemaMoveBallTriggersEffector);
    install_schema_subgoal(s, "goalPickUpBall", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPutDownBall", goalHandsFull, tol_goal_blue_on_one);
    install_schema_scope(s, THEME);
    install_schema_scope(s, TARGET);

    s = install_schema(cs, "schemaPickUpBall", "goalPickUpBall", ACTION_PICK_UP_THEME, LOW);
    install_schema_object_trigger(s, THEME, schemaPickUpBallTriggersTheme);
    install_schema_effector_trigger(s, schemaPickUpBallTriggersEffector);

    s = install_schema(cs, "schemaPutDownBall", "goalPutDownBall", ACTION_PUT_DOWN_TARGET, LOW);
    install_schema_object_trigger(s, TARGET, schemaPutDownBallTriggersTarget);
    install_schema_effector_trigger(s, schemaPutDownBallTriggersEffector);
}

/******************************************************************************/

void tol_initialise_wm(ContentionScheduler *cs)
{

}

/******************************************************************************/
