/*******************************************************************************

    File:     wcst.c
    Author:   Rick Cooper
    Created:  28/03/07
    Edited:   03/04/07
    Contents: Routines for administering and scoring the WCST 

    Public procedures:
        WcstTaskData *wcst_create_task()
        void wcst_destroy_task(WcstTaskData *task)
        void wcst_initialise_task(WcstTaskData *task)
        void wcst_initialise_next_trial(ContentionScheduler *cs, WcstTaskData *task)
        Boolean wcst_trial_complete(WcstTaskData *task)
        Boolean wcst_task_complete(WcstTaskData *task)
        void wcst_step_event(ContentionScheduler *cs, WcstTaskData *task)
        void wcst_synchronise_world(ContentionScheduler *cs, WcstTaskData *task)
        void wcst_action_pick_up(WcstTaskData *task, Object *theme, Object *target)
        void wcst_action_put_down(WcstTaskData *task, Object *theme, Object *target)

*******************************************************************************/

#include "ss.h"
#include "wcst.h"
#include <string.h>
#include <math.h>

/******************************************************************************/
/* Section 1: Data structure creation/disposal -------------------------------*/

WcstTaskData *wcst_create_task()
{
    return((WcstTaskData *)malloc(sizeof(WcstTaskData)));
}

void wcst_destroy_task(WcstTaskData *task)
{
    if (task != NULL) {
        free(task);
    }
}

/******************************************************************************/
/* Section 2: Task initialisation --------------------------------------------*/

static void initialise_card(Card *card, int n, int c, int f, int x, int y)
{
    card->number = n;
    card->colour = c;
    card->form = f;
    card->position_x = x;
    card->position_y = y;
}

/*----------------------------------------------------------------------------*/

static void initialise_card_lists(WcstTaskData *task)
{
    int i;

    initialise_card(&(task->card_target[0]), 1, 1, 1, 1, 1);
    initialise_card(&(task->card_target[1]), 2, 2, 2, 1, 2);
    initialise_card(&(task->card_target[2]), 3, 3, 3, 1, 3);
    initialise_card(&(task->card_target[3]), 4, 4, 4, 1, 4);

    initialise_card(&(task->card_sorted[0]), 0, 0, 0, 2, 1);
    initialise_card(&(task->card_sorted[1]), 0, 0, 0, 2, 2);
    initialise_card(&(task->card_sorted[2]), 0, 0, 0, 2, 3);
    initialise_card(&(task->card_sorted[3]), 0, 0, 0, 2, 4);

    initialise_card(&(task->card_list[0]), 1, 2, 1, -1, -1);
    initialise_card(&(task->card_list[1]), 4, 1, 3, -1, -1);
    initialise_card(&(task->card_list[2]), 2, 4, 1, -1, -1);
    initialise_card(&(task->card_list[3]), 1, 1, 4, -1, -1);
    initialise_card(&(task->card_list[4]), 4, 2, 2, -1, -1);
    initialise_card(&(task->card_list[5]), 1, 3, 3, -1, -1);
    initialise_card(&(task->card_list[6]), 4, 4, 1, -1, -1);
    initialise_card(&(task->card_list[7]), 3, 1, 4, -1, -1);
    initialise_card(&(task->card_list[8]), 4, 2, 3, -1, -1);
    initialise_card(&(task->card_list[9]), 2, 3, 4, -1, -1);
    initialise_card(&(task->card_list[10]), 1, 4, 2, -1, -1);
    initialise_card(&(task->card_list[11]), 3, 1, 1, -1, -1);
    initialise_card(&(task->card_list[12]), 2, 4, 3, -1, -1);
    initialise_card(&(task->card_list[13]), 1, 3, 2, -1, -1);
    initialise_card(&(task->card_list[14]), 3, 2, 1, -1, -1);
    initialise_card(&(task->card_list[15]), 4, 4, 4, -1, -1);
    initialise_card(&(task->card_list[16]), 2, 1, 2, -1, -1);
    initialise_card(&(task->card_list[17]), 3, 3, 4, -1, -1);
    initialise_card(&(task->card_list[18]), 4, 1, 1, -1, -1);
    initialise_card(&(task->card_list[19]), 1, 3, 4, -1, -1);
    initialise_card(&(task->card_list[20]), 2, 4, 2, -1, -1);
    initialise_card(&(task->card_list[21]), 3, 2, 3, -1, -1);
    initialise_card(&(task->card_list[22]), 2, 3, 2, -1, -1);
    initialise_card(&(task->card_list[23]), 3, 4, 1, -1, -1);
    initialise_card(&(task->card_list[24]), 4, 1, 2, -1, -1);
    initialise_card(&(task->card_list[25]), 2, 3, 1, -1, -1);
    initialise_card(&(task->card_list[26]), 3, 4, 3, -1, -1);
    initialise_card(&(task->card_list[27]), 4, 1, 4, -1, -1);
    initialise_card(&(task->card_list[28]), 2, 2, 2, -1, -1);
    initialise_card(&(task->card_list[29]), 1, 1, 3, -1, -1);
    initialise_card(&(task->card_list[30]), 4, 2, 4, -1, -1);
    initialise_card(&(task->card_list[31]), 1, 1, 2, -1, -1);
    initialise_card(&(task->card_list[32]), 3, 4, 4, -1, -1);
    initialise_card(&(task->card_list[33]), 4, 3, 3, -1, -1);
    initialise_card(&(task->card_list[34]), 1, 2, 2, -1, -1);
    initialise_card(&(task->card_list[35]), 4, 3, 1, -1, -1);
    initialise_card(&(task->card_list[36]), 2, 4, 4, -1, -1);
    initialise_card(&(task->card_list[37]), 3, 3, 2, -1, -1);
    initialise_card(&(task->card_list[38]), 1, 2, 4, -1, -1);
    initialise_card(&(task->card_list[39]), 4, 4, 3, -1, -1);
    initialise_card(&(task->card_list[40]), 1, 1, 1, -1, -1);
    initialise_card(&(task->card_list[41]), 3, 4, 2, -1, -1);
    initialise_card(&(task->card_list[42]), 4, 3, 4, -1, -1);
    initialise_card(&(task->card_list[43]), 2, 2, 1, -1, -1);
    initialise_card(&(task->card_list[44]), 3, 1, 2, -1, -1);
    initialise_card(&(task->card_list[45]), 2, 3, 3, -1, -1);
    initialise_card(&(task->card_list[46]), 1, 4, 4, -1, -1);
    initialise_card(&(task->card_list[47]), 3, 3, 1, -1, -1);
    initialise_card(&(task->card_list[48]), 2, 2, 3, -1, -1);
    initialise_card(&(task->card_list[49]), 4, 3, 2, -1, -1);
    initialise_card(&(task->card_list[50]), 1, 4, 1, -1, -1);
    initialise_card(&(task->card_list[51]), 3, 2, 2, -1, -1);
    initialise_card(&(task->card_list[52]), 2, 1, 4, -1, -1);
    initialise_card(&(task->card_list[53]), 4, 2, 1, -1, -1);
    initialise_card(&(task->card_list[54]), 1, 4, 3, -1, -1);
    initialise_card(&(task->card_list[55]), 3, 2, 4, -1, -1);
    initialise_card(&(task->card_list[56]), 2, 1, 3, -1, -1);
    initialise_card(&(task->card_list[57]), 1, 3, 1, -1, -1);
    initialise_card(&(task->card_list[58]), 3, 1, 3, -1, -1);
    initialise_card(&(task->card_list[59]), 2, 2, 4, -1, -1);
    initialise_card(&(task->card_list[60]), 4, 4, 2, -1, -1);
    initialise_card(&(task->card_list[61]), 3, 3, 3, -1, -1);
    initialise_card(&(task->card_list[62]), 2, 1, 1, -1, -1);
    initialise_card(&(task->card_list[63]), 1, 2, 3, -1, -1);
#if NUM_CARDS == 128
    initialise_card(&(task->card_list[64]), 1, 2, 1, -1, -1);
    initialise_card(&(task->card_list[65]), 4, 1, 3, -1, -1);
    initialise_card(&(task->card_list[66]), 2, 4, 1, -1, -1);
    initialise_card(&(task->card_list[67]), 1, 1, 4, -1, -1);
    initialise_card(&(task->card_list[68]), 4, 2, 2, -1, -1);
    initialise_card(&(task->card_list[69]), 1, 3, 3, -1, -1);
    initialise_card(&(task->card_list[70]), 4, 4, 1, -1, -1);
    initialise_card(&(task->card_list[71]), 3, 1, 4, -1, -1);
    initialise_card(&(task->card_list[72]), 4, 2, 3, -1, -1);
    initialise_card(&(task->card_list[73]), 2, 3, 4, -1, -1);
    initialise_card(&(task->card_list[74]), 1, 4, 2, -1, -1);
    initialise_card(&(task->card_list[75]), 3, 1, 1, -1, -1);
    initialise_card(&(task->card_list[76]), 2, 4, 3, -1, -1);
    initialise_card(&(task->card_list[77]), 1, 3, 2, -1, -1);
    initialise_card(&(task->card_list[78]), 3, 2, 1, -1, -1);
    initialise_card(&(task->card_list[79]), 4, 4, 4, -1, -1);
    initialise_card(&(task->card_list[80]), 2, 1, 2, -1, -1);
    initialise_card(&(task->card_list[81]), 3, 3, 4, -1, -1);
    initialise_card(&(task->card_list[82]), 4, 1, 1, -1, -1);
    initialise_card(&(task->card_list[83]), 1, 3, 4, -1, -1);
    initialise_card(&(task->card_list[84]), 2, 4, 2, -1, -1);
    initialise_card(&(task->card_list[85]), 3, 2, 3, -1, -1);
    initialise_card(&(task->card_list[86]), 2, 3, 2, -1, -1);
    initialise_card(&(task->card_list[87]), 3, 4, 1, -1, -1);
    initialise_card(&(task->card_list[88]), 4, 1, 2, -1, -1);
    initialise_card(&(task->card_list[89]), 2, 3, 1, -1, -1);
    initialise_card(&(task->card_list[90]), 3, 4, 3, -1, -1);
    initialise_card(&(task->card_list[91]), 4, 1, 4, -1, -1);
    initialise_card(&(task->card_list[92]), 2, 2, 2, -1, -1);
    initialise_card(&(task->card_list[93]), 1, 1, 3, -1, -1);
    initialise_card(&(task->card_list[94]), 4, 2, 4, -1, -1);
    initialise_card(&(task->card_list[95]), 1, 1, 2, -1, -1);
    initialise_card(&(task->card_list[96]), 3, 4, 4, -1, -1);
    initialise_card(&(task->card_list[97]), 4, 3, 3, -1, -1);
    initialise_card(&(task->card_list[98]), 1, 2, 2, -1, -1);
    initialise_card(&(task->card_list[99]), 4, 3, 1, -1, -1);
    initialise_card(&(task->card_list[100]), 2, 4, 4, -1, -1);
    initialise_card(&(task->card_list[101]), 3, 3, 2, -1, -1);
    initialise_card(&(task->card_list[102]), 1, 2, 4, -1, -1);
    initialise_card(&(task->card_list[103]), 4, 4, 3, -1, -1);
    initialise_card(&(task->card_list[104]), 1, 1, 1, -1, -1);
    initialise_card(&(task->card_list[105]), 3, 4, 2, -1, -1);
    initialise_card(&(task->card_list[106]), 4, 3, 4, -1, -1);
    initialise_card(&(task->card_list[107]), 2, 2, 1, -1, -1);
    initialise_card(&(task->card_list[108]), 3, 1, 2, -1, -1);
    initialise_card(&(task->card_list[109]), 2, 3, 3, -1, -1);
    initialise_card(&(task->card_list[110]), 1, 4, 4, -1, -1);
    initialise_card(&(task->card_list[111]), 3, 3, 1, -1, -1);
    initialise_card(&(task->card_list[112]), 2, 2, 3, -1, -1);
    initialise_card(&(task->card_list[113]), 4, 3, 2, -1, -1);
    initialise_card(&(task->card_list[114]), 1, 4, 1, -1, -1);
    initialise_card(&(task->card_list[115]), 3, 2, 2, -1, -1);
    initialise_card(&(task->card_list[116]), 2, 1, 4, -1, -1);
    initialise_card(&(task->card_list[117]), 4, 2, 1, -1, -1);
    initialise_card(&(task->card_list[118]), 1, 4, 3, -1, -1);
    initialise_card(&(task->card_list[119]), 3, 2, 4, -1, -1);
    initialise_card(&(task->card_list[120]), 2, 1, 3, -1, -1);
    initialise_card(&(task->card_list[121]), 1, 3, 1, -1, -1);
    initialise_card(&(task->card_list[122]), 3, 1, 3, -1, -1);
    initialise_card(&(task->card_list[123]), 2, 2, 4, -1, -1);
    initialise_card(&(task->card_list[124]), 4, 4, 2, -1, -1);
    initialise_card(&(task->card_list[125]), 3, 3, 3, -1, -1);
    initialise_card(&(task->card_list[126]), 2, 1, 1, -1, -1);
    initialise_card(&(task->card_list[127]), 1, 2, 3, -1, -1);
#endif

    for (i = 0; i < NUM_CARDS; i++) {
        task->features_matched[i] = 0;
    }
}

/*----------------------------------------------------------------------------*/

static void initialise_trial(WcstTaskData *task, int trial)
{
    task->chosen_deck = 0;
    task->feedback = NULL;
    task->features_matched[trial] = 0;
    initialise_card(&(task->current_card), 0, 0, 0, -1, -1);
    task->event_count = 0;
    task->trial_complete = FALSE;
}

/******************************************************************************/

void wcst_initialise_task(WcstTaskData *task)
{
    initialise_card_lists(task);
    initialise_trial(task, 0);
    task->trial_count = 0;
    task->criterion = SORT_COLOUR;
    task->correct_consecutive_sorts = 0;

    /* Initialise scoring: */
    task->score_total_correct_sorts = 0;
    task->score_categories_obtained = 0;
    task->score_errors_perseverations = 0;
    task->score_errors_ppr = 0;
    task->score_errors_ppc = 0;
    task->score_errors_set_loss = 0;
    task->score_errors_other = 0;
}

/******************************************************************************/

void wcst_initialise_next_trial(ContentionScheduler *cs, WcstTaskData *task)
{
    task->trial_count++;
    initialise_trial(task, task->trial_count);
}

/******************************************************************************/
/* Section 3: Detect trial/task completion -----------------------------------*/

Boolean wcst_trial_complete(WcstTaskData *task)
{
    return(task->trial_complete);
}

/*----------------------------------------------------------------------------*/

Boolean wcst_task_complete(WcstTaskData *task)
{
//    return(wcst_trial_complete(task) && ((task->trial_count >= (NUM_CARDS-1)) || (task->score_categories_obtained == 6)));
    return(wcst_trial_complete(task) && (task->trial_count >= (NUM_CARDS-1)));
}

/******************************************************************************/
/* Section 4: Score the current trial ----------------------------------------*/

static Criterion previous_criterion(Criterion criterion)
{
    if (criterion == SORT_COLOUR) {
        return(SORT_NUMBER);
    }
    else if (criterion == SORT_FORM) {
        return(SORT_COLOUR);
    }
    else {
        return(SORT_FORM);
    }
}

/*----------------------------------------------------------------------------*/

static Boolean match_includes_criterion(int match, Criterion criterion)
{
    switch (criterion) {
        case SORT_COLOUR: {
            return(match & 1);
        }
        case SORT_FORM: {
            return(match & 2);
        }
        case SORT_NUMBER: {
            return(match & 4);
        }
        default: {
            return(FALSE);
        }
    }
}

/*----------------------------------------------------------------------------*/

static Boolean ppr_error(WcstTaskData *task)
{
    if ((task->trial_count < 1) || task->correct[task->trial_count-1]) {
        return(FALSE);
    }
    else if (task->features_matched[task->trial_count] == task->features_matched[task->trial_count-1]) {
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static Boolean ppc_error(WcstTaskData *task)
{
    if (task->score_categories_obtained == 0) {
        return(FALSE);
    }
    else if (match_includes_criterion(task->features_matched[task->trial_count], previous_criterion(task->criterion))) {
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static Criterion unique_matched_criterion(WcstTaskData *task, int i)
{
    /* Was the card sorted on trial i unambiguously correct? */
    if (!task->correct[i]) {
        return(SORT_NULL);
    }
    else if (task->features_matched[i] == 1) {
        /* Match colour and nothing else */
        return(SORT_COLOUR);
    }
    else if (task->features_matched[i] == 2) {
        /* Match form and nothing else */
        return(SORT_FORM);
    }
    else if (task->features_matched[i] == 4) {
        /* Match number and nothing else */
        return(SORT_NUMBER);
    }
    else {
        return(SORT_NULL);
    }
}

static Boolean unambiguous_sort(WcstTaskData *task, int i)
{
    return(unique_matched_criterion(task, i) != SORT_NULL);
}

static void wcst_score_trial(WcstTaskData *task)
{
    if (task->correct[task->trial_count]) {
        task->score_total_correct_sorts++;
        task->correct_consecutive_sorts++;
    }
    else {
        Boolean accounted = FALSE;

        if ((task->correct_consecutive_sorts > 2) && (unambiguous_sort(task, task->trial_count) || unambiguous_sort(task, task->trial_count - 1) || unambiguous_sort(task, task->trial_count - 2))) {
            task->score_errors_set_loss++;
            accounted = TRUE;
        }
        if (ppc_error(task)) {
            task->score_errors_ppc++;
            accounted = TRUE;
        }
        if (ppr_error(task)) {
            task->score_errors_ppr++;
            accounted = TRUE;
        }
        if (!accounted) {
            task->score_errors_other++;
        }

        task->correct_consecutive_sorts = 0;
    }

    /* Change criterion if appropriate: */
    if (task->correct_consecutive_sorts >= CRITERION) {
        task->score_categories_obtained++;
        task->correct_consecutive_sorts = 0;
        if (task->criterion == SORT_COLOUR) {
            task->criterion = SORT_FORM;
        }
        else if (task->criterion == SORT_FORM) {
            task->criterion = SORT_NUMBER;
        }
        else {
            task->criterion = SORT_COLOUR;
        }
    }

    task->trial_complete = TRUE;
}

/******************************************************************************/
/* Section 5: WCST events ----------------------------------------------------*/

static void wcst_present_card(WcstTaskData *task)
{
    task->current_card.number = task->card_list[task->trial_count].number;
    task->current_card.colour = task->card_list[task->trial_count].colour;
    task->current_card.form = task->card_list[task->trial_count].form;
    task->current_card.position_x = 3;
    task->current_card.position_y = 1;
    task->chosen_deck = 0;
}

/******************************************************************************/

static void wcst_generate_feedback(WcstTaskData *task)
{
    task->features_matched[task->trial_count] = 0;
    if (task->card_sorted[task->chosen_deck-1].colour == task->card_target[task->chosen_deck-1].colour) {
        task->features_matched[task->trial_count] += 1;
    }
    if (task->card_sorted[task->chosen_deck-1].form == task->card_target[task->chosen_deck-1].form) {
        task->features_matched[task->trial_count] += 2;
    }
    if (task->card_sorted[task->chosen_deck-1].number == task->card_target[task->chosen_deck-1].number) {
        task->features_matched[task->trial_count] += 4;
    }

    if (match_includes_criterion(task->features_matched[task->trial_count], task->criterion)) {
        task->correct[task->trial_count] = TRUE;
        task->feedback = "Correct";
    }
    else {
        task->correct[task->trial_count] = FALSE;
        task->feedback = "Wrong";
    }
}

/******************************************************************************/

#ifdef WM_DISCRETE

void wcst_run_perceptual_module(ContentionScheduler *cs, WcstTaskData *task)
{
    Chunk *fb, *criterion;

    /* FIX ME: Decay existing working memory elements:                               */

    /* FIX ME: Add to working memory all elements supported by the external display: */

    /* Now handle feedback: */
    if (task->feedback != NULL) {

        if ((fb = wm_query(cs, chunk_build("Feedback", NULL))) != NULL) { wm_delete(cs, fb); }

        wm_add(cs, chunk_build("Feedback", task->feedback));
    }
    if (wm_match(cs, chunk_build("Feedback", "Wrong")) && ((criterion = wm_query(cs, chunk_build("Strategy", NULL))) != NULL)) { 
        wm_add(cs, chunk_build("Reject", criterion->value));
    }
}

char *wcst_generate_strategy(ContentionScheduler *cs, WcstTaskData *task)
{
    GList *pool = NULL;
    char *s;

    /* Put all plausible sorting criteria into a pool: */

    if (wm_match(cs, chunk_build("Feedback", "Correct"))) {
        if (!wm_match(cs, chunk_build("Reject", "Colour"))) {
            pool = g_list_append(pool, "schemaMatchColour");
        }
        if (!wm_match(cs, chunk_build("Reject", "Number"))) {
            pool = g_list_append(pool, "schemaMatchNumber");
        }
        if (!wm_match(cs, chunk_build("Reject", "Form"))) {
            pool = g_list_append(pool, "schemaMatchForm");
        }
    }
    else if (wm_match(cs, chunk_build("Feedback", "Wrong"))) {
        if (!wm_match(cs, chunk_build("Reject", "Colour"))) {
            pool = g_list_append(pool, "schemaMatchColour");
        }
        if (!wm_match(cs, chunk_build("Reject", "Number"))) {
            pool = g_list_append(pool, "schemaMatchNumber");
        }
        if (!wm_match(cs, chunk_build("Reject", "Form"))) {
            pool = g_list_append(pool, "schemaMatchForm");
        }
    }

fprintf(stdout, "%d strategies possible\n", g_list_length(pool));

    /* Select at random from the pool: */
    if (pool == NULL) {
        s = "schemaRandomPlacement";
    }
    else if (g_list_length(pool) == 1) {
        s = (char *)g_list_nth_data(pool, 0);
    }
    else {
        int p = (int) floor(g_list_length(pool) * (random() / pow(2, 31)));
        s = (char *)g_list_nth_data(pool, p);
    }

    g_list_free(pool);
    return(s);
}

static void wcst_sort_card(ContentionScheduler *cs, WcstTaskData *task)
{
    char *s;
    Chunk *sc;

    wcst_run_perceptual_module(cs, task);

    if ((s = temporary_schema_get(cs)) == NULL) {
        s = wcst_generate_strategy(cs, task);
        if ((sc = wm_query(cs, chunk_build("Strategy", NULL))) != NULL) { wm_delete(cs, sc); }
        wm_add(cs, chunk_build("Strategy", s));
        temporary_schema_set(cs, s);
    }
    cs_schema_set_intended(cs, s, TRUE);
    cs_run_task(cs, task, 50);
}

#endif
#ifdef WM_CONTINUOUS

static void wcst_sort_card(ContentionScheduler *cs, WcstTaskData *task)
{
    Wme *w;

    // FIX ME: Simple card sorting: assume that schemas exist and
    // are in working memory, then just select the most active

    if ((w = wm_select_most_active_element(cs)) != NULL) {
        cs_schema_set_intended(cs, w->schema_name, TRUE);
    }
    cs_run_task(cs, task, 50);
}

#endif

/******************************************************************************/

#ifdef WM_DISCRETE

static Boolean state_supports_strategy(WcstTaskData *task, char *criterion)
{
    if (strcmp(criterion, "Colour") == 0) {
        return(task->card_sorted[task->chosen_deck-1].colour == task->card_target[task->chosen_deck-1].colour);
    }
    else if (strcmp(criterion, "Number") == 0) {
       return(task->card_sorted[task->chosen_deck-1].number == task->card_target[task->chosen_deck-1].number);
    }
    else if (strcmp(criterion, "Form") == 0) {
        return(task->card_sorted[task->chosen_deck-1].form == task->card_target[task->chosen_deck-1].form);
    }
    else {
        return(FALSE);
    }
}

static void wcst_assimilate_feedback(ContentionScheduler *cs, WcstTaskData *task)
{
    /* This is basically the monitoring process. It should be possible      */
    /* to impair this, perhaps so that monitoring fails probabilistically.  */

    wcst_run_perceptual_module(cs, task);

    if (wm_match(cs, chunk_build("Feedback", "Wrong"))) {
        temporary_schema_reject_current(cs);

        if (state_supports_strategy(task, "Colour")) {
            wm_add(cs, chunk_build("Reject", "Colour"));
        }
        if (state_supports_strategy(task, "Number")) {
            wm_add(cs, chunk_build("Reject", "Number"));
        }
        if (state_supports_strategy(task, "Form")) {
            wm_add(cs, chunk_build("Reject", "Form"));
        }
    }
    else {
        /* Assume we've got it right: */
        if (wm_match(cs, chunk_build("Reject", "Colour"))) {
            wm_delete(cs, chunk_build("Reject", "Colour"));
        }
        if (wm_match(cs, chunk_build("Reject", "Number"))) {
            wm_delete(cs, chunk_build("Reject", "Number"));
        }
        if (wm_match(cs, chunk_build("Reject", "Form"))) {
            wm_delete(cs, chunk_build("Reject", "Form"));
        }
    }
}

#endif
#ifdef WM_CONTINUOUS

static Boolean state_supports_strategy(WcstTaskData *task, Wme *w)
{
    if (strcmp(w->schema_name, "schemaMatchColour") == 0) {
        return(task->card_sorted[task->chosen_deck-1].colour == task->card_target[task->chosen_deck-1].colour);
    }
    else if (strcmp(w->schema_name, "schemaMatchNumber") == 0) {
       return(task->card_sorted[task->chosen_deck-1].number == task->card_target[task->chosen_deck-1].number);
    }
    else if (strcmp(w->schema_name, "schemaMatchForm") == 0) {
        return(task->card_sorted[task->chosen_deck-1].form == task->card_target[task->chosen_deck-1].form);
    }
    else {
        return(FALSE);
    }
}

static void wcst_assimilate_feedback(ContentionScheduler *cs, WcstTaskData *task)
{
    WmeList *wl;

    // FIX ME: We assume schemas exist and are in working memory.
    // Assimilating feedback is then a matter of exciting those
    // schemas that are consistent with the positive feedback or
    // inhibiting those that are consistent with negative feedback

    if (task->feedback == NULL) {
        fprintf(stdout, "Error: No feedback!\n");
    }
    else if (strcmp(task->feedback, "Correct") == 0) {
        /* Feedback is positive: excite all WM nodes consistent with  */
        /* the existing world. */

        for (wl = cs->wmes; wl != NULL; wl = wl->tail) {
            if (state_supports_strategy(task, wl->head)) {
                wm_excite_node(cs, wl->head, parameter.wm_excite);
            }
        }
    }
    else if (strcmp(task->feedback, "Wrong") == 0) {
        /* Feedback is negative: inhibit all WM nodes consistent with */
        /* the existing world. */

        for (wl = cs->wmes; wl != NULL; wl = wl->tail) {
            if (state_supports_strategy(task, wl->head)) {
                wm_excite_node(cs, wl->head, -parameter.wm_inhibit);
            }
        }
    }
}

#endif

/******************************************************************************/

static Boolean unambiguous(int matches, Criterion *criterion)
{
    if (matches == 1) {
        *criterion = SORT_COLOUR;
        return(TRUE);
    }
    else if (matches == 2) {
        *criterion = SORT_FORM;
        return(TRUE);
    }
    else if (matches == 4) {
        *criterion = SORT_NUMBER;
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/*----------------------------------------------------------------------------*/

static Boolean extract_established_criterion(WcstTaskData *task, Criterion *criterion)
{
    int i;

    for (i = 0; i < NUM_CARDS; i++) {
        if (!task->correct[i] && unambiguous(task->features_matched[i], criterion)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static void wcst_rescore_ppc_errors(WcstTaskData *task)
{
    Criterion criterion;
    int i;

    task->score_errors_ppc = 0;
    if (extract_established_criterion(task, &criterion)) {
        for (i = 0; i < NUM_CARDS; i++) { 
            if (!task->correct[i] && match_includes_criterion(task->features_matched[i], criterion)) {
                task->score_errors_ppc++;
            }
        }
    }
}

/*----------------------------------------------------------------------------*/

static int perseverative_criterion_changes(WcstTaskData *task, int i)
{
    /* We've just got one wrong, and it was unambiguous, so we          */
    /* should know not to use this criterion. If, however, it is the    */
    /* start of a continuous string of responses to this criterion,     */
    /* including two further unambiguous responses, then it would seem  */
    /* that the subject has changed his/her criterion, so return TRUE.  */

    int unambiguous_responses = 1;
    Criterion new_criterion = unique_matched_criterion(task, i);

    while (++i < NUM_CARDS) {
        if (!match_includes_criterion(task->features_matched[i], new_criterion)) {
            return(0);
        }
        else if (unambiguous_sort(task, i)) {
            unambiguous_responses++;
        }
        if (unambiguous_responses > 2) {
            return(1);
        }
    }
    return(0);
}

static void wcst_score_classical_perseverations(WcstTaskData *task)
{
    int i, seq = 0;
    int successive_unambiguous_responses = 0;
    Criterion previous_criterion = SORT_NULL;

    for (i = 0; i < NUM_CARDS; i++) {
        if (task->correct[i]) {
            task->perseverative_error[i] = FALSE;
            seq++;
            if (unambiguous_sort(task, i)) {
                successive_unambiguous_responses++;
            }
            if (seq == 10) {
                if (previous_criterion == SORT_NULL) {

                    previous_criterion = SORT_COLOUR;
		}
		else if (previous_criterion == SORT_COLOUR) {
                    previous_criterion = SORT_FORM;
		}
		else if (previous_criterion == SORT_FORM) {
                    previous_criterion = SORT_NUMBER;
		}
		else if (previous_criterion == SORT_NUMBER) {
                    previous_criterion = SORT_COLOUR;
		}
                seq = 0;
            }
        }
        else {
            if (match_includes_criterion(task->features_matched[i], previous_criterion)) {
                if (unambiguous_sort(task, i)) {
                    successive_unambiguous_responses++;
                }
                if (successive_unambiguous_responses > 1) {
                    task->perseverative_error[i] = TRUE;
                    task->score_errors_perseverations++;
		}
		else {
                    task->perseverative_error[i] = FALSE;
		}
            }
            else if (unambiguous_sort(task, i)) {
                task->perseverative_error[i] = FALSE;
                if (previous_criterion == 0) {
                    previous_criterion = unique_matched_criterion(task, i);
                    successive_unambiguous_responses = 1;
                }
                else if (perseverative_criterion_changes(task, i)) {
                    previous_criterion = unique_matched_criterion(task, i);
                    successive_unambiguous_responses = 1;
                }
            }
            else {
                task->perseverative_error[i] = FALSE;
            }
            seq = 0;
	}
    }
}

/******************************************************************************/

void wcst_step_event(ContentionScheduler *cs, WcstTaskData *task)
{
    task->event_count++;
    if (task->event_count == 1) {
        wcst_present_card(task);
    }
    else if (task->event_count == 2) {
        wcst_sort_card(cs, task);
    }
    else if (task->event_count == 3) {
        wcst_generate_feedback(task);
    }
    else if (task->event_count == 4) {
        wcst_assimilate_feedback(cs, task);
    }
    else {
        wcst_score_trial(task);

        /* If the entire task is completed without obtaining any categories */
        /* then rescore PPC errors:                                         */

        if (wcst_task_complete(task)) {
            wcst_score_classical_perseverations(task);
            if (task->score_categories_obtained == 0) {
                wcst_rescore_ppc_errors(task);
            }
        }
    }
}

/******************************************************************************/
/* Section 6: Synchronise CS and external world representations --------------*/

static Object *set_card_features(ContentionScheduler *cs, char *name, Card *card)
{
    Object *this = objectGetByName(cs, name);

    if (this != NULL) {
        this->features = 0;

        objectSetPosition(this, card->position_x, card->position_y);

        if (card->number == 1) {
            this->features = this->features | IS_NUMBER_ONE;
        }
        else if (card->number == 2) {
            this->features = this->features | IS_NUMBER_TWO;
        }
        else if (card->number == 3) {
            this->features = this->features | IS_NUMBER_THREE;
        }
        else if (card->number == 4) {
            this->features = this->features | IS_NUMBER_FOUR;
        }

        if (card->colour == 1) {
            this->features = this->features | IS_COLOUR_RED;
        }
        else if (card->colour == 2) {
            this->features = this->features | IS_COLOUR_GREEN;
        }
        else if (card->colour == 3) {
            this->features = this->features | IS_COLOUR_YELLOW;
        }
        else if (card->colour == 4) {
            this->features = this->features | IS_COLOUR_BLUE;
        }

        if (card->form == 1) {
            this->features = this->features | IS_FORM_TRIANGLE;
        }
        else if (card->form == 2) {
            this->features = this->features | IS_FORM_STAR;
        }
        else if (card->form == 3) {
            this->features = this->features | IS_FORM_CROSS;
        }
        else if (card->form == 4) {
            this->features = this->features | IS_FORM_CIRCLE;
        }
    }
    return(this);
}

/*----------------------------------------------------------------------------*/

static Object *set_feedback(ContentionScheduler *cs, char *feedback)
{
    Object *this = objectGetByName(cs, "feedback");

    if (this != NULL) {
        if (feedback == NULL) {
            this->features = 0;
	}
        else if (strcmp(feedback, "Correct") == 0) {
            this->features = FEEDBACK_CORRECT;
        }
        else if (strcmp(feedback, "Wrong") == 0) {
            this->features = FEEDBACK_WRONG;
        }
        else {
            this->features = 0;
        }
    }
    return(this);
}

/*----------------------------------------------------------------------------*/

void wcst_synchronise_world(ContentionScheduler *cs, WcstTaskData *task)
{
    set_card_features(cs, "target1", &(task->card_target[0]));
    set_card_features(cs, "target2", &(task->card_target[1]));
    set_card_features(cs, "target3", &(task->card_target[2]));
    set_card_features(cs, "target4", &(task->card_target[3]));

    set_card_features(cs, "sorted1", &(task->card_sorted[0]));
    set_card_features(cs, "sorted2", &(task->card_sorted[1]));
    set_card_features(cs, "sorted3", &(task->card_sorted[2]));
    set_card_features(cs, "sorted4", &(task->card_sorted[3]));

    set_card_features(cs, "current", &(task->current_card));
    
    set_feedback(cs, task->feedback);
}

/******************************************************************************/
/* Section 7: Dump the analysis results: -------------------------------------*/

void wcst_dump_analyses(FILE *fp, WcstTaskData *task)
{
    int cards_sorted = wcst_trial_complete(task) ? task->trial_count+1 :  task->trial_count;

    fprintf(fp, "\n");
    fprintf(fp, "Cards Sorted:            %4d\n", cards_sorted);
    fprintf(fp, "Correct Sorts:           %4d\n", task->score_total_correct_sorts);
    fprintf(fp, "Categories Obtained:     %4d\n", task->score_categories_obtained);
    fprintf(fp, "Errors (Perseverations): %4d\n", task->score_errors_perseverations);
    fprintf(fp, "Errors (PPR):            %4d\n", task->score_errors_ppr);
    fprintf(fp, "Errors (PPC):            %4d\n", task->score_errors_ppc);
    fprintf(fp, "Errors (Set Loss):       %4d\n", task->score_errors_set_loss);
    fprintf(fp, "Errors (Other):          %4d\n", task->score_errors_other);
}

/******************************************************************************/
/* Section 8: Actions that alter the world: ----------------------------------*/

void wcst_action_pick_up(WcstTaskData *task, Object *theme, Effector *effector)
{
    objectSetPosition(theme, 0.0, 0.0);
}

void wcst_action_put_down(WcstTaskData *task, Object *theme, Object *target)
{
    if (strcmp(target->name, "target1") == 0) {
        task->chosen_deck = 1;
    }
    else if (strcmp(target->name, "target2") == 0) {
        task->chosen_deck = 2;
    }
    else if (strcmp(target->name, "target3") == 0) {
        task->chosen_deck = 3;
    }
    else if (strcmp(target->name, "target4") == 0) {
        task->chosen_deck = 4;
    }
    else {
        task->chosen_deck = 0;
    }

    if (task->chosen_deck > 0) {
        task->card_sorted[task->chosen_deck-1].number = task->current_card.number;
        task->current_card.number = 0; 
        task->card_sorted[task->chosen_deck-1].colour = task->current_card.colour;
        task->current_card.colour = 0;
        task->card_sorted[task->chosen_deck-1].form = task->current_card.form;
        task->current_card.form = 0;

        objectSetPosition(theme, 2.0, task->chosen_deck);
    }
}

/******************************************************************************/
