/*******************************************************************************

    File:     wcst_schema.c
    Author:   Rick Cooper
    Created:  30/03/07
    Edited:   03/04/07
    Contents: Create/maintain schemas for the WCST task

    Public procedures:
        void install_wcst_schemas(ContentionScheduler *cs)
        void wcst_initialise_wm(ContentionScheduler *cs)

*******************************************************************************/

#include "ss.h"
#include "wcst.h"

/******************************************************************************/
/* Section 1: WCST-specific object test/manipulation functions ---------------*/

static int get_value_of_card_feature(Object *card, Criterion feature)
{
    if (feature == SORT_COLOUR) {
        if (card->features & IS_COLOUR_RED) {
            return(IS_COLOUR_RED);
        }
        else if (card->features & IS_COLOUR_GREEN) {
            return(IS_COLOUR_GREEN);
        }
        else if (card->features & IS_COLOUR_YELLOW) {
            return(IS_COLOUR_YELLOW);
        }
        else if (card->features & IS_COLOUR_BLUE) {
            return(IS_COLOUR_BLUE);
        }
    }
    else if (feature == SORT_FORM) {
        if (card->features & IS_FORM_TRIANGLE) {
            return(IS_FORM_TRIANGLE);
        }
        else if (card->features & IS_FORM_STAR) {
            return(IS_FORM_STAR);
        }
        else if (card->features & IS_FORM_CROSS) {
            return(IS_FORM_CROSS);
        }
        else if (card->features & IS_FORM_CIRCLE) {
            return(IS_FORM_CIRCLE);
        }
    }
    else if (feature == SORT_NUMBER) {
        if (card->features & IS_NUMBER_ONE) {
            return(IS_NUMBER_ONE);
        }
        else if (card->features & IS_NUMBER_TWO) {
            return(IS_NUMBER_TWO);
        }
        else if (card->features & IS_NUMBER_THREE) {
            return(IS_NUMBER_THREE);
        }
        else if (card->features & IS_NUMBER_FOUR) {
            return(IS_NUMBER_FOUR);
        }
    }
    return(0);
}

static Boolean cards_match_on_feature(Object *c1, Object *c2, Criterion feature)
{
    return(get_value_of_card_feature(c1, feature) == get_value_of_card_feature(c2, feature));
}

static Boolean wcst_held_card_matches_feature(ContentionScheduler *cs, Object *o, Criterion feature)
{
    ObjectList *os;

    for (os = cs->objects; os != NULL; os = os->tail) {
        if (isHeldBoolean(os->head) && cards_match_on_feature(o, os->head, feature)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static Boolean wcst_is_card(Object *this)
{
    return(strcmp(this->name, "feedback") != 0);
}

static Boolean wcst_holds_card(ContentionScheduler *cs)
{
    ObjectList *os;

    for (os = cs->objects; os != NULL; os = os->tail) {
        if (isHeldBoolean(os->head) && wcst_is_card(os->head)) {
            return(TRUE);
	}
    }
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

static GoalType wcst_goal_card_sorted(ContentionScheduler *cs)
{
    Object *this = objectGetByName(cs, "current");

    if ((this != NULL) && (this->features > 0) && isOnTable(this) && (this->position[0] < 3.0)) {
        return(GOAL_TRUE);
    }
    else {
        return(GOAL_FALSE);
    }
}

/******************************************************************************/
/* Section 2: Schema triggering functions ------------------------------------*/

static double schemaMatchCardTriggersTheme(ContentionScheduler *cs, Object *o)
{
/*  An object is triggered as a theme for matching if it is a           */
/*  "to-be-sorted" card - that is, if it's name is "current" and        */
/*  it has non-zero features. (Once the current card has been sorted    */
/*  its features are reset to zero.)                                    */

    if ((strcmp(o->name, "current") == 0) && (o->features > 0)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaMatchColourTriggersTarget(ContentionScheduler *cs, Object *o)
{
/*  An object is triggered as a target for matching to colour if it is  */
/*  a "target" card and there is a held card which matches the object   */
/*  on the colour dimension.                                            */
    
    if (strncmp(o->name, "target", 6) != 0) {
        return(0.0);
    }
    else if (wcst_held_card_matches_feature(cs, o, SORT_COLOUR)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaMatchFormTriggersTarget(ContentionScheduler *cs, Object *o)
{
/*  An object is triggered as a target for matching to form if it is    */
/*  a "target" card and there is a held card which matches the object   */
/*  on the form dimension.                                              */

    if (strncmp(o->name, "target", 6) != 0) {
        return(0.0);
    }
    else if (wcst_held_card_matches_feature(cs, o, SORT_FORM)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaMatchNumberTriggersTarget(ContentionScheduler *cs, Object *o)
{
/*  An object is triggered as a target for matching to number if it is  */
/*  a "target" card and there is a held card which matches the object   */
/*  on the number dimension.                                            */

    if (strncmp(o->name, "target", 6) != 0) {
        return(0.0);
    }
    else if (wcst_held_card_matches_feature(cs, o, SORT_NUMBER)) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaPickUpCardTriggersTheme(ContentionScheduler *cs, Object *o)
{
/*  An object is triggered as a theme for picking up if it is a card    */
/*  of any kind and it is not already held.                             */

    if (isHeldBoolean(o)) {
        return(0.0);
    }
    else if ((strcmp(o->name, "current") == 0) && (o->features > 0)) {
        return(1.0);
    }
    else if (strncmp(o->name, "target", 6) == 0) {
        return(1.0);
    }
    else if (strncmp(o->name, "sorted", 6) == 0) {
        return(1.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaPlaceCardTriggersTarget(ContentionScheduler *cs, Object *o)
{
/*  Objects are triggered as targets for the placement of a card if      */
/*  they are target cards (name is "target1" ... "target4"). If they     */
/*  also match the held card on one or more features they get an exta    */
/*  boost, proportional to the number of matching features.              */

    if (!wcst_holds_card(cs)) {
        return(0.0);
    }
    else if (strncmp(o->name, "target", 6) == 0) {
        int matches = 1;

        if (wcst_held_card_matches_feature(cs, o, SORT_COLOUR)) {
            matches++;
        }
        if (wcst_held_card_matches_feature(cs, o, SORT_NUMBER)) {
            matches++;
        }
        if (wcst_held_card_matches_feature(cs, o, SORT_FORM)) {
            matches++;
        }
        return(matches / 4.0);
    }
    else {
        return(0.0);
    }
}

/*----------------------------------------------------------------------------*/

static double schemaMatchCardTriggersEffector(ContentionScheduler *cs, Effector *e)
{
    return(0.0);
}

/*----------------------------------------------------------------------------*/

static double schemaPickUpTriggersEffector(ContentionScheduler *cs, Effector *e)
{
/*  Effectors are triggered for pick-up actions if they hold nothing. */

    return(isEmptyEffectorFuzzy(e));
}

/*----------------------------------------------------------------------------*/

static double schemaPlaceCardTriggersEffector(ContentionScheduler *cs, Effector *e)
{
/*  Effectors are triggered for put-down actions if they hold something. */
    
    return(isFullEffectorFuzzy(e));
}

/******************************************************************************/
/* Section 3: Install WCST schemas -------------------------------------------*/

void install_wcst_schemas(ContentionScheduler *cs)
{
    Schema *s;

    s = install_schema(cs, "schemaMatchColour", "goalWCST", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMatchCardTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMatchColourTriggersTarget);
    install_schema_effector_trigger(s, schemaMatchCardTriggersEffector);
    install_schema_subgoal(s, "goalPickUpCard", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPlaceCard", goalHandsFull, wcst_goal_card_sorted);

    s = install_schema(cs, "schemaMatchForm", "goalWCST", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMatchCardTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMatchFormTriggersTarget);
    install_schema_effector_trigger(s, schemaMatchCardTriggersEffector);
    install_schema_subgoal(s, "goalPickUpCard", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPlaceCard", goalHandsFull, wcst_goal_card_sorted);

    s = install_schema(cs, "schemaMatchNumber", "goalWCST", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMatchCardTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaMatchNumberTriggersTarget);
    install_schema_effector_trigger(s, schemaMatchCardTriggersEffector);
    install_schema_subgoal(s, "goalPickUpCard", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPlaceCard", goalHandsFull, wcst_goal_card_sorted);

    s = install_schema(cs, "schemaRandomPlacement", "goalWCST", ACTION_NONE, MEDIUM);
    install_schema_object_trigger(s, THEME, schemaMatchCardTriggersTheme);
    install_schema_object_trigger(s, TARGET, schemaPlaceCardTriggersTarget);
    install_schema_effector_trigger(s, schemaMatchCardTriggersEffector);
    install_schema_subgoal(s, "goalPickUpCard", goalHandsEmpty, goalHandsFull);
    install_schema_subgoal(s, "goalPlaceCard", goalHandsFull, wcst_goal_card_sorted);

    s = install_schema(cs, "schemaPickUpCard", "goalPickUpCard", ACTION_PICK_UP_THEME, LOW);
    install_schema_object_trigger(s, THEME, schemaPickUpCardTriggersTheme);
    install_schema_effector_trigger(s, schemaPickUpTriggersEffector);

    s = install_schema(cs, "schemaPlaceCard", "goalPlaceCard", ACTION_PUT_DOWN_TARGET, LOW);
    install_schema_object_trigger(s, TARGET, schemaPlaceCardTriggersTarget);
    install_schema_effector_trigger(s, schemaPlaceCardTriggersEffector);
}

/******************************************************************************/

void wcst_initialise_wm(ContentionScheduler *cs)
{
    /* Create initial WM representations of strategies (FIX ME) */
#ifdef WM_CONTINUOUS
    wm_install_node(cs, "schemaMatchColour");
    wm_install_node(cs, "schemaMatchNumber");
    wm_install_node(cs, "schemaMatchForm");
#endif
}

/******************************************************************************/
