
#include "ss.h"
#include "cs.h"
#include "chunk.h"

#define DEBUG_LEVEL 0

/*----------------------------------------------------------------------------*/

static void chunk_dump_to_filep(Chunk *content, FILE *fp)
{
    fprintf(fp, "<< ");
    while (content != NULL) {
        fprintf(fp, "<%s %s> ", content->attribute, content->value);
        content = content->next;
    }
    fprintf(fp, ">>");
}

void wm_dump_to_filep(ContentionScheduler *cs, FILE *fp)
{
    WmeList *this;

    for (this = cs->wmes; this != NULL; this = this->tail) {
        fprintf(fp, "%5.2f: ", this->head->activation[cs->cycle]);
        chunk_dump_to_filep(this->head->content, fp);
        fprintf(fp, "\n");
    }
}

static void wm_log(int cycle, int level, char *operation, Chunk *cnk)
{
    if (level < DEBUG_LEVEL) {
        fprintf(stderr, "%5d: WM: %s ", cycle, operation);
        if (cnk != NULL) {
            chunk_dump_to_filep(cnk, stderr);
        }
        fprintf(stderr, "\n");
    }
}

/*----------------------------------------------------------------------------*/

char *chunk_get_value(Chunk *chunk, char *attribute)
{
    while (chunk != NULL) {
        if (strcmp(chunk->attribute, attribute) == 0) {
            return(chunk->value);
        }
        chunk = chunk->next;
    }
    return(NULL);
}

/*----------------------------------------------------------------------------*/

static Boolean av_match(Chunk *target, char *attribute, char *value)
{
    while (target != NULL) {
        if (strcmp(target->attribute, attribute) == 0) {
            if ((value == NULL) || (strcmp(target->value, value) == 0)) {
                return(TRUE);
            }
        }
        target = target->next;
    }
    return(FALSE);
}

static Boolean chunk_match(Chunk *template, Chunk *target)
{
    while (template != NULL) {
        if (!av_match(target, template->attribute, template->value)) {
            return(FALSE);
        }
        template = template->next;
    }
    return(TRUE);
}

/*----------------------------------------------------------------------------*/

Chunk *chunk_build(char *attribute, char *value)
{
    Chunk *new = (Chunk *)malloc(sizeof(Chunk));

    if (new != NULL) {
        new->attribute = string_copy(attribute);
        new->value = (value == NULL ? NULL : string_copy(value));
        new->next = NULL;
    }
    return(new);
}

Boolean chunk_extend(Chunk *chunk, char *attribute, char *value)
{
    Chunk *new;

    if (chunk == NULL) {
        return(FALSE);
    }
    else if ((new = chunk_build(attribute, value)) == NULL) {
        return(FALSE);
    }
    else {
        while (chunk->next != NULL) {
            chunk = chunk->next;
        }
        chunk->next = new;
        return(TRUE);
    }
}

void   chunk_dispose(Chunk *chunk)
{
    while (chunk != NULL) {
        Chunk *tmp = chunk->next;

        free(chunk->attribute);
        if (chunk->value != NULL) {
            free(chunk->value);
        }
        free(chunk);
        chunk = tmp;
    }
}

/*----------------------------------------------------------------------------*/

void wm_add(ContentionScheduler *cs, Chunk *cnk)
{
    WmeList *new;

    if ((new = (WmeList *)malloc(sizeof(WmeList))) == NULL) {
        wm_log(cs->cycle, 0, "ADD: MALLOC 1 FAILED", cnk);
    }
    else if ((new->head = (Wme *)malloc(sizeof(Wme))) == NULL) {
        wm_log(cs->cycle, 0, "ADD: MALLOC 2 FAILED", cnk);
    }
    else {
        new->head->content = cnk;
        new->head->timestamp = cs->cycle;
        new->head->highlighted = FALSE;
        new->head->activation[cs->cycle] = 1.0;
        new->tail = cs->wmes;
        cs->wmes = new;
        wm_log(cs->cycle, 1, "ADD", cnk);
    }
}

/*----------------------------------------------------------------------------*/

void wm_delete(ContentionScheduler *cs, Chunk *template)
{
    WmeList *this, *prev;

    prev = NULL;
    for (this = cs->wmes; this != NULL; this = this->tail) {
        if (chunk_match(template, this->head->content)) {
            if (prev == NULL) {
                cs->wmes = this->tail;
            }
            else {
                prev->tail = this->tail;
	    }
            wm_log(cs->cycle, 1, "DELETE", this->head->content);
            chunk_dispose(this->head->content);
            free(this->head);
            free(this);

            return;
	}
        prev = this;
    }
    wm_log(cs->cycle, 0, "DELETE FAILED", template);
}

/*----------------------------------------------------------------------------*/

void wm_delete_all(ContentionScheduler *cs, Chunk *template)
{
    WmeList *this, *prev;

    prev = NULL;
    this = cs->wmes;
    while (this != NULL) {
        if (chunk_match(template, this->head->content)) {
            WmeList *next = this->tail;
            if (prev == NULL) {
                cs->wmes = next;
            }
            else {
                prev->tail = next;
	    }
            wm_log(cs->cycle, 1, "DELETE", this->head->content);
            chunk_dispose(this->head->content);
            free(this->head);
            free(this);
            this = next;
	}
	else {
	    prev = this;
            this = this->tail;
        }
    }
}

/*----------------------------------------------------------------------------*/

Boolean wm_match(ContentionScheduler *cs, Chunk *template)
{
    WmeList *this;

    for (this = cs->wmes; this != NULL; this = this->tail) {
        if (chunk_match(template, this->head->content)) {
            wm_log(cs->cycle, 2, "MATCH", this->head->content);
            return(TRUE);
        }
    }
    wm_log(cs->cycle, 2, "MATCH FAILED", template);
    return(FALSE);
}

/*----------------------------------------------------------------------------*/

Chunk *wm_query(ContentionScheduler *cs, Chunk *template)
{
    WmeList *this;
    Chunk *result = NULL;
    double result_act = 0.0;

    for (this = cs->wmes; this != NULL; this = this->tail) {
        if (chunk_match(template, this->head->content)) {
            if (this->head->activation[cs->cycle] > result_act) {
                result = this->head->content;
                result_act = this->head->activation[cs->cycle];
            }
        }
    }

    if (result != NULL) {
        wm_log(cs->cycle, 2, "QUERY", result);
    }
    else {
        wm_log(cs->cycle, 2, "QUERY FAILED", template);
    }

    return(result);
}

/*----------------------------------------------------------------------------*/

void wm_destroy(ContentionScheduler *cs)
{

    wm_log(cs->cycle, 0, "DESTROY", NULL);
    while (cs->wmes != NULL) {
        WmeList *this = cs->wmes;
        chunk_dispose(this->head->content);
        free(this->head);
        free(this);
        cs->wmes = this->tail;
    }
}

/*----------------------------------------------------------------------------*/

void wm_step_activations(ContentionScheduler *cs)
{
    WmeList *wl;

    for (wl = cs->wmes; wl != NULL; wl = wl->tail) {
        wl->head->activation[cs->cycle] = wl->head->activation[cs->cycle-1];
    }
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
