/*******************************************************************************

    File:     world.c
    Author:   Rick Cooper
    Created:  28/03/07
    Edited:   29/03/07
    Contents: Interface routines for running the SS world model

    Public procedures:
        GlobalVariables *world_create(TaskId task)
        void world_destroy(GlobalVariables *globals)
        Boolean world_initialise_task(GlobalVariables *globals)
        void world_initialise_next_trial(GlobalVariables *globals)
        void world_step_event(GlobalVariables *globals)
        Boolean world_trial_complete(GlobalVariables *globals)
        Boolean world_task_complete(GlobalVariables *globals)

*******************************************************************************/

#include "ss.h"
#include "wcst.h"
#include "tol.h"
#include "rng.h"

/******************************************************************************/
/* Section 1: Data structure creation/destruction ----------------------------*/

GlobalVariables *world_create(TaskId task)
{
    GlobalVariables *globals;

    if ((globals = (GlobalVariables *)malloc(sizeof(GlobalVariables))) != NULL) {
        globals->window = NULL;
        globals->world_da = NULL;
        globals->analysis_da = NULL;
        globals->cs_graphs[0] = NULL;
        globals->cs_graphs[1] = NULL;
        globals->cs_graphs[2] = NULL;
        globals->dynamic_canvases = FALSE;
        globals->gc = NULL;
        globals->cs = NULL;
        globals->task = task;
    }
    return(globals);
}

/*----------------------------------------------------------------------------*/

void world_destroy(GlobalVariables *globals)
{
    free(globals);
}

/******************************************************************************/
/* Section 2: Run model events -----------------------------------------------*/

Boolean world_initialise_task(GlobalVariables *globals)
{
    cs_initialise_task(globals->cs, globals->task);
    wm_destroy(globals->cs);
    switch (globals->task) {
        case TASK_WCST: {
            wcst_initialise_task((WcstTaskData *)globals->task_data);
            wcst_initialise_wm(globals->cs);
	    break;
        }
        case TASK_TOL: {
            tol_initialise_task((TolTaskData *)globals->task_data);
            tol_initialise_wm(globals->cs);
	    break;
        }
        case TASK_RNG: {
            rng_initialise_task((RngTaskData *)globals->task_data);
            rng_initialise_wm(globals->cs);
	    break;
        }
        default: {
            break;
        }
    }
    return(TRUE);
}

/*----------------------------------------------------------------------------*/

void world_initialise_next_trial(GlobalVariables *globals)
{
    switch (globals->task) {
        case TASK_WCST: {
            cs_reset_intended_schema_subgoals(globals->cs);
            wcst_initialise_next_trial(globals->cs, (WcstTaskData *)globals->task_data);
	    break;
        }
        case TASK_TOL: {
            tol_initialise_next_trial(globals->cs, (TolTaskData *)globals->task_data);
	    break;
        }
        case TASK_RNG: {
            cs_reset_intended_schema_subgoals(globals->cs);
            rng_initialise_next_trial(globals->cs, (RngTaskData *)globals->task_data);
	    break;
        }
        default: {
        }
    }
}

/*----------------------------------------------------------------------------*/

void world_step_event(GlobalVariables *globals)
{
    switch (globals->task) {
        case TASK_WCST: {
            wcst_synchronise_world(globals->cs, (WcstTaskData *)globals->task_data);
            wcst_step_event(globals->cs, (WcstTaskData *)globals->task_data);
	    break;
        }
        case TASK_TOL: {
            tol_synchronise_world(globals->cs, (TolTaskData *)globals->task_data);
            tol_step_event(globals->cs, (TolTaskData *)globals->task_data);
	    break;
	}
        case TASK_RNG: {
            rng_synchronise_world(globals->cs, (RngTaskData *)globals->task_data);
            rng_step_event(globals->cs, (RngTaskData *)globals->task_data);
	    break;
	}
        default: {
            break;
        }
    }
#ifdef WM_CONTINUOUS
    /* Propogate all changes to WM activations: */
    wm_recalculate_activations(globals->cs);
#endif
}

/******************************************************************************/
/* Section 3: Detect trial/task completion -----------------------------------*/

Boolean world_trial_complete(GlobalVariables *globals)
{
    switch (globals->task) {
        case TASK_WCST: {
            return(wcst_trial_complete((WcstTaskData *)globals->task_data));
	    break;
        }
        case TASK_TOL: {
            return(tol_trial_complete((TolTaskData *)globals->task_data));
	    break;
        }
        case TASK_RNG: {
            return(rng_trial_complete((RngTaskData *)globals->task_data));
	    break;
        }
        default: {
            return(TRUE);
        }
    }
}

/*----------------------------------------------------------------------------*/

Boolean world_task_complete(GlobalVariables *globals)
{
    switch (globals->task) {
        case TASK_WCST: {
            return(wcst_task_complete((WcstTaskData *)globals->task_data));
	    break;
        }
        case TASK_TOL: {
            return(tol_task_complete((TolTaskData *)globals->task_data));
	    break;
        }
        case TASK_RNG: {
            return(rng_task_complete((RngTaskData *)globals->task_data));
	    break;
        }
        default: {
            return(TRUE);
        }
    }
}

/******************************************************************************/
