/*******************************************************************************

    File:     x_cs.c
    Author:   Rick Cooper
    Created:  28/03/07
    Edited:   29/03/07
    Contents: Routines for displaying and running the SS world model in X

    Public procedures:
        GtkWidget *x_cs_graphs_create_page(GlobalVariables *globals)
        void x_cs_graphs_redraw(GlobalVariables *globals)

*******************************************************************************/

#include "ss.h"
#include <math.h>
#include <gtk/gtk.h>
#include "lib_gtkx.h"

#define xMin  31
#define xMax  11
#define yMin  10
#define yMax  20

/******************************************************************************/
/* Section 1: Redraw the world canvas ----------------------------------------*/

static int calculate_range(int count)
{
    if (count > 10) {
	/* If count < 10^(n+1), round up to nearest 10^n */

	double n = floor(log(count) / log(10.0));
	int scale = (int) pow(10.0, n);

	return(scale * (1 + (count / scale)));
    }
    else {
        return(10);
    }
}

static void draw_plot_line(GdkPixmap *canvas, GdkGC *gc, double *values, int time0, int count, double x_scale, double y_scale, int height)
{
    short x0, y0, x1, y1;

    if (time0 < count) {
        x0 = (time0 * x_scale) + xMin;
        y0 = (height - yMax - (values[time0] * y_scale));

        while (time0++ < count) {
            x1 = (time0 * x_scale) + xMin;
            y1 = (height - yMax - values[time0] * y_scale);
    	    gdk_draw_line(canvas, gc, x0, y0, x1, y1);
            x0 = x1;
            y0 = y1;
        }
    }
}

static void x_cs_graph_draw_schemas(GdkPixmap *canvas, GdkGC *gc, ContentionScheduler *cs)
{
    SchemaList *ss;
    int width, height, range;
    double x_scale, y_scale;

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);
    range = calculate_range(cs->cycle);
    x_scale = (width-xMin-xMax) / (double) range;
    y_scale = (height-yMin-yMax) / (double) (maxAct - minAct);

    for (ss = cs->schemas; ss; ss = ss->tail)    {
        double *act = ss->head->activation;

        if (ss->head->highlighted) {
            gdk_gc_set_foreground(gc, &colour[18]); /* Red */
        }
        else {
            gdk_gc_set_foreground(gc, &colour[2]); /* Royal blue */
        }
        draw_plot_line(canvas, gc, act, 0, cs->cycle, x_scale, y_scale, height);
    }
}

static void x_cs_graph_draw_objects(GdkPixmap *canvas, GdkGC *gc, ContentionScheduler *cs)
{
    ObjectList *os;
    int width, height, role, range;
    double x_scale, y_scale;

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);
    range = calculate_range(cs->cycle);
    x_scale = (width-xMin-xMax) / (double) range;
    y_scale = (height-yMin-yMax) / (double) (maxAct - minAct);

    role = (int) g_object_get_data(G_OBJECT(canvas), "role");
    if ((role < 0) || (role >= MAX_ROLES)) {
        role = 0;
    }

    for (os = cs->objects; os; os = os->tail)    {
        double *act = os->head->activation[role];
    
        if (os->head->highlighted) {
            gdk_gc_set_foreground(gc, &colour[18]); /* Red */
        }
        else {
            gdk_gc_set_foreground(gc, &colour[2]); /* Royal blue */
        }
        draw_plot_line(canvas, gc, act, os->head->time0, cs->cycle, x_scale, y_scale, height);
    }
}

static void x_cs_graph_draw_effectors(GdkPixmap *canvas, GdkGC *gc, ContentionScheduler *cs)
{
    EffectorList *es;
    int width, height, range;
    double x_scale, y_scale;

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);
    range = calculate_range(cs->cycle);
    x_scale = (width-xMin-xMax) / (double) range;
    y_scale = (height-yMin-yMax) / (double) (maxAct - minAct);

    for (es = cs->effectors; es; es = es->tail)    {
        double *act = es->head->activation;

        if (es->head->highlighted) {
            gdk_gc_set_foreground(gc, &colour[18]); /* Red */
        }
        else {
            gdk_gc_set_foreground(gc, &colour[2]); /* Royal blue */
        }
        draw_plot_line(canvas, gc, act, 0, cs->cycle, x_scale, y_scale, height);
    }
}

static void x_cs_graph_redraw(GtkWidget *da, GdkGC *gc, ContentionScheduler *cs)
{
    GdkPixmap *canvas;
    PangoLayout *layout;
    int width, height, i;
    int x_min, x_max;
    void (*draw_data)();

    gtk_widget_realize(da);
    canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(da), "pixmap");
    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);

    x_min = 0;
    x_max = calculate_range(cs->cycle);

    /* Clear the canvas: */
    gdk_gc_set_foreground(gc, &colour[26]);
    gdk_draw_rectangle(canvas, gc, TRUE, 0, 0, width, height);

    /* Draw the graph's grid lines: */
    gdk_gc_set_foreground(gc, &colour[27]); /* Light grey */
    for (i = 1; i < 5; i++) {
        int y = yMin + (i * (height-yMin-yMax)) / 5;
        gdk_draw_line(canvas, gc, xMin, y, width-xMax, y);
    }
    for (i = 1; i < 5; i++) {
        int x = xMin + (i * (width-xMin-xMax)) / 5;
        gdk_draw_line(canvas, gc, x, yMin, x, height-yMax);
    }

    /* Draw the graph's frame: */
    gdk_gc_set_foreground(gc, &colour[0]);
    gdk_draw_rectangle(canvas, gc, FALSE, xMin, yMin, width-xMin-xMax, height-yMin-yMax);

    layout = gtk_widget_create_pango_layout(da, "");
    /* Draw vertical axis tick marks: */
    for (i = 0; i < 6; i++) {
        int w, h, y;
        char buffer[8];

        y = yMin + (i * (height-yMin-yMax)) / 5;
        g_snprintf(buffer, 8, "%5.2f", minAct + (5 - i) * (maxAct - minAct) / 5);
        pango_layout_set_text(layout, buffer, strlen(buffer));
        pango_layout_get_pixel_size(layout, &w, &h);
        gdk_draw_line(canvas, gc, xMin-3, y, xMin, y);
        gdk_draw_layout(canvas, gc, xMin-4-w, y-h/2, layout);
    }
    /* Draw horizontal axis tick marks: */
    for (i = 0; i < 6; i++) {
        int w, h, x;
        char buffer[8];

        x = xMin + (i * (width-xMin-xMax)) / 5;
        g_snprintf(buffer, 8, "%d", x_min + i * (x_max - x_min) / 5);
        pango_layout_set_text(layout, buffer, strlen(buffer));
        pango_layout_get_pixel_size(layout, &w, &h);
        gdk_draw_line(canvas, gc, x, height-yMax+3, x, height-yMax);
        gdk_draw_layout(canvas, gc, x-w/2, height-yMax+5, layout);
    }
    g_object_unref(layout);

    /* Draw the graph's data: */
    draw_data = (void (*)())g_object_get_data(G_OBJECT(da), "draw_function");
    draw_data(canvas, gc, cs);

    /* Return the GC to its former state: */
    gdk_gc_set_foreground(gc, &colour[0]);

    gdk_draw_drawable(da->window, gc, canvas, 0, 0, 0, 0, width, height);
    gtk_widget_queue_draw_area(da, 0, 0, width, height);
    gtkx_flush_events();
}

void x_cs_graphs_redraw(GlobalVariables *globals)
{
    x_cs_graph_redraw(globals->cs_graphs[0], globals->gc, globals->cs);
    x_cs_graph_redraw(globals->cs_graphs[1], globals->gc, globals->cs);
    x_cs_graph_redraw(globals->cs_graphs[2], globals->gc, globals->cs);
}

/******************************************************************************/
/* Section 2: Create menus for the CS graphs: --------------------------------*/

/* Highlight callback functions ----------------------------------------------*/

static void highlight_schema_none(GtkWidget *button, GlobalVariables *globals)
{
    SchemaList *tmp;

    for (tmp = globals->cs->schemas; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    x_cs_graph_redraw(globals->cs_graphs[0], globals->gc, globals->cs);
}

static void highlight_schema(GtkWidget *button, GlobalVariables *globals)
{
    Schema *tmp = g_object_get_data(G_OBJECT(button), "schema");

    tmp->highlighted = !(tmp->highlighted);
    x_cs_graph_redraw(globals->cs_graphs[0], globals->gc, globals->cs);
}

static void highlight_schema_all(GtkWidget *button, GlobalVariables *globals)
{
    SchemaList *tmp;

    for (tmp = globals->cs->schemas; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    x_cs_graph_redraw(globals->cs_graphs[0], globals->gc, globals->cs);
}

/*----------------------------------------------------------------------------*/

static void highlight_object_none(GtkWidget *button, GlobalVariables *globals)
{
    ObjectList *tmp;

    for (tmp = globals->cs->objects; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    x_cs_graph_redraw(globals->cs_graphs[1], globals->gc, globals->cs);
}

static void highlight_object(GtkWidget *button, GlobalVariables *globals)
{
    Object *tmp = g_object_get_data(G_OBJECT(button), "object");

    tmp->highlighted = !(tmp->highlighted);
    x_cs_graph_redraw(globals->cs_graphs[1], globals->gc, globals->cs);
}

static void highlight_object_all(GtkWidget *button, GlobalVariables *globals)
{
    ObjectList *tmp;

    for (tmp = globals->cs->objects; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    x_cs_graph_redraw(globals->cs_graphs[1], globals->gc, globals->cs);
}

/*----------------------------------------------------------------------------*/

static void highlight_effector_none(GtkWidget *button, GlobalVariables *globals)
{
    EffectorList *tmp;

    for (tmp = globals->cs->effectors; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    x_cs_graph_redraw(globals->cs_graphs[2], globals->gc, globals->cs);
}

static void highlight_effector(GtkWidget *button, GlobalVariables *globals)
{
    Effector *tmp = g_object_get_data(G_OBJECT(button), "effector");

    tmp->highlighted = !(tmp->highlighted);
    x_cs_graph_redraw(globals->cs_graphs[2], globals->gc, globals->cs);
}

static void highlight_effector_all(GtkWidget *button, GlobalVariables *globals)
{
    EffectorList *tmp;

    for (tmp = globals->cs->effectors; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    x_cs_graph_redraw(globals->cs_graphs[2], globals->gc, globals->cs);
}

/*----------------------------------------------------------------------------*/

static Boolean schemaMenuGenerate(GtkWidget *item, GdkEvent *event, GlobalVariables *globals)
{
    GtkWidget *menu, *mi;
    SchemaList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label("None");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_schema_none), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    for (tmp = globals->cs->schemas; tmp != NULL; tmp = tmp->tail) {
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
        mi = gtk_menu_item_new_with_label(buffer);
        g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_schema), globals);
        g_object_set_data(G_OBJECT(mi), "schema", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_schema_all), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static Boolean objectMenuGenerate(GtkWidget *item, GdkEvent *event, GlobalVariables *globals)
{
    GtkWidget *menu, *mi;
    ObjectList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_object_none), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    for (tmp = globals->cs->objects; tmp != NULL; tmp = tmp->tail) {
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
        mi = gtk_menu_item_new_with_label(buffer);
        g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_object), globals);
        g_object_set_data(G_OBJECT(mi), "object", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_object_all), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static Boolean effectorMenuGenerate(GtkWidget *item, GdkEvent *event, GlobalVariables *globals)
{
    GtkWidget *menu, *mi;
    EffectorList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_effector_none), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    for (tmp = globals->cs->effectors; tmp != NULL; tmp = tmp->tail) {
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->name);
        mi = gtk_menu_item_new_with_label(buffer);
        g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_effector), globals);
        g_object_set_data(G_OBJECT(mi), "effector", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_effector_all), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

/******************************************************************************/
/* Section 3: Create the canvas and set canvas event handlers ----------------*/

static int cs_graph_event_configure(GtkWidget *da, GdkEventConfigure *event, GlobalVariables *globals)
{
    gtkx_canvas_reallocate(da);
    x_cs_graph_redraw(da, globals->gc, globals->cs);
    return(1);
}

/*----------------------------------------------------------------------------*/

static GtkWidget *x_cs_create_graph(GlobalVariables *globals, void draw_data())
{
    GtkWidget *da = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(da), "configure_event", G_CALLBACK(cs_graph_event_configure), globals);
    g_signal_connect(G_OBJECT(da), "expose_event", G_CALLBACK(gtkx_canvas_event_expose), NULL);
    g_signal_connect(G_OBJECT(da), "destroy", G_CALLBACK(gtkx_canvas_callback_destroy), NULL);
    g_object_set_data(G_OBJECT(da), "draw_function", draw_data);
    gtk_widget_show(da);
    return(da);
}

/*----------------------------------------------------------------------------*/

static GtkWidget *x_cs_graph_pack_hbox(GtkWidget *page)
{
    GtkWidget *hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);
    return(hbox);
}

/*----------------------------------------------------------------------------*/

static void x_cs_graph_pack_label(GtkWidget *box, char *label)
{
    GtkWidget *item = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(box), GTK_WIDGET(item), FALSE, FALSE, 0);
    gtk_widget_show(item);
}

/*----------------------------------------------------------------------------*/

static void x_cs_graph_pack_highlight_menu(GtkWidget *box, Boolean generate_menu(), GlobalVariables *globals)
{
    GtkWidget *menu_bar = gtk_menu_bar_new();
    GtkWidget *menu = gtk_menu_new();
    GtkWidget *item = gtk_menu_item_new_with_label("Highlight");

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
    g_signal_connect(G_OBJECT(item), "enter_notify_event", G_CALLBACK(generate_menu), globals);
    gtk_widget_show(item);

    gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), item);
    gtk_widget_show(menu_bar);

    gtk_box_pack_end(GTK_BOX(box), GTK_WIDGET(menu_bar), FALSE, FALSE, 0);
}

/*----------------------------------------------------------------------------*/

static void select_object_role_callback(GtkWidget *item, GlobalVariables *globals)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(globals->cs_graphs[1]), "pixmap");
    int role = (int) g_object_get_data(G_OBJECT(item), "role");

    g_object_set_data(G_OBJECT(canvas), "role", (void *)role);
    x_cs_graph_redraw(globals->cs_graphs[1], globals->gc, globals->cs);
}

static void x_cs_graph_pack_role_options(GtkWidget *hbox, GlobalVariables *globals)
{
    /* The object role menu: */

    GSList *rb_grp = NULL;
    GtkWidget *label, *mi;
    int i;

    label = gtk_label_new("            Role: ");
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    for (i = 0; i < MAX_ROLES-1; i++) {
        mi = gtk_radio_button_new_with_label(rb_grp, roleName[i]);
        gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(mi), FALSE, FALSE, 5);
        g_signal_connect(G_OBJECT(mi), "toggled", G_CALLBACK(select_object_role_callback), globals);
        g_object_set_data(G_OBJECT(mi), "role", (void *)i);
        gtk_widget_show(mi);
        rb_grp = gtk_radio_button_group(GTK_RADIO_BUTTON(mi));
    }
}

/*----------------------------------------------------------------------------*/

GtkWidget *x_cs_graphs_create_page(GlobalVariables *globals)
{
    GtkWidget *page, *hbox;
    
    page = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(page), 5);

    /* Schema graph/menu: */

    hbox = x_cs_graph_pack_hbox(page);
    x_cs_graph_pack_label(hbox, "Schemas:");
    x_cs_graph_pack_highlight_menu(hbox, schemaMenuGenerate, globals);
    globals->cs_graphs[0] = x_cs_create_graph(globals, x_cs_graph_draw_schemas);
    gtk_box_pack_start(GTK_BOX(page), globals->cs_graphs[0], TRUE, TRUE, 0);

    /* Object graph/menu: */

    hbox = x_cs_graph_pack_hbox(page);
    x_cs_graph_pack_label(hbox, "Objects:");
    x_cs_graph_pack_role_options(hbox, globals);
    x_cs_graph_pack_highlight_menu(hbox, objectMenuGenerate, globals);
    globals->cs_graphs[1] = x_cs_create_graph(globals, x_cs_graph_draw_objects);
    gtk_box_pack_start(GTK_BOX(page), globals->cs_graphs[1], TRUE, TRUE, 0);

    /* Effector graph/menu: */

    hbox = x_cs_graph_pack_hbox(page);
    x_cs_graph_pack_label(hbox, "Effectors:");
    x_cs_graph_pack_highlight_menu(hbox, effectorMenuGenerate, globals);
    globals->cs_graphs[2] = x_cs_create_graph(globals, x_cs_graph_draw_effectors);
    gtk_box_pack_start(GTK_BOX(page), globals->cs_graphs[2], TRUE, TRUE, 0);

    // FIX ME: Add status bar with cycle count?

    return(page);
}

/******************************************************************************/
