/*******************************************************************************

    File:     x_frame.c
    Author:   Rick Cooper
    Created:  28/03/07
    Edited:   29/03/07
    Contents: Routines for maintaining the X interface to SS

    Public procedures:
        Boolean widgets_create(GlobalVariables *globals)
        void widgets_initialise(GlobalVariables *globals)
        void x_redraw_canvases(GlobalVariables *globals, Boolean force)

*******************************************************************************/

#include "ss.h"
#include <gtk/gtk.h>
#include "lib_gtkx.h"

/******************************************************************************/

extern void x_task_initialise(GtkWidget *caller, GlobalVariables *globals);
extern void x_task_run_event(GtkWidget *caller, GlobalVariables *globals);
extern void x_task_run_trial(GtkWidget *caller, GlobalVariables *globals);
extern void x_task_run_block(GtkWidget *caller, GlobalVariables *globals);

extern GtkWidget *x_world_create_page(GlobalVariables *globals);
extern void x_world_redraw(GlobalVariables *globals);
extern GtkWidget *x_analysis_create_page(GlobalVariables *globals);
extern void x_analysis_redraw(GlobalVariables *globals);
extern GtkWidget *x_cs_graphs_create_page(GlobalVariables *globals);
extern void x_cs_graphs_redraw(GlobalVariables *globals);
extern GtkWidget *x_cs_logs_create_page(ContentionScheduler *cs);
extern GtkWidget *x_wm_graph_create_page(GlobalVariables *globals);
extern void x_wm_graph_redraw(GlobalVariables *globals);

/******************************************************************************/
/* Section 1: Create the interface -------------------------------------------*/

static void redraw_all_canvases(GtkWidget *caller, GlobalVariables *globals)
{
    x_world_redraw(globals);
    x_analysis_redraw(globals);
    x_cs_graphs_redraw(globals);
    x_wm_graph_redraw(globals);
}

/*----------------------------------------------------------------------------*/

void x_toggle_dynamic_display(GtkWidget *caller, GlobalVariables *globals)
{
    globals->dynamic_canvases = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(caller));
}

/*----------------------------------------------------------------------------*/

void x_select_task(GtkWidget *caller, GlobalVariables *globals)
{
    task_destroy(globals);
    globals->task = (TaskId) gtk_combo_box_get_active(GTK_COMBO_BOX(caller));
    task_create(globals);
    gtk_list_store_clear(GTK_LIST_STORE(gtk_tree_view_get_model(globals->cs->cl_actions)));
    if (!world_initialise_task(globals)) {
        io_message(globals->window, "World initialisation failed.\nSwitching to NULL task.");
        gtk_combo_box_set_active(GTK_COMBO_BOX(caller), 0);
    }
    else {
        cs_reinitialise(globals->cs, "tmp.log");
        redraw_all_canvases(NULL, globals);
    }
}

/*----------------------------------------------------------------------------*/

static gint event_delete(GtkWidget *caller, GdkEvent *event, GlobalVariables *globals)
{
    return(0);    // Return 1 if we need to abort
}

/*----------------------------------------------------------------------------*/

static void callback_destroy(GtkWidget *caller, GlobalVariables *globals)
{
    gtk_main_quit();
}

/*----------------------------------------------------------------------------*/

Boolean widgets_create(GlobalVariables *globals)
{
    if ((globals->window = gtk_window_new(GTK_WINDOW_TOPLEVEL)) == NULL) {
        return(FALSE);
    }
    else {
        GtkWidget *notebook, *page, *hbox, *toolbar, *tmp;
	GtkToolItem *tool_item;
	GtkTooltips *tooltips;
        int position = 0;

        gtk_window_set_default_size(GTK_WINDOW(globals->window), 600, 700);
        gtk_window_set_title(GTK_WINDOW(globals->window), "X_SS");
        g_signal_connect(G_OBJECT(globals->window), "delete_event", G_CALLBACK(event_delete), globals);
        g_signal_connect(G_OBJECT(globals->window), "destroy", G_CALLBACK(callback_destroy), globals);

        tooltips = gtk_tooltips_new();

        page = gtk_vbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(globals->window), page);
        gtk_widget_show(page);

        /* The top-most menu & tool bar: */

        hbox = gtk_hbox_new(FALSE, 0);
        gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
        gtk_widget_show(hbox);

        /* The left-hand buttons: */

        tmp = gtk_label_new("Task:");
        gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
        gtk_widget_show(tmp);

        tmp = gtk_combo_box_new_text();
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), "None");
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), "WCST");
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), "ToL");
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), "RNG");
        gtk_combo_box_set_active(GTK_COMBO_BOX(tmp), 0);
        g_signal_connect(G_OBJECT(tmp), "changed", G_CALLBACK(x_select_task), globals);
        gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
        gtk_widget_show(tmp);

        tmp = gtk_check_button_new_with_label("Dynamic display");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), globals->dynamic_canvases);
        g_signal_connect(G_OBJECT(tmp), "toggled", G_CALLBACK(x_toggle_dynamic_display), globals);
        gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 50);
        gtk_widget_show(tmp);

        /* The toolbar: */

        toolbar = gtk_toolbar_new();
        gtk_box_pack_start(GTK_BOX(hbox), toolbar, TRUE, TRUE, 0);
        gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
        gtk_widget_show(toolbar);

        tool_item = gtk_tool_button_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS); // Initialise
        g_signal_connect(G_OBJECT(tool_item), "clicked", G_CALLBACK(x_task_initialise), globals);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_tooltips_set_tip(tooltips, GTK_WIDGET(tool_item), "Reinitialise the model", NULL);
        gtk_widget_show(GTK_WIDGET(tool_item));

        tool_item = gtk_tool_button_new_from_stock(GTK_STOCK_MEDIA_PLAY); // Run Event
        g_signal_connect(G_OBJECT(tool_item), "clicked", G_CALLBACK(x_task_run_event), globals);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_tooltips_set_tip(tooltips, GTK_WIDGET(tool_item), "Run one event", NULL);
        gtk_widget_show(GTK_WIDGET(tool_item));

        tool_item = gtk_tool_button_new_from_stock(GTK_STOCK_MEDIA_FORWARD); // Run Trial
        g_signal_connect(G_OBJECT(tool_item), "clicked", G_CALLBACK(x_task_run_trial), globals);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_tooltips_set_tip(tooltips, GTK_WIDGET(tool_item), "Run until the end of the current trial", NULL);
        gtk_widget_show(GTK_WIDGET(tool_item));

        tool_item = gtk_tool_button_new_from_stock(GTK_STOCK_MEDIA_NEXT); // Run Block
        g_signal_connect(G_OBJECT(tool_item), "clicked", G_CALLBACK(x_task_run_block), globals);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_tooltips_set_tip(tooltips, GTK_WIDGET(tool_item), "Run until the end of task", NULL);
        gtk_widget_show(GTK_WIDGET(tool_item));

        tool_item = gtk_separator_tool_item_new();
        gtk_separator_tool_item_set_draw(GTK_SEPARATOR_TOOL_ITEM(tool_item), FALSE);
        gtk_tool_item_set_expand(GTK_TOOL_ITEM(tool_item), TRUE);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_widget_show(GTK_WIDGET(tool_item));

        tool_item = gtk_tool_button_new_from_stock(GTK_STOCK_REFRESH); // Redraw canvases
        g_signal_connect(G_OBJECT(tool_item), "clicked", G_CALLBACK(redraw_all_canvases), globals);
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tool_item, position++);
        gtk_tooltips_set_tip(tooltips, GTK_WIDGET(tool_item), "Refresh all canvases", NULL);
        gtk_widget_show(GTK_WIDGET(tool_item));

        /* The notebook: */

        notebook = gtk_notebook_new();
        gtk_box_pack_start(GTK_BOX(page), notebook, TRUE, TRUE, 0);
        gtk_widget_show(notebook);

        tmp = x_world_create_page(globals);
        gtkx_notebook_append_page(notebook, tmp, "World");

        tmp = x_cs_graphs_create_page(globals);
        gtkx_notebook_append_page(notebook, tmp, "CS Graphs");

        tmp = x_cs_logs_create_page(globals->cs);
        gtkx_notebook_append_page(notebook, tmp, "CS Logs");

        tmp = x_wm_graph_create_page(globals);
        gtkx_notebook_append_page(notebook, tmp, "WM Graphs");

        tmp = x_analysis_create_page(globals);
        gtkx_notebook_append_page(notebook, tmp, "Analyses");

        return(TRUE);
    }
}

/******************************************************************************/
/* Section 2: Initialise the interface ---------------------------------------*/

void widgets_initialise(GlobalVariables *globals)
{
    gtk_widget_realize(globals->window);
    initialise_colours(globals->window);
    globals->gc = create_gcs(globals->window);
}

/******************************************************************************/
/* Section 3: Redraw all canvasses -------------------------------------------*/

void x_redraw_canvases(GlobalVariables *globals, Boolean force)
{
    if (force) {
        redraw_all_canvases(NULL, globals);
    }
}

/******************************************************************************/
