/*******************************************************************************

    File:     x_rng.c
    Author:   Rick Cooper
    Created:  29/03/07
    Edited:   04/04/07
    Contents: Routines displaying the ToL in X

    Public procedures:
        void x_rng_redraw_world(GlobalVariables *globals, TolTaskData *task)

*******************************************************************************/

#include "ss.h"
#include <math.h>
#include <gtk/gtk.h>
#include "lib_gtkx.h"
#include "rng.h"

/******************************************************************************/
/* Section 1: Redraw the world -----------------------------------------------*/

static void draw_rng_disks(cairo_t *cr, PangoLayout *layout, int width, int height, RngDisks *disks)
{
    while (disks != NULL) {
        short r, g, b;
        int x, y;

	r = disks->colour / 9;
	g = (disks->colour - (9 * r)) / 3;
	b = disks->colour - (9 * r) - (3 * g);

        cairo_set_source_rgb(cr, r / 2.0, g / 2.0, b / 2.0);

        x = width / 2 + disks->x;
        y = height / 2 + disks->y;

        cairo_move_to(cr, x-25, y);
        cairo_arc(cr, x, y, 25, 0, 2*M_PI);
        cairo_fill(cr);

        cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
        if (disks->id >= 0) {
            int w, h;
            char label[2] = {'\0', '\0'};

            label[0] = 'A' + disks->id;
            pango_layout_set_text(layout, label, 1);
            pango_layout_get_size(layout, &w, &h);
            cairo_move_to(cr, x-w/(2*PANGO_SCALE), y-h/(2*PANGO_SCALE));
            pango_cairo_show_layout(cr, layout);


        }
        else {
            cairo_move_to(cr, x-3, y-3);
            cairo_line_to(cr, x+3, y+3);
            cairo_move_to(cr, x-3, y+3);
            cairo_line_to(cr, x+3, y-3);
            cairo_stroke(cr);
        }
        disks = disks->next;
    }
}

/*----------------------------------------------------------------------------*/

void x_rng_redraw_world(GlobalVariables *globals, cairo_t *cr, RngTaskData *task)
{
    GtkWidget *da = globals->world_da;
    int width = da->allocation.width;
    int height = da->allocation.height;
    PangoLayout *layout;
    char buffer[64];

    /* Black canvas: */
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, 0, 0, width, height);
    cairo_fill(cr);

    /* Set up the text layout: */
    layout = gtk_widget_create_pango_layout(globals->world_da, "");
    
    /* Draw current state of the world: */

    draw_rng_disks(cr, layout, width, height, task->disks);

    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    g_snprintf(buffer, 64, "Trial is %d.%02d", task->trial_count + 1, task->event_count + 1);
    pango_layout_set_text(layout, buffer, strlen(buffer));
    cairo_move_to(cr, 20, 600);
    pango_cairo_show_layout(cr, layout);

    /* Release the text layout: */
    g_object_unref(layout);
}

/******************************************************************************/
/* Section 2: Redraw analyses ------------------------------------------------*/

static void x_rng_centre_text(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, int x, int y, char *label)
{
    int w, h;

    pango_layout_set_text(layout, label, strlen(label));
    pango_layout_get_pixel_size(layout, &w, &h);
    gdk_draw_layout(canvas, gc, x-w/2, y, layout);
}

/*----------------------------------------------------------------------------*/

#define W 70

static void x_rng_draw_row_data(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, RngTaskData *task, int i, int x0, int y0)
{
    char buffer[16];

    g_snprintf(buffer, 16, "%5.3f", task->score_r[i]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer);
    g_snprintf(buffer, 16, "%5.3f", task->score_rng[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer);
    g_snprintf(buffer, 16, "%5.3f", task->score_rr[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer);
    g_snprintf(buffer, 16, "%5.3f", task->score_oa[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer);
    g_snprintf(buffer, 16, "%5.3f", task->score_aa[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); 
    g_snprintf(buffer, 16, "%5.3f", task->score_tpi[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); 
    g_snprintf(buffer, 16, "%5.3f", task->score_rg[i]); x0 = x0+W;
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); 
}

/*----------------------------------------------------------------------------*/

static double mean(double scores[RNG_BLOCKS], int count)
{
    int i;
    double s = 0;

    for (i = 0; i < count; i++) {
        s += scores[i];
    }
    return(s / (double) count);
}

static double std_err(double scores[RNG_BLOCKS], int count)
{
    int i;
    double s = 0, ss = 0, std;

    for (i = 0; i < count; i++) {
        s += scores[i];
        ss += (scores[i] * scores[i]);
    }
    std = sqrt((ss - (s * s / (double) count)) / (double) (count - 1.0));

    return(std / sqrt(count));
}

static void x_rng_draw_mean_data(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, RngTaskData *task, int x0, int y0)
{
    char buffer[16];

    x_rng_centre_text(canvas, gc, layout, x0-30, y0, "Mean");

    g_snprintf(buffer, 16, "%5.3f", mean(task->score_r, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_rng, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_rr, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_oa, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_aa, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_tpi, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", mean(task->score_rg, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
}

static void x_rng_draw_stderr_data(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, RngTaskData *task, int x0, int y0)
{
    char buffer[16];

    x_rng_centre_text(canvas, gc, layout, x0-30, y0, "S.E.");

    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_r, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_rng, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_rr, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_oa, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_aa, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_tpi, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "(%5.3f)", std_err(task->score_rg, task->block_count));
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
}

static void x_rng_draw_subject_data(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, int x0, int y0, char *label, double scores[7])
{
    char buffer[16];

    x_rng_centre_text(canvas, gc, layout, x0-30, y0, label);

    g_snprintf(buffer, 16, "%5.3f", scores[0]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[1]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[2]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[3]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[4]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[5]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
    g_snprintf(buffer, 16, "%5.3f", scores[6]);
    x_rng_centre_text(canvas, gc, layout, x0+40, y0, buffer); x0 = x0+W;
}

/*----------------------------------------------------------------------------*/

void x_rng_redraw_analysis(GlobalVariables *globals, RngTaskData *task)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(globals->analysis_da), "pixmap");
    PangoLayout *layout = gtk_widget_create_pango_layout(globals->analysis_da, "");;
    int width, height;
    int x0, y0, i, y1;

    double scores_ctl[7] = {0.962, 0.300, 0.014, 0.131, 0.259, 44.0, 9.731};
    double scores_ds[7]  = {2.408, 0.410, 0.004, 0.136, 0.328, 36.3, 9.540};
    double scores_2b[7]  = {1.979, 0.461, 0.002, 0.097, 0.424, 26.7, 9.561};
    double scores_ss[7]  = {1.196, 0.388, 0.005, 0.130, 0.334, 35.0, 9.654};

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);

    gdk_gc_set_foreground(globals->gc, &colour[26]);
    gdk_draw_rectangle(canvas, globals->gc, TRUE, 0, 0, width, height);
    gdk_gc_set_foreground(globals->gc, &colour[0]);
    
    x0 = MAX(20, (width - (7*W)) / 2); y0 = 30;

    if (task->block_count > 0) {
        y1 = y0 + 37 + 23 + task->block_count * 18;
    }
    else {
        y1 = y0 + 23;
    }

    gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);
    y0 = y0 + 4;
    /* Vertical table lines and column labels: */
    gdk_draw_line(canvas, globals->gc, x0, y0-4, x0, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35, y0, "R");
    gdk_draw_line(canvas, globals->gc, x0+1*W, y0-4, x0+1*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+1*W, y0, "RNG");
    gdk_draw_line(canvas, globals->gc, x0+2*W, y0-4, x0+2*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+2*W, y0, "RR");
    gdk_draw_line(canvas, globals->gc, x0+3*W, y0-4, x0+3*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+3*W, y0, "OA");
    gdk_draw_line(canvas, globals->gc, x0+4*W, y0-4, x0+4*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+4*W, y0, "AA");
    gdk_draw_line(canvas, globals->gc, x0+5*W, y0-4, x0+5*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+5*W, y0, "TPI");
    gdk_draw_line(canvas, globals->gc, x0+6*W, y0-4, x0+6*W, y1);
    x_rng_centre_text(canvas, globals->gc, layout, x0+35+6*W, y0, "RG");
    gdk_draw_line(canvas, globals->gc, x0+7*W, y0-4, x0+7*W, y1);

    y0 = y0 + 16;
    gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);
    y0 = y0 + 3;
    gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);

    /* Fill in the data for each block: */
    for (i = 0; i < task->block_count; i++) {
        y0 = y0 + 2;
        x_rng_draw_row_data(canvas, globals->gc, layout, task, i, x0, y0);
        y0 = y0 + 16;
        gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);
    }

    /* Show mean scores across blocks: */
    if (task->block_count > 0) {
        y0 = y0 + 3;
        gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);
        y0 = y0 + 2;
        x_rng_draw_mean_data(canvas, globals->gc, layout, task, x0, y0);
        y0 = y0 + 16;
        x_rng_draw_stderr_data(canvas, globals->gc, layout, task, x0, y0);
        y0 = y0 + 16;
        gdk_draw_line(canvas, globals->gc, x0, y0, x0+7*W, y0);
    }

    x_rng_draw_subject_data(canvas, globals->gc, layout, x0, y0+100, "CTL", scores_ctl);
    x_rng_draw_subject_data(canvas, globals->gc, layout, x0, y0+120, "DS", scores_ds);
    x_rng_draw_subject_data(canvas, globals->gc, layout, x0, y0+140, "2B", scores_2b);
    x_rng_draw_subject_data(canvas, globals->gc, layout, x0, y0+160, "SS", scores_ss);

    g_object_unref(layout);
}

/******************************************************************************/
