/*******************************************************************************

    File:     x_tol.c
    Author:   Rick Cooper
    Created:  29/03/07
    Edited:   04/04/07
    Contents: Routines displaying the ToL in X

    Public procedures:
        void x_tol_redraw_world(GlobalVariables *globals, TolTaskData *task)

*******************************************************************************/

#include "ss.h"
#include <gtk/gtk.h>
#include "lib_gtkx.h"
#include "tol.h"

/******************************************************************************/
/* Section 1: Redraw the world -----------------------------------------------*/

static void draw_tol_ball(GdkPixmap *canvas, GdkGC *gc, int x0, int y0, BallPosition *p, int hx, int c)
{
    int x, y;

    gdk_gc_set_foreground(gc, &colour[c]);
    if ((p->peg == 0.0) && (p->position == 0.0)) {
        /* The ball is held: Set coordinates appropriately */
        x = x0 + 290 + 45 * hx;
        y = y0 - 160;
    }
    else {
        x = x0 + 350 - 100 * p->peg;
        y = y0 + 17 - 40 * p->position;
    }
    gdk_draw_arc(canvas, gc, TRUE, x-20, y-20, 40, 40, 0, 23040);
}

/*----------------------------------------------------------------------------*/

static void draw_tol_apparatus(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, int x0, int y0, char *label, TolConfiguration *configuration)
{
    gdk_gc_set_foreground(gc, &colour[13]); /* Grey */

    /* The base: */
    gdk_draw_rectangle(canvas, gc, TRUE, x0+10, y0-3, 280, 7);

    /* The pegs: */
    gdk_draw_rectangle(canvas, gc, TRUE, x0+247, y0-45, 7, 42);
    gdk_draw_rectangle(canvas, gc, TRUE, x0+147, y0-85, 7, 82);
    gdk_draw_rectangle(canvas, gc, TRUE, x0+47, y0-125, 7, 122);

    /* The label: */
    gdk_gc_set_foreground(gc, &colour[0]);
    pango_layout_set_text(layout, label, strlen(label));
    gdk_draw_layout(canvas, gc, x0-130, y0-70, layout);

    /* The balls: */
    draw_tol_ball(canvas, gc, x0, y0, &(configuration->red_ball), 1, 18);
    draw_tol_ball(canvas, gc, x0, y0, &(configuration->green_ball), 2, 6);
    draw_tol_ball(canvas, gc, x0, y0, &(configuration->blue_ball), 3, 2);
}

/*----------------------------------------------------------------------------*/

void x_tol_redraw_world(GlobalVariables *globals, TolTaskData *task)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(globals->world_da), "pixmap");
    PangoLayout *layout;
    TolProblem *p = &(task->problems[task->trial_count]);
    int width, height;
    char buffer[64];

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);

    /* Clear the canvas: */
    gdk_gc_set_foreground(globals->gc, &colour[26]);
    gdk_draw_rectangle(canvas, globals->gc, TRUE, 0, 0, width, height);

    /* Set up the text layout: */
    layout = gtk_widget_create_pango_layout(globals->world_da, "");
    pango_layout_set_width(layout, 100*PANGO_SCALE);
    
    /* Draw current and desired configurations: */
    draw_tol_apparatus(canvas, globals->gc, layout, 150, 200, "Current Configuration:", &(p->current_state));
    draw_tol_apparatus(canvas, globals->gc, layout, 150, 500, "Desired Configuration:", &(p->desired_state));

    gdk_gc_set_foreground(globals->gc, &colour[0]);
    g_snprintf(buffer, 64, "Trial is %d.%02d", task->trial_count + 1, task->event_count + 1);
    pango_layout_set_text(layout, buffer, strlen(buffer));
    gdk_draw_layout(canvas, globals->gc, 20, 600, layout);

    /* Release the text layout: */
    g_object_unref(layout);
}

/******************************************************************************/
/* Section 2: Redraw analyses ------------------------------------------------*/

static void x_tol_centre_text(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, int x, int y, char *label)
{
    int w, h;

    pango_layout_set_text(layout, label, strlen(label));
    pango_layout_get_pixel_size(layout, &w, &h);
    gdk_draw_layout(canvas, gc, x-w/2, y, layout);
}

/*----------------------------------------------------------------------------*/

static void x_tol_draw_row_data(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, int i, TolProblem *problem, int x0, int y0)
{
    char buffer[16];

    g_snprintf(buffer, 16, "Problem %d", i + 1);
    x_tol_centre_text(canvas, gc, layout, x0+50, y0, buffer);
    g_snprintf(buffer, 16, "%4d", problem->moves_minimum);
    x_tol_centre_text(canvas, gc, layout, x0+150, y0, buffer);
    g_snprintf(buffer, 16, "%4d", problem->moves_actual);
    x_tol_centre_text(canvas, gc, layout, x0+250, y0, buffer);
    g_snprintf(buffer, 16, "%4d", problem->rule_breaks);
    x_tol_centre_text(canvas, gc, layout, x0+350, y0, buffer);
    g_snprintf(buffer, 16, "%4d", problem->balls_correct);
    x_tol_centre_text(canvas, gc, layout, x0+450, y0, buffer); 
}

/*----------------------------------------------------------------------------*/

static void x_tol_draw_column_means(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, TolTaskData *task, int x0, int y0)
{
    int mm = 0, ma = 0, rb = 0, bc = 0, i;
    char buffer[16];

    x_tol_centre_text(canvas, gc, layout, x0+ 50, y0, "Mean");
    for (i = 0; i <= task->trial_count; i++) {
        mm += task->problems[i].moves_minimum;
        ma += task->problems[i].moves_actual;
        rb += task->problems[i].rule_breaks;
        bc += task->problems[i].balls_correct;
    }
    g_snprintf(buffer, 16, "%6.3f", mm / (double) i);
    x_tol_centre_text(canvas, gc, layout, x0+150, y0, buffer);
    g_snprintf(buffer, 16, "%6.3f", ma / (double) i);
    x_tol_centre_text(canvas, gc, layout, x0+250, y0, buffer);
    g_snprintf(buffer, 16, "%6.3f", rb / (double) i);
    x_tol_centre_text(canvas, gc, layout, x0+350, y0, buffer);
    g_snprintf(buffer, 16, "%6.3f", bc / (double) i);
    x_tol_centre_text(canvas, gc, layout, x0+450, y0, buffer); 
}

/*----------------------------------------------------------------------------*/

void x_tol_redraw_analysis(GlobalVariables *globals, TolTaskData *task)
{
    GdkPixmap *canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(globals->analysis_da), "pixmap");
    PangoLayout *layout = gtk_widget_create_pango_layout(globals->analysis_da, "");;
    int width, height;
    int x0, y0, i;

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);

    gdk_gc_set_foreground(globals->gc, &colour[26]);
    gdk_draw_rectangle(canvas, globals->gc, TRUE, 0, 0, width, height);
    gdk_gc_set_foreground(globals->gc, &colour[0]);
    
    x0 = MAX(20, (width - 500) / 2); y0 = 60;

    gdk_draw_line(canvas, globals->gc, x0, y0-10, x0+500, y0-10);
    /* Vertical table lines and column labels: */
    gdk_draw_line(canvas, globals->gc, x0, y0-10, x0, y0+86+(30*task->trial_count));
    x_tol_centre_text(canvas, globals->gc, layout, x0+ 50, y0, "Problem Id");
    gdk_draw_line(canvas, globals->gc, x0+100, y0-10, x0+100, y0+86+(30*task->trial_count));
    x_tol_centre_text(canvas, globals->gc, layout, x0+150, y0, "Min. Moves");
    gdk_draw_line(canvas, globals->gc, x0+200, y0-10, x0+200, y0+86+(30*task->trial_count));
    x_tol_centre_text(canvas, globals->gc, layout, x0+250, y0, "Actual Moves");
    gdk_draw_line(canvas, globals->gc, x0+300, y0-10, x0+300, y0+86+(30*task->trial_count));
    x_tol_centre_text(canvas, globals->gc, layout, x0+350, y0, "Rule Breaks");
    gdk_draw_line(canvas, globals->gc, x0+400, y0-10, x0+400, y0+86+(30*task->trial_count));
    x_tol_centre_text(canvas, globals->gc, layout, x0+450, y0, "Correct Balls");
    gdk_draw_line(canvas, globals->gc, x0+500, y0-10, x0+500, y0+86+(30*task->trial_count));

    y0 = y0 + 30;
    gdk_draw_line(canvas, globals->gc, x0, y0-10, x0+500, y0-10);
    y0 = y0 + 3;
    gdk_draw_line(canvas, globals->gc, x0, y0-10, x0+500, y0-10);

    /* Fill in the data for each problem (row): */

    for (i = 0; i <= task->trial_count; i++) {
        x_tol_draw_row_data(canvas, globals->gc, layout, i, &(task->problems[i]), x0, y0);
        gdk_draw_line(canvas, globals->gc, x0, y0+20, x0+500, y0+20);
        y0 = y0 + 30;
    }
    y0 = y0 - 27;
    gdk_draw_line(canvas, globals->gc, x0, y0+20, x0+500, y0+20);

    /* Fill in column means: */
    x_tol_draw_column_means(canvas, globals->gc, layout, task, x0, y0+30);
    gdk_draw_line(canvas, globals->gc, x0, y0+50, x0+500, y0+50);

    g_object_unref(layout);
}

/******************************************************************************/
