/*******************************************************************************

    File:     x_wm.c
    Author:   Rick Cooper
    Created:  19/04/07
    Edited:   19/04/07
    Contents: Routines for displaying WM activation for the SS model in X

    Public procedures:
        GtkWidget *x_wm_graph_create_page(GlobalVariables *globals)
        void x_wm_graph_redraw(GlobalVariables *globals)

*******************************************************************************/

#include "ss.h"
#include <math.h>
#include <gtk/gtk.h>
#include "lib_gtkx.h"

#define xMin  32
#define xMax  15
#define yMin  10
#define yMax  20

/******************************************************************************/
/* Section 1: Redraw the world canvas ----------------------------------------*/

static int calculate_range(int count)
{
    if (count > 10) {
	/* If count < 10^(n+1), round up to nearest 10^n */

        double n = floor(log(count) / log(10.0));
        int scale = (int) pow(10.0, n);

        return(scale * (1 + (count / scale)));
    }
    else {
        return(10);
    }
}

#ifdef WM_CONTINUOUS
static void draw_plot_line(GdkPixmap *canvas, GdkGC *gc, double *values, int time0, int count, double x_scale, double y_scale, int height)
{
    short x0, y0, x1, y1;

    if (time0 < count) {
        x0 = (time0 * x_scale) + xMin;
        y0 = (height - yMax - (values[time0] * y_scale));

        while (time0++ < count) {
            x1 = (time0 * x_scale) + xMin;
            y1 = (height - yMax - values[time0] * y_scale);
    	    gdk_draw_line(canvas, gc, x0, y0, x1, y1);
            x0 = x1;
            y0 = y1;
        }
    }
}

static void draw_key_entry(GdkPixmap *canvas, GdkGC *gc, PangoLayout *layout, char *content, int x, int y)
{
    int w, h;

    gdk_draw_line(canvas, gc, x+3, y, x+13, y);
    pango_layout_set_text(layout, content, strlen(content));
    pango_layout_get_size(layout, &w, &h);
    x = x - w/PANGO_SCALE;
    pango_layout_set_text(layout, content, strlen(content));
    gdk_draw_layout(canvas, gc, x, y-7, layout);
}
#endif

static void x_wm_graph_draw(GdkPixmap *canvas, GdkGC *gc, ContentionScheduler *cs, PangoLayout *layout)
{
    WmeList *wmes;
    int width, height, range, i = 0;
    double x_scale, y_scale;

    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);
    range = calculate_range(cs->cycle);
    x_scale = (width-xMin-xMax) / (double) range;
    y_scale = (height-yMin-yMax) / (double) (maxAct - minAct);

    for (wmes = cs->wmes; wmes; wmes = wmes->tail)    {
#ifdef WM_CONTINUOUS
        double *act = wmes->head->activation;

        gdk_gc_set_foreground(gc, &colour[wmes->head->colour % 26]);
        draw_plot_line(canvas, gc, act, wmes->head->timestamp, cs->cycle, x_scale, y_scale, height);
        draw_key_entry(canvas, gc, layout, wmes->head->schema_name, width-70, 30+20*i);
#endif
        i++;
    }
}

void x_wm_graph_redraw(GlobalVariables *globals)
{
    ContentionScheduler *cs = globals->cs;
    GtkWidget *da = globals->wm_graph;
    GdkGC *gc = globals->gc;

    GdkPixmap *canvas;
    PangoLayout *layout;
    int width, height, i;
    int x_min, x_max;
    void (*draw_data)();

    gtk_widget_realize(da);
    canvas = (GdkPixmap *)g_object_get_data(G_OBJECT(da), "pixmap");
    gdk_drawable_get_size((GdkDrawable *)canvas, &width, &height);

    x_min = 0;
    x_max = calculate_range(cs->cycle);

    /* Clear the canvas: */
    gdk_gc_set_foreground(gc, &colour[26]);
    gdk_draw_rectangle(canvas, gc, TRUE, 0, 0, width, height);

    /* Draw the graph's grid lines: */
    gdk_gc_set_foreground(gc, &colour[27]); /* Light grey */
    for (i = 1; i < 5; i++) {
        int y = yMin + (i * (height-yMin-yMax)) / 5;
        gdk_draw_line(canvas, gc, xMin, y, width-xMax, y);
    }
    for (i = 1; i < 5; i++) {
        int x = xMin + (i * (width-xMin-xMax)) / 5;
        gdk_draw_line(canvas, gc, x, yMin, x, height-yMax);
    }

    /* Draw the graph's frame: */
    gdk_gc_set_foreground(gc, &colour[0]);
    gdk_draw_rectangle(canvas, gc, FALSE, xMin, yMin, width-xMin-xMax, height-yMin-yMax);

    layout = gtk_widget_create_pango_layout(da, "");
    /* Draw vertical axis tick marks: */
    for (i = 0; i < 6; i++) {
        int w, h, y;
        char buffer[8];

        y = yMin + (i * (height-yMin-yMax)) / 5;
        g_snprintf(buffer, 8, "%5.2f", minAct + (5 - i) * (maxAct - minAct) / 5);
        pango_layout_set_text(layout, buffer, strlen(buffer));
        pango_layout_get_pixel_size(layout, &w, &h);
        gdk_draw_line(canvas, gc, xMin-3, y, xMin, y);
        gdk_draw_layout(canvas, gc, xMin-4-w, y-h/2, layout);
    }
    /* Draw horizontal axis tick marks: */
    for (i = 0; i < 6; i++) {
        int w, h, x;
        char buffer[8];

        x = xMin + (i * (width-xMin-xMax)) / 5;
        g_snprintf(buffer, 8, "%d", x_min + i * (x_max - x_min) / 5);
        pango_layout_set_text(layout, buffer, strlen(buffer));
        pango_layout_get_pixel_size(layout, &w, &h);
        gdk_draw_line(canvas, gc, x, height-yMax+3, x, height-yMax);
        gdk_draw_layout(canvas, gc, x-w/2, height-yMax+5, layout);
    }

    /* Draw the graph's data: */
    draw_data = (void (*)())g_object_get_data(G_OBJECT(da), "draw_function");
    draw_data(canvas, gc, cs, layout);
    g_object_unref(layout);

    /* Return the GC to its former state: */
    gdk_gc_set_foreground(gc, &colour[0]);

    gdk_draw_drawable(da->window, gc, canvas, 0, 0, 0, 0, width, height);
    gtk_widget_queue_draw_area(da, 0, 0, width, height);
    gtkx_flush_events();
}

/******************************************************************************/
/* Section 2: Create menus for the CS graphs: --------------------------------*/

/* Highlight callback functions ----------------------------------------------*/

static void highlight_wme_none(GtkWidget *button, GlobalVariables *globals)
{
    WmeList *tmp;

    for (tmp = globals->cs->wmes; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = FALSE;
    }
    x_wm_graph_redraw(globals);
}

static void highlight_wme(GtkWidget *button, GlobalVariables *globals)
{
    Wme *tmp = g_object_get_data(G_OBJECT(button), "wme");

    tmp->highlighted = !(tmp->highlighted);
    x_wm_graph_redraw(globals);
}

static void highlight_wme_all(GtkWidget *button, GlobalVariables *globals)
{
    WmeList *tmp;

    for (tmp = globals->cs->wmes; tmp != NULL; tmp = tmp->tail) {
	tmp->head->highlighted = TRUE;
    }
    x_wm_graph_redraw(globals);
}

/*----------------------------------------------------------------------------*/

static Boolean wmeMenuGenerate(GtkWidget *item, GdkEvent *event, GlobalVariables *globals)
{
    GtkWidget *menu, *mi;
    WmeList *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label("None");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_wme_none), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    for (tmp = globals->cs->wmes; tmp != NULL; tmp = tmp->tail) {
#ifdef WM_CONTINUOUS
        g_snprintf(buffer, 64, "%c %s", tmp->head->highlighted?'*':' ', tmp->head->schema_name);
#else
        g_snprintf(buffer, 64, "%c %s: %s", tmp->head->highlighted?'*':' ', tmp->head->content->attribute, tmp->head->content->value);
#endif
        mi = gtk_menu_item_new_with_label(buffer);
        g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_wme), globals);
        g_object_set_data(G_OBJECT(mi), "wme", tmp->head);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    g_signal_connect(G_OBJECT(mi), "activate", G_CALLBACK(highlight_wme_all), globals);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

/******************************************************************************/
/* Section 3: Create the canvas and set canvas event handlers ----------------*/

static int wm_graph_event_configure(GtkWidget *da, GdkEventConfigure *event, GlobalVariables *globals)
{
    gtkx_canvas_reallocate(da);
    x_wm_graph_redraw(globals);
    return(1);
}

/*----------------------------------------------------------------------------*/

static GtkWidget *x_wm_create_graph(GlobalVariables *globals, void draw_data())
{
    GtkWidget *da = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(da), "configure_event", G_CALLBACK(wm_graph_event_configure), globals);
    g_signal_connect(G_OBJECT(da), "expose_event", G_CALLBACK(gtkx_canvas_event_expose), NULL);
    g_signal_connect(G_OBJECT(da), "destroy", G_CALLBACK(gtkx_canvas_callback_destroy), NULL);
    g_object_set_data(G_OBJECT(da), "draw_function", draw_data);
    gtk_widget_show(da);
    return(da);
}

/*----------------------------------------------------------------------------*/

static GtkWidget *x_wm_graph_pack_hbox(GtkWidget *page)
{
    GtkWidget *hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);
    return(hbox);
}

/*----------------------------------------------------------------------------*/

static void x_wm_graph_pack_label(GtkWidget *box, char *label)
{
    GtkWidget *item = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(box), GTK_WIDGET(item), FALSE, FALSE, 0);
    gtk_widget_show(item);
}

/*----------------------------------------------------------------------------*/

static void x_wm_graph_pack_highlight_menu(GtkWidget *box, Boolean generate_menu(), GlobalVariables *globals)
{
    GtkWidget *menu_bar = gtk_menu_bar_new();
    GtkWidget *menu = gtk_menu_new();
    GtkWidget *item = gtk_menu_item_new_with_label("Highlight");

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
    g_signal_connect(G_OBJECT(item), "enter_notify_event", G_CALLBACK(generate_menu), globals);
    gtk_widget_show(item);

    gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), item);
    gtk_widget_show(menu_bar);

    gtk_box_pack_end(GTK_BOX(box), GTK_WIDGET(menu_bar), FALSE, FALSE, 0);
}

/*----------------------------------------------------------------------------*/

static void slider_changed(GtkAdjustment *adjustment, double *value)
{
    *value = (double) adjustment->value;
}

static void x_wm_slider_create(GtkWidget *box, char *text, double *value)
{
    GtkWidget *hbox, *label, *slider;
    GtkObject *adj;

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    label = gtk_label_new(text);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    adj = gtk_adjustment_new(*value, 0.0, 1.0, 0.01, 0.01, 0.01);

    slider = gtk_hscale_new(GTK_ADJUSTMENT(adj));
    gtk_scale_set_digits(GTK_SCALE(slider), 2);
    gtk_scale_set_draw_value(GTK_SCALE(slider), TRUE);
    gtk_scale_set_value_pos(GTK_SCALE(slider), GTK_POS_RIGHT);
    gtk_widget_set_usize(slider, 400, -1);
    gtk_box_pack_end(GTK_BOX(hbox), slider, FALSE, FALSE, 0);
    gtk_widget_show(slider);

    g_signal_connect(G_OBJECT(adj), "value_changed", G_CALLBACK(slider_changed), value);
}

static void x_wm_pack_sliders(GtkWidget *box)
{
    x_wm_slider_create(box, "Positive feedback excitation:", &parameter.wm_excite);
    x_wm_slider_create(box, "Negative feedback inhibition:", &parameter.wm_inhibit);
    x_wm_slider_create(box, "WM Noise (s.d.):", &parameter.wm_noise);
    x_wm_slider_create(box, "WM persistence:", &parameter.wm_persistence);
    x_wm_slider_create(box, "WM lateral inhibition:", &parameter.wm_inhibition);
}

/*----------------------------------------------------------------------------*/

GtkWidget *x_wm_graph_create_page(GlobalVariables *globals)
{
    GtkWidget *page, *box;
    
    page = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(page), 5);

    /* WME parameter sliders: */

    box = gtk_vbox_new(TRUE, 5);
    x_wm_pack_sliders(box);
    gtk_box_pack_start(GTK_BOX(page), box, FALSE, FALSE, 0);
    gtk_widget_show(box);

    box = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(page), box, FALSE, FALSE, 4);
    gtk_widget_show(box);

    /* WME graph/menu: */

    box = x_wm_graph_pack_hbox(page);
    x_wm_graph_pack_label(box, "WMEs:");
    x_wm_graph_pack_highlight_menu(box, wmeMenuGenerate, globals);
    globals->wm_graph = x_wm_create_graph(globals, x_wm_graph_draw);
    gtk_box_pack_start(GTK_BOX(page), globals->wm_graph, TRUE, TRUE, 0);

    // FIX ME: Add status bar with cycle count?

    return(page);
}

/******************************************************************************/
